using System;
using DocumentHandle = System.Int32;

namespace Indexer.Engine
{
	using CachedDocumentId = System.UInt32;
	
	public interface IIndexObserver
	{
		void DocumentCountChanged(int documentCount);
		void FoundDocumentCountChanged(int documentCount);
		void OpenProgress(string message, int count);
		bool ShowProgress(int index);
		void ShowMessage(string message);
		void CloseProgress();
		bool InterruptRequested();
	}
	
	public class Index: IDisposable
	{
		private const string RecommendedAnalyzerVersion = "3";
		
		private IndexStore _store = new IndexStore();
		private Explorer.Model.Reader _io = new Explorer.Model.Reader();
		
		public Explorer.Model.Reader IO
		{
			get
			{
				return _io;
			}
		}

		private void InitializeEmptyIndex()
		{
			WriteConfigString("Index/Analyzer.Version", RecommendedAnalyzerVersion);
			Modified = false;
		}
		
		public Index()
		{
			InitializeEmptyIndex();
		}
		
		public Index(IIndexObserver observer)
		{
			_observer = observer;
			InitializeEmptyIndex();
		}
		
		public Index(string fileName)
		{
			Open(fileName);
		}
		
		public Index(IIndexObserver observer, string fileName)
		{
			_observer = observer;
			Open(fileName);
		}
		
		public Worker GetWorker()
		{
			return _store.GetWorker();
		}
		
		private void Open(string fileName)
		{
			_fileName = System.IO.Path.GetFullPath(fileName);
			try
			{
				_store.OpenArchive(new ZipProgress(this, true), _fileName);
			}
			catch(Exception)
			{
				_store.Clear();
				throw;
			}
		}
		
		#region Observer interface
		private IIndexObserver _observer = null;
		public IIndexObserver Observer
		{
			get
			{
				return _observer;
			}
			set
			{
				_observer = value;
			}
		}
		
		private int _documentCount;
		private bool _sendDocumentCount = false;
		private void DocumentCountChanged(int documentCount, bool flush)
		{
			if (flush)
			{
				_sendDocumentCount = false;
				if (_observer != null)
				{
					_observer.DocumentCountChanged(documentCount);
				}
			}
			else
			{
				_documentCount = documentCount;
				_sendDocumentCount = true;
			}
		}

		private int _foundDocumentCount;
		private bool _sendFoundDocumentCount = false;
		private void FoundDocumentCountChanged(int documentCount, bool flush)
		{
			if (flush)
			{
				_sendFoundDocumentCount = false;
				if (_observer != null)
				{
					_observer.FoundDocumentCountChanged(documentCount);
				}
			}
			else
			{
				_foundDocumentCount = documentCount;
				_sendFoundDocumentCount = true;
			}
		}
		
		private void OpenProgress(string message, int count)
		{
			if (_observer != null)
			{
				_observer.OpenProgress(message, count);
				_sendFoundDocumentCount = false;
				_sendDocumentCount = false;
			}
		}
		
		private bool ShowProgress(int index)
		{
			if (_observer != null)
			{
				if (_sendFoundDocumentCount)
				{
					_sendFoundDocumentCount = false;
					_observer.FoundDocumentCountChanged(_foundDocumentCount);
				}
				if (_sendDocumentCount)
				{
					_sendDocumentCount = false;
					_observer.DocumentCountChanged(_documentCount);
				}
				return _observer.ShowProgress(index);
			}
			else
			{
				return true;
			}
		}
		
		private bool InterruptRequested()
		{
			if (_observer != null)
			{
				return _observer.InterruptRequested();
			}
			else
			{
				return false;
			}
		}
		
		private void CloseProgress()
		{
			ShowMessage(null);
			if (_observer != null)
			{
				_observer.CloseProgress();
			}
		}
		
		private System.Collections.Generic.Stack<string> _showMessageStack = new System.Collections.Generic.Stack<string>();
		private string _lastShowMessage;
		private void ShowMessage(string message)
		{
			if (_observer != null)
			{
				_lastShowMessage = message;
				_observer.ShowMessage(message);
			}
		}
		
		private void OpenMessage(string message)
		{
			_showMessageStack.Push(_lastShowMessage);
			ShowMessage(message);
		}
		
		private void CloseMessage()
		{
			ShowMessage(_showMessageStack.Pop());
		}
		#endregion
		
		private string _fileName = null;
		public string FileName
		{
			get
			{
				return _fileName;
			}
		}
		
		#region Configuration stuff
		
		#region Settings I/O
		private const string _settingsFile = "index.ini";
		
		private class Settings: System.Configuration.IniFile
		{
			private IndexStore _store;
			private string _name;
			public Settings(IndexStore store, string name)
			{
				_store = store;
				_name = name;
				base.Initialize();
			}
			
			protected override System.IO.Stream CreateInputStream()
			{
				return _store.OpenInputStream(_name, IndexStore.InputStreamAccessMode.MultipleSequentialScan);
			}
			
			protected override System.IO.Stream CreateOutputStream()
			{
				return _store.CreateOutputStream(_name);
			}
		}
		
		private string ReadConfigString(string key)
		{
			try
			{
				System.Configuration.ISettingsReader config = new Settings(_store, _settingsFile);
				try
				{
					return config.ReadString(key, null);
				}
				finally
				{
					config.Dispose();
				}
			}
			catch(Exception)
			{
				return null;
			}
		}
		
		private bool ReadConfigValue(string key, out int value)
		{
			string buffer = ReadConfigString(key);
			if (!String.IsNullOrEmpty(buffer) && int.TryParse(buffer, out value))
			{
				return true;
			}
			else
			{
				value = 0;
				return false;
			}
		}

		private bool ReadConfigValue(string key, out long value)
		{
			string buffer = ReadConfigString(key);
			if (!String.IsNullOrEmpty(buffer) && long.TryParse(buffer, out value))
			{
				return true;
			}
			else
			{
				value = 0;
				return false;
			}
		}
		
		private void WriteConfigValue(string key, int value)
		{
			WriteConfigString(key, value.ToString());
		}
		
		private void WriteConfigValue(string key, long value)
		{
			WriteConfigString(key, value.ToString());
		}
		
		private void WriteConfigString(string key, string value)
		{
			System.Configuration.ISettingsWriter config = new Settings(_store, _settingsFile);
			try
			{
				if (String.IsNullOrEmpty(value))
				{
					config.Delete(key);
					Modified = true;
				}
				else
				{
					if (ReadConfigString(key) != value)
					{
						config.WriteString(key, value);
						Modified = true;
					}
				}
			}
			finally
			{
				config.Dispose();
			}
		}
		#endregion
		
		public string[] Seeds
		{
			get
			{
				string list = ReadConfigString("Index/Seeds");
				if (String.IsNullOrEmpty(list))
				{
					return new string[0];
				}
				else
				{
					return list.Split(';');
				}
			}
		}

		public string[] Mounts
		{
			get
			{
				string list = ReadConfigString("Index/Mounts");
				if (String.IsNullOrEmpty(list))
				{
					return new string[0];
				}
				else
				{
					return list.Split(';');
				}
			}
		}
		
		public long Size
		{
			get
			{
				return _store.Size;
			}
		}

		public DateTime LastWriteTime
		{
			get
			{
				bool empty = true;
				DateTime result = new DateTime();
				foreach(string name in _store.List())
				{
					DateTime candidate = IndexStore.FileModifiedToDateTime(_store.FileModified(name));
					if (empty)
					{
						result = candidate;
						empty = false;
					}
					else
					{
						if (candidate > result)
						{
							result = candidate;
						}
					}
				}
				if (empty)
				{
					return new System.IO.DirectoryInfo(_store.Folder).LastWriteTime;
				}
				else
				{
					return result;
				}
			}
		}
		
		private void ValidateWordMetrics(bool useConfigurationIfAvailable)
		{
			if (_terms < 0)
			{
				if (useConfigurationIfAvailable &&
				    ReadConfigValue("Metrics/Words.Count", out _terms) &&
				    ReadConfigValue("Metrics/Words.Size", out _termLetters))
				{
					return;
				}
				
				_terms = 0;
				_termLetters = 0;
				
				foreach(string word in GetWordEnumerator())
				{
					_terms++;
					_termLetters += word.Length;
				}
			}
		}
		
		private void SaveMetrics()
		{
			ValidateWordMetrics(false);
			
			WriteConfigValue("Metrics/Words.Count", _terms);
			WriteConfigValue("Metrics/Words.Size", _termLetters);
			WriteConfigValue("Metrics/Documents.Size", DocumentsSize);
		}
		
		public void Prepare()
		{
			ValidateWordMetrics(true);
			Documents.Sorted = true;
		}
		
		public void Initialized(bool commit)
		{
			_store.Initialized();
			if (commit)
			{
				MountFileSystem();
			}
		}
		
		private bool _usesNativeIO = false;
		public bool UsesNativeIO
		{
			get
			{
				return _usesNativeIO;
			}
		}
		
		private void MountFileSystem()
		{
			string[] archives = Mounts;
			if (archives.Length == 0)
			{
				_io.FileSystem = new Explorer.Engine.NativeFileSystem();
				_usesNativeIO = true;
			}
			else
			{
				ShowMessage("Read zip");
				try
				{
					_io.FileSystem = null;
					
					Explorer.Engine.MountFileSystem fileSystem = new Explorer.Engine.MountFileSystem('a');
					try
					{
						fileSystem.Add("a:\\", Explorer.Engine.MountFileSystem.FileSystemAccess.Read).Clear();
						foreach(string archive in archives)
						{
							fileSystem["a:\\"].Add(new Explorer.Engine.ZipFileSystem('c', archive), "c:\\", Explorer.Engine.MountFileSystem.FileSystemMode.Merge, Explorer.Engine.MountFileSystem.FileSystemAccess.Read);
						}
					}
					catch(Exception)
					{
						fileSystem.Dispose();
						throw;
					}
					
					_io.FileSystem = fileSystem;
					_usesNativeIO = false;
					
					AddSeed("a:\\");
				}
				finally
				{
					ShowMessage(null);
				}
			}
		}
		
		public int WordCount
		{
			get
			{
				ValidateWordMetrics(true);
				return _terms;
			}
		}
		
		public int WordLetters
		{
			get
			{
				ValidateWordMetrics(true);
				return _termLetters;
			}
		}
		
		public int DocumentsCount
		{
			get
			{
				return Documents.Count;
			}
		}
		
		public long DocumentsSize
		{
			get
			{
				if (_documentsSize < 0)
				{
					if (!ReadConfigValue("Metrics/Documents.Size", out _documentsSize))
					{
						// not known
						_documentsSize = -1;
					}
				}
				return _documentsSize;
			}
			protected set
			{
				_documentsSize = value;	
			}
		}

		private CachedDocumentId _cachedDocumentIdSeed = NoCacheId;
		private CachedDocumentId LastCachedDocumentId
		{
			get
			{
				if (_cachedDocumentIdSeed == NoCacheId)
				{
					long id;
					if (ReadConfigValue("Index/LastCachedDocumentId", out id))
					{
						_cachedDocumentIdSeed = (CachedDocumentId)id;
					}
				}
				return _cachedDocumentIdSeed;
			}
			set
			{
				_cachedDocumentIdSeed = value;
				WriteConfigValue("Index/LastCachedDocumentId", _cachedDocumentIdSeed);
			}
		}
		
		private CachedDocumentId AllocCachedId()
		{
			// TODO: prevent wrap-around and accidental overwrite
			LastCachedDocumentId = LastCachedDocumentId + 1;
			return LastCachedDocumentId;
		}

		private void ReleaseCachedId(CachedDocumentId cacheId)
		{
			if (cacheId != NoCacheId)
			{
				if (cacheId == LastCachedDocumentId)
				{
					LastCachedDocumentId = LastCachedDocumentId - 1;
				}
				_store.DeleteFile(GetCachedFileName(cacheId));
			}
		}
		
		private bool AddSeed(string seedName)
		{
			seedName = seedName.ToLower();
			string[] currentSeeds = Seeds;
			foreach(string seed in currentSeeds)
			{
				if (seedName == seed)
				{
					// duplicate entry
					return false;
				}
			}
			string newSeeds = String.Join(";", currentSeeds);
			if (String.IsNullOrEmpty(newSeeds))
			{
				newSeeds = seedName;
			}
			else
			{
				newSeeds = newSeeds + ";" + seedName;
			}
			WriteConfigString("Index/Seeds", newSeeds);
			return true;
		}
		
		private bool AddMount(string mountName)
		{
			mountName = mountName.ToLower();
			string[] currentMounts = Mounts;
			foreach(string mount in currentMounts)
			{
				if (mountName == mount)
				{
					// duplicate entry
					return false;
				}
			}
			string newMounts = String.Join(";", currentMounts);
			if (String.IsNullOrEmpty(newMounts))
			{
				newMounts = mountName;
			}
			else
			{
				newMounts = newMounts + ";" + mountName;
			}
			WriteConfigString("Index/Mounts", newMounts);
			return true;
		}
		#endregion

		private const string _pathSortFile = "filenames.sort";
		
		private class DocumentEnumerator: System.Collections.Generic.IEnumerator<DocumentHandle>, System.Collections.Generic.IEnumerable<DocumentHandle>
		{
			private DocumentCollection _collection;
			private int _index;
			private int _count;
			
			public DocumentEnumerator(DocumentCollection collection)
			{
				_collection = collection;
				_count = _collection.Count;
				Reset();
			}
			
			public DocumentHandle Current 
			{
				get 
				{
					return _collection.GetItemHandle(_index);
				}
			}
			
			object System.Collections.IEnumerator.Current 
			{
				get 
				{
					return _collection.GetItemHandle(_index);
				}
			}
			
			public void Dispose()
			{
				_collection = null;
			}
			
			public bool MoveNext()
			{
				_index++;
				if (_index < _count)
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			
			public void Reset()
			{
				_index = -1;
			}
			
			public System.Collections.Generic.IEnumerator<DocumentHandle> GetEnumerator()
			{
				return this;
			}
			
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
			{
				return this;
			}
		}

		[FlagsAttribute]
		public enum DocumentKind
		{
			Binary  = 0,			// file is NOT indexed
			Indexed = 1 << 0,		// file is indexed
			Deleted = 1 << 1,		// file is deleted
			Cached =  1 << 2		// file is cached (e.g. it's translation to text is indexed, not the original file)
		}
		
		public abstract class DocumentCollection: System.Collections.Generic.IEnumerable<DocumentHandle>
		{
			internal abstract DocumentHandle GetItemHandle(int index);
			internal abstract Lucene.Net.Documents.Document GetDocument(DocumentHandle handle);

			internal Explorer.Model.Reader _io;
			
			private DocumentCollection()
			{
			}
			
			protected DocumentCollection(Explorer.Model.Reader io)
			{
				_io = io;
			}
			
			public DocumentKind GetItemKind(int index)
			{
				return GetKind(GetItemHandle(index));
			}
			
			internal virtual DocumentKind GetKind(DocumentHandle handle)
			{
				DocumentFlags flags = GetFlags(handle);
				DocumentKind result;
				if (flags.Indexed)
				{
					result = DocumentKind.Indexed;
				}
				else
				{
					result = DocumentKind.Binary;
				}
				if (flags.Cached)
				{
					result |= DocumentKind.Cached;
				}
				if (!_io.FileExists(GetPath(handle)))
				{
					result |= DocumentKind.Deleted;
				}
				return result;
			}
			
			internal virtual DocumentFlags GetFlags(DocumentHandle handle)
			{
				return new DocumentFlags(GetDocument(handle).GetBinaryValue(DocumentFields.Flags));
			}
			
			protected virtual Lucene.Net.Documents.Document GetItemDocument(int index)
			{
				return GetDocument(GetItemHandle(index));
			}
			
			internal string GetItemPath(int index)
			{
				return GetPath(GetItemHandle(index));
			}
			
			internal virtual string GetPath(DocumentHandle handle)
			{
				return GetDocument(handle).Get(DocumentFields.Path);
			}

			public virtual bool Contains(DocumentHandle handle)
			{
				return IndexOf(handle) >= 0;
			}
			
			internal abstract int IndexOf(DocumentHandle handle);
			protected abstract int GetCount();

			public string this[int index]
			{
				get
				{
					return GetItemPath(index);
				}
			}
			
			public int Count
			{
				get
				{
					return GetCount();
				}
			}
			
			protected bool _sorted = false;
			public bool Sorted
			{
				get
				{
					return _sorted;
				}
				set
				{
					if (_sorted != value)
					{
						if (value)
						{
							ApplySort();
							_sorted = true;
						}
						else
						{
							RemoveSort();
							_sorted = false;
						}
					}
				}
			}
			
			protected virtual void ApplySort()
			{
				throw new NotImplementedException();
			}
			
			protected virtual void RemoveSort()
			{
			}
			
			protected virtual System.Collections.Generic.IEnumerator<DocumentHandle> CreateDocumentEnumerator()
			{
				return new DocumentEnumerator(this);
			}
			
			public System.Collections.Generic.IEnumerator<DocumentHandle> GetEnumerator()
			{
				return CreateDocumentEnumerator();
			}
			
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
			{
				return CreateDocumentEnumerator();
			}
		}
		
		public class AllDocumentsCollection: DocumentCollection
		{
			private System.Collections.Generic.List<DocumentHandle> _handles = null;
			private System.Collections.Generic.List<string> _pathCache = null;
			private System.Collections.Generic.List<int> _handlePositions = null;
			private System.Collections.MultiBitArray _kindCache = null;
			
			internal override int IndexOf(DocumentHandle handle)
			{
				if (_handlePositions == null)
				{
					return (int)handle;
				}
				else
				{
					return _handlePositions[(int)handle];
				}
			}
			
			internal override Lucene.Net.Documents.Document GetDocument(DocumentHandle handle)
			{
				return _parent.Reader.Document(handle);
			}

			internal override DocumentHandle GetItemHandle(int index)
			{
				return Handles[index];
			}
			
			private Index _parent;
			public AllDocumentsCollection(Index parent): base(parent._io)
			{
				_parent = parent;
			}
			
			public override bool Contains(DocumentHandle handle)
			{
				return true;
			}
			
			protected override int GetCount()
			{
				if (_handles == null)
				{
					return _parent.Reader.NumDocs();
				}
				else
				{
					return _handles.Count;
				}
			}
			
			internal override string GetPath(DocumentHandle handle)
			{
				PrepareHandles();
				return _pathCache[(int)handle];
			}
			
			internal override DocumentKind GetKind(DocumentHandle handle)
			{
				const int NotDefined = 0;
				const int IndexedAndExists = 1;
				const int IndexedAndNotExists = 2;
				const int NotIndexedAndExists = 3;
				const int NotIndexedAndNotExists = 4;
				const int CachedAndIndexedAndExists = 5;
				const int CachedAndIndexedAndNotExists = 6;
				const int MaxValue = CachedAndIndexedAndNotExists;

				if (_kindCache == null)
				{
					_kindCache = new System.Collections.MultiBitArray(Count, MaxValue);
				}

				switch (_kindCache[(int)handle])
				{
					case NotDefined:
						DocumentKind result = base.GetKind(handle);
						switch (result)
						{
							case DocumentKind.Indexed:
								_kindCache[(int)handle] = IndexedAndExists;
								break;
								
							case DocumentKind.Indexed | DocumentKind.Deleted:
								_kindCache[(int)handle] = IndexedAndNotExists;
								break;
								
							case DocumentKind.Binary:
								_kindCache[(int)handle] = NotIndexedAndExists;
								break;
								
							case DocumentKind.Binary | DocumentKind.Deleted:
								_kindCache[(int)handle] = NotIndexedAndNotExists;
								break;
								
							case DocumentKind.Indexed | DocumentKind.Cached:
								_kindCache[(int)handle] = CachedAndIndexedAndExists;
								break;
								
							case DocumentKind.Indexed | DocumentKind.Cached | DocumentKind.Deleted:
								_kindCache[(int)handle] = CachedAndIndexedAndNotExists;
								break;
								
							default:
								// TODO: implement
								throw new NotSupportedException();
								
						}
						return result;
						
					case IndexedAndExists:
						return DocumentKind.Indexed;
						
					case IndexedAndNotExists:
						return DocumentKind.Indexed | DocumentKind.Deleted;
						
					case NotIndexedAndExists:
						return DocumentKind.Binary;
						
					case NotIndexedAndNotExists:
						return DocumentKind.Binary | DocumentKind.Deleted;
						
					case CachedAndIndexedAndExists:
						return DocumentKind.Indexed | DocumentKind.Cached;
						
					case CachedAndIndexedAndNotExists:
						return DocumentKind.Indexed | DocumentKind.Cached | DocumentKind.Deleted;
						
					default:
						// TODO: assert(false)
						throw new NotSupportedException();
				}
			}

			private void PrepareHandles()
			{
				if (_handles == null)
				{
					int count = GetCount();
					_handles = new System.Collections.Generic.List<DocumentHandle>(count);
					_pathCache = new System.Collections.Generic.List<string>(count);
					for(int i = 0; i < count; i++)
					{
						_handles.Add((DocumentHandle)i);
						_pathCache.Add(base.GetPath((DocumentHandle)i));
					}
				}
			}
			
			private System.Collections.Generic.List<DocumentHandle> Handles
			{
				get
				{
					PrepareHandles();
					return _handles;
				}
			}

			public static int Compare(string sx, string sy)
			{
				int result = String.Compare(System.IO.Path.GetDirectoryName(sx), System.IO.Path.GetDirectoryName(sy), true);
				if (result == 0)
				{
					result = String.Compare(System.IO.Path.GetFileName(sx), System.IO.Path.GetFileName(sy), true);
				}
				return result;
			}
			
			private class HandleComparer: System.Collections.Generic.Comparer<DocumentHandle>
			{
				private AllDocumentsCollection _collection;
				public HandleComparer(AllDocumentsCollection collection)
				{
					_collection = collection;
				}
				
				public override int Compare(DocumentHandle x, DocumentHandle y)
				{
					return AllDocumentsCollection.Compare(_collection.GetPath(x), _collection.GetPath(y));
				}
			}
			
			protected override void ApplySort()
			{
				int count = Handles.Count;
				if (count == 0)
				{
					return;
				}
				if (_parent._store.FileLength(_pathSortFile) == count * sizeof(Int32))
				{
					System.IO.Stream raw = _parent._store.OpenInputStream(_pathSortFile, IndexStore.InputStreamAccessMode.SingleShotSequentialScan);
					try
					{
						System.IO.BinaryReader r = new System.IO.BinaryReader(raw);
						try
						{
							for(int i = 0; i < count; i++)
							{
								Handles[i] = (DocumentHandle)r.ReadInt32();
							}
						}
						finally
						{
							r.Close();
						}
					}
					finally
					{
						raw.Dispose();
					}
				}
				else
				{
					_parent.OpenMessage("Sort" /* TODO: localize */);
					try
					{
						Handles.Sort(new HandleComparer(this));
						
						System.IO.Stream raw = _parent._store.CreateOutputStream(_pathSortFile);
						try
						{
							System.IO.BinaryWriter w = new System.IO.BinaryWriter(raw);
							try
							{
								for(int i = 0; i < count; i++)
								{
									w.Write((int)Handles[i]);
								}
							}
							finally
							{
								w.Close();
							}
						}
						finally
						{
							raw.Dispose();
						}
						_parent.Modified = true;
					}
					finally
					{
						_parent.CloseMessage();
					}
				}
				_handlePositions = new System.Collections.Generic.List<int>(count);
				for(int position = 0; position < count; position++)
				{
					_handlePositions.Add(0);
				}
				for(int position = 0; position < count; position++)
				{
					_handlePositions[(int)Handles[position]] = position;
				}
			}
			
			protected override void RemoveSort()
			{
				_handlePositions = null;
			}
			
			protected override System.Collections.Generic.IEnumerator<DocumentHandle> CreateDocumentEnumerator()
			{
				return Handles.GetEnumerator();
			}
		}
		
		public class SearchResultsCollection: DocumentCollection
		{
			private Lucene.Net.Search.Hits _parent;
			internal SearchResultsCollection(Lucene.Net.Search.Hits parent, Explorer.Model.Reader io): base(io)
			{
				_parent = parent;
			}
			
			protected override int GetCount()
			{
				return _parent.Length();
			}
			
			internal override Lucene.Net.Documents.Document GetDocument(DocumentHandle handle)
			{
				// TODO: is it ok?
				throw new NotImplementedException();
			}
			
			internal override int IndexOf(DocumentHandle handle)
			{
				// TODO: is it ok?
				throw new NotImplementedException();
			}
			
			protected override Lucene.Net.Documents.Document GetItemDocument(int index)
			{
				return _parent.Doc(index);
			}
			
			internal override DocumentHandle GetItemHandle(int index)
			{
				return _parent.Id(index);
			}
		}

		private class FilterEnumerator: System.Collections.Generic.IEnumerator<DocumentHandle>, System.Collections.Generic.IEnumerable<DocumentHandle>
		{
			private System.Collections.BitArray _handleBits;
			private int _index;
			private int _toGo;
			private int _count;
			
			public FilterEnumerator(System.Collections.BitArray handleBits, int count)
			{
				_handleBits = handleBits;
				_count = count;
				Reset();
			}
			
			public DocumentHandle Current 
			{
				get 
				{
					return (DocumentHandle)_index;
				}
			}
			
			object System.Collections.IEnumerator.Current 
			{
				get 
				{
					return (DocumentHandle)_index;
				}
			}
			
			public void Dispose()
			{
				_handleBits = null;
			}
			
			public bool MoveNext()
			{
				if (_toGo <= 0)
				{
					return false;
				}
				_index++;
				while (!_handleBits.Get(_index))
				{
					_index++;
				}
				_toGo--;
				return true;
			}
			
			public void Reset()
			{
				_index = -1;
				_toGo = _count;
			}
			
			public System.Collections.Generic.IEnumerator<DocumentHandle> GetEnumerator()
			{
				return this;
			}
			
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
			{
				return this;
			}
		}
		
		public class FilterCollection: DocumentCollection
		{
			private DocumentCollection _masterCollection;
			private System.Collections.BitArray _handleBits;
			private int _count = 0;
			
			public FilterCollection(DocumentCollection masterCollection): base(masterCollection._io)
			{
				_masterCollection = masterCollection;
				_handleBits = new System.Collections.BitArray(_masterCollection.Count);
			}
			
			private class Filter: Lucene.Net.Search.Filter
			{
				private System.Collections.BitArray _bits;
				public Filter(System.Collections.BitArray bits)
				{
					_bits = bits;
				}
				
				public override System.Collections.BitArray Bits(Lucene.Net.Index.IndexReader reader)
				{
					return _bits;
				}
			}
			
			public Lucene.Net.Search.Filter CreateFilter()
			{
				return new Filter(_handleBits);
			}
			
			protected override int GetCount()
			{
				return _count;
			}
			
			internal override int IndexOf(DocumentHandle handle)
			{
				// TODO: is it ok?
				throw new NotImplementedException();
			}
			
			public override bool Contains(DocumentHandle handle)
			{
				return _handleBits.Get((int)handle);
			}
			
			internal override Lucene.Net.Documents.Document GetDocument(DocumentHandle handle)
			{
				return _masterCollection.GetDocument(handle);
			}
			
			internal override DocumentHandle GetItemHandle(int index)
			{
				// TODO: is it ok?
				throw new NotImplementedException();
			}
			
			protected override System.Collections.Generic.IEnumerator<DocumentHandle> CreateDocumentEnumerator()
			{
				return new FilterEnumerator(_handleBits, _count);
			}
			
			public void Or(DocumentHandle handle)
			{
				if (!_handleBits.Get(handle))
				{
					_handleBits.Set(handle, true);
					_count++;
				}
			}
			
			public void And(FilterCollection handles)
			{
				if (_count == 0)
				{
					return;
				}
				int count = _handleBits.Count;
				for(int i = 0; i < count; i++)
				{
					if (!handles._handleBits.Get(i))
					{
						if (_handleBits.Get(i))
						{
							_handleBits.Set(i, false);
							_count--;
							if (_count == 0)
							{
								return;
							}
						}
					}
				}
			}
		}
		
		public class CustomCollection: DocumentCollection
		{
			private DocumentCollection _masterCollection;
			private System.Collections.Generic.List<DocumentHandle> _handles = new System.Collections.Generic.List<DocumentHandle>();
			
			public CustomCollection(): base(null)
			{
				_masterCollection = null;
			}
			
			public CustomCollection(DocumentCollection masterCollection): base(masterCollection._io)
			{
				_masterCollection = masterCollection;
			}
			
			protected override int GetCount()
			{
				if (_masterBits == null)
				{
					return _handles.Count;
				}
				else
				{
					return _count;
				}
			}
			
			internal override int IndexOf(DocumentHandle handle)
			{
				return _handles.IndexOf(handle);
			}
			
			internal override int GetItemHandle(int index)
			{
				return _handles[index];
			}
			
			internal override Lucene.Net.Documents.Document GetDocument(DocumentHandle handle)
			{
				return _masterCollection.GetDocument(handle);
			}
			
			internal override DocumentKind GetKind(DocumentHandle handle)
			{
				return _masterCollection.GetKind(handle);
			}
			
			private System.Collections.BitArray _masterBits = null;
			private int _count;
			public void BeginUpdate()
			{
				_masterBits = new System.Collections.BitArray(_masterCollection.Count);
				_count = 0;
				_sorted = true;
			}
			
			public void Add(DocumentHandle handle)
			{
				_masterBits.Set(_masterCollection.IndexOf(handle), true);
				_count++;
			}
			
			public void Or(DocumentHandle handle)
			{
				int index = _masterCollection.IndexOf(handle);
				if (!_masterBits.Get(index))
				{
					_masterBits.Set(index, true);
					_count++;
				}
			}
			
			public void EndUpdate()
			{
				_handles = new System.Collections.Generic.List<DocumentHandle>(_count);
				
				if (_count > 0)
				{
					for(int i = 0; i < _masterBits.Count; i++)
					{
						if (_masterBits[i])
						{
							_handles.Add(_masterCollection.GetItemHandle(i));
							_count--;
							if (_count == 0)
							{
								break;
							}
						}
					}
				}
				_masterBits = null;
			}
			
			protected override System.Collections.Generic.IEnumerator<DocumentHandle> CreateDocumentEnumerator()
			{
				return _handles.GetEnumerator();
			}
		}
		
		private int _terms = -1;
		private int _termLetters = -1;
		private long _documentsSize = -1;
		
		private DocumentCollection _documents = null;
		public DocumentCollection Documents
		{
			get
			{
				if (_documents == null)
				{
					_documents = new AllDocumentsCollection(this);
					// _documents.Sorted = true;
				}
				return _documents;
			}
		}

		private void InvalidateDocumentCache(bool invalidatePersistentData)
		{
			_documents = null;
			_terms = -1;
			_termLetters = -1;
			
			if (invalidatePersistentData)
			{
				_store.DeleteFile(_pathSortFile);
			}
			
			//_documentsSize = -1;
		}
		
		private void Detach()
		{
			Detach(false);
		}
		
		private void Detach(bool invalidatePersistentData)
		{
			Reader = null;
			Writer = null;
			InvalidateDocumentCache(invalidatePersistentData);
		}
		
		private Lucene.Net.Index.IndexModifier _writer = null;
		private Lucene.Net.Index.IndexReader _reader = null;
		private int _updateCounter = 0;
		
		private bool IsEmpty
		{
			get
			{
				return !_store.FileExists("segments");
			}
		}
		
		private Lucene.Net.Index.IndexReader Reader
		{
			get
			{
				if (_reader == null)
				{
					if (IsEmpty)
					{
						bool oldModified = Modified;
						BeginUpdate();
						EndUpdate();
						Modified = oldModified;
					}
					_reader = Lucene.Net.Index.IndexReader.Open(_store);
				}
				return _reader;
			}
			set
			{
				if (value != null)
				{
					throw new ArgumentException();
				}
				if (_reader != null)
				{
					_reader.Close();
					_reader = null;
				}
			}
		}
		
		private class DebugInfoStream: System.IO.StreamWriter
		{
			private Index _index;
			
			public DebugInfoStream(Index index): base(System.IO.Stream.Null)
			{
				_index = index;
			}

			public override void Write(string value)
			{
				#if DEBUG
				System.Diagnostics.Debug.Write(value);
				#endif
				if (value == "merging segments")
				{
					_index.OpenMessage("Optimize" /* TODO: localize */);
				}
				else if (value == "segments merged")
				{
					_index.CloseMessage();
				}
			}
			
			public override void WriteLine(string value)
			{
				Write(value);
				#if DEBUG
				System.Diagnostics.Debug.WriteLine("");
				#endif
			}
			
			public override System.Text.Encoding Encoding 
			{
				get 
				{ 
					return System.Text.Encoding.Default; 
				}
			}
		}
		
		private System.IO.StreamWriter _infoStream = null;
		private System.IO.StreamWriter InfoStream
		{
			get
			{
				if (_infoStream == null)
				{
					_infoStream = new DebugInfoStream(this);
				}
				return _infoStream;
			}
		}
		
		private class HookedIndexWriter: Lucene.Net.Index.IndexWriter
		{
			private Index _index;
			
			public HookedIndexWriter(Index index, Lucene.Net.Store.Directory d, Lucene.Net.Analysis.Analyzer a, bool create) : base(d, a, create)
			{
				_index = index;
			}
		}
		
		private class ConfiguredIndexModifier: Lucene.Net.Index.IndexModifier
		{
			private Index _index;
			
			public ConfiguredIndexModifier(Index index, Lucene.Net.Store.Directory dir, Lucene.Net.Analysis.Analyzer analyzer, bool create)
				: base(dir, analyzer, create)
			{
				_index = index;
				if (indexWriter != null)
				{
					indexWriter.Close();
					indexWriter = null;
				}
				indexWriter = CreateAndConfigureWriter(directory, analyzer, create);
			}
			
			private Lucene.Net.Index.IndexWriter CreateAndConfigureWriter(Lucene.Net.Store.Directory directory, Lucene.Net.Analysis.Analyzer analyzer, bool create)
			{
				// This is needed to override implicit word-count limit
				maxFieldLength = int.MaxValue;
				infoStream = _index.InfoStream;
				
				Lucene.Net.Index.IndexWriter result = new HookedIndexWriter(_index, directory, analyzer, create);
				result.SetInfoStream(infoStream);
				result.SetUseCompoundFile(useCompoundFile);
				result.SetMaxBufferedDocs(maxBufferedDocs);
				result.SetMaxFieldLength(maxFieldLength);
				result.SetMergeFactor(mergeFactor);
				return result;
			}
			
			protected override void CreateIndexWriter()
			{
				// This code is merely copy/paste from Lucene.Net.Index.IndexModifier.CreateIndexWriter()
				// modified to instantiate "HookedIndexWriter" and set infinite field length limit
				if (indexWriter == null)
				{
					if (indexReader != null)
					{
						indexReader.Close();
						indexReader = null;
					}
					indexWriter = CreateAndConfigureWriter(directory, analyzer, false);
				}				
			}
		}
		
		private Lucene.Net.Index.IndexModifier Writer
		{
			get
			{
				if (_writer == null)
				{
					// preserve settings. Writer may clear working folder if empty
					string settings;
					try
					{
						System.IO.TextReader r = System.Text.FileStream.CreateReader(_store.OpenInputStream(_settingsFile, IndexStore.InputStreamAccessMode.SingleShotSequentialScan));
						try
						{
							settings = r.ReadToEnd();
						}
						finally
						{
							r.Dispose();
						}
					}
					catch(Exception)
					{
						settings = "";
					}
					
					_writer = new ConfiguredIndexModifier(this, _store, CreateAnalyzer(AnalyzerMode.TokenizeDocument), IsEmpty);

					System.IO.TextWriter w = new System.IO.StreamWriter(_store.CreateOutputStream(_settingsFile), System.Text.Encoding.Unicode /* TODO: choose ASCII if sufficient*/);
					try
					{
						w.Write(settings);
					}
					finally
					{
						w.Dispose();
					}
				}
				return _writer;
			}
			set
			{
				if (value != null)
				{
					throw new ArgumentException();
				}
				if (_writer != null)
				{
					_writer.Close();
					_writer = null;
				}
			}
		}
		
		public void BeginUpdate()
		{
			if (_updateCounter == 0)
			{
				Reader = null;
				// Following hit just ensures that writer is created
				if (Writer == null)
				{
					throw new NullReferenceException();
				}
			}
			_updateCounter++;
		}
		
		public void EndUpdate()
		{
			UpdaterStop();
				
			if (_updateCounter == 1)
			{
				if (Modified)
				{
					DocumentCountChanged(Writer.DocCount(), true);
					Writer.Optimize();
				}
				Writer = null;
				InvalidateDocumentCache(Modified);
				SaveMetrics();
				Prepare();

				_updateCounter = 0;
			}
			else
			{
				_updateCounter--;
			}
		}

		#region Tokenizer
		private enum AnalyzerMode
		{
			TokenizeQuery,
			TokenizeDocument
		}
		
		private class OldSourceCodeTokenizer: Lucene.Net.Analysis.LetterTokenizer
		{
			public OldSourceCodeTokenizer(System.IO.TextReader in_Renamed) : base(in_Renamed)
			{
			}
			
			protected override bool IsTokenChar(char c)
			{
				return Char.IsLetterOrDigit(c);
			}
			
			protected override char Normalize(char c)
			{
				return Char.ToLower(c);
			}
		}
		
		private abstract class AnalyzerBase: Lucene.Net.Analysis.Analyzer
		{
			public AnalyzerMode Mode;
			
			public virtual bool SupportsPhrases(AnalyzerMode mode)
			{
				return true;
			}
		}
		
		private class OldSourceCodeAnalyzer: AnalyzerBase
		{
			public override Lucene.Net.Analysis.TokenStream TokenStream(string fieldName, System.IO.TextReader reader)
			{
				return new OldSourceCodeTokenizer(reader);
			}
		}

		private class SourceCodeTokenizer: Lucene.Net.Analysis.LetterTokenizer
		{
			public SourceCodeTokenizer(System.IO.TextReader in_Renamed) : base(in_Renamed)
			{
			}
			
			protected override bool IsTokenChar(char c)
			{
				return (c == '_') || Char.IsLetterOrDigit(c);
			}
			
			protected override char Normalize(char c)
			{
				return Char.ToLower(c);
			}
		}
		
		private class SourceCodeAnalyzer: AnalyzerBase
		{
			public override Lucene.Net.Analysis.TokenStream TokenStream(string fieldName, System.IO.TextReader reader)
			{
				return new SourceCodeTokenizer(reader);
			}
		}
		
		private class ReducingSourceCodeTokenizer: SourceCodeTokenizer
		{
			private class CaseSensitiveSortedList: System.Collections.Generic.List<string>, System.Collections.Generic.IComparer<string>
			{
				public int Compare(string x, string y)
				{
					return String.Compare(x, y, false);
				}
				
				/// <summary>
				/// Updates the string dictionary if needed
				/// </summary>
				/// <param name="value">new value to add</param>
				/// <returns>true - value was added, false - value was already in</returns>
				public bool AddIfNotThere(string value)
				{
					int index = BinarySearch(value, this);
 					if (index < 0)
 					{
 						Insert(~index, value);
 						return true;
 					}
 					else
 					{
 						return false;
 					}
 				}
			}
			
			private CaseSensitiveSortedList _duplicates;
			private int _maxDigits;
			
			public ReducingSourceCodeTokenizer(System.IO.TextReader in_Renamed, bool skipDuplicates, int maxDigits) : base(in_Renamed)
			{
				if (skipDuplicates)
				{
					_duplicates = new CaseSensitiveSortedList();
				}
				else
				{
					_duplicates = null;
				}
				_maxDigits = maxDigits;
			}
			
			private static string _aChars = "aeiouy";
			private static string _mChars = "bcdfghjklmnpqrstvwxz";

			// TODO: following code would reject e.g. "RHttpSession" because of the "rhttp" prefix
			private static bool SoundsLikeWord(string token)
			{
				char repeater = char.MinValue;
				int repeatCount = 0;
				int countOfAs = 0;
				int countOfMs = 0;
				
				for(int i = 0; i < token.Length; i++)
				{
					char c = token[i];
					if (c == repeater)
					{
						repeatCount++;
						if (repeatCount >= 4 /* TODO: TUNE: sequence of 4 repeated characters is not a word */)
						{
							return false;
						}
					}
					else
					{
						repeater = c;
						repeatCount = 1;
					}
					c = char.ToLowerInvariant(c);
					if (_aChars.IndexOf(c) >= 0)
					{
						countOfAs++;
						countOfMs = 0;
						if (countOfAs >= 5 /* TODO: TUNE: sequence of 5 repeated "A"-class characters is not a word */)
						{
							return false;
						}
					}
					else if (_mChars.IndexOf(c) >= 0)
					{
						countOfMs++;
						countOfAs = 0;
						if (countOfMs >= 5 /* TODO: TUNE: sequence of 5 repeated "M"-class characters is not a word */)
						{
							return false;
						}
					}
					else
					{
						countOfAs = 0;
						countOfMs = 0;
					}
				}
				return true;
			}
			
			protected virtual bool IsAcceptable(string token)
			{
			case1:
				// reject sequences of more than "_maxDigits" hex digits starting with a digit
				if (Char.IsDigit(token[0]))
				{
					if (token.Length > _maxDigits)
					{
						for(int i = 1; i < token.Length; i++)
						{
							char c = Char.ToUpper(token[i]);
							if ( ((c >= '0') && (c <= '9')) || ((c >= 'A') && (c <= 'F')) )
							{
								continue;
							}
							else
							{
								goto case2;
							}
						}
						return false;
					}
				}
			case2:
/* TODO:
				if (!SoundsLikeWord(token))
				{
					return false;
				}
*/				
			case3:
				return true;
			}
			
			public override Lucene.Net.Analysis.Token Next()
			{
				Lucene.Net.Analysis.Token result;
				while (true)
				{
					result = base.Next();
					if (result == null)
					{
						return null;
					}
					
					string word = result.TermText();
					if (IsAcceptable(word))
					{
						if (_duplicates == null)
						{
							return result;
						}
						else
						{
							if (_duplicates.AddIfNotThere(word))
							{
								return result;
							}
							else
							{
								// token was already recognized some time before -> skip it to
								// reduce token position table
							}
						}
					}
				}
			}
			
			public override void Close()
			{
				base.Close();
				if (_duplicates != null)
				{
					_duplicates.Clear();
					_duplicates = null;
				}
			}
		}

		private class ReducingSourceCodeAnalyzer: AnalyzerBase
		{
			protected bool CanSkipDuplicates(AnalyzerMode mode)
			{
				switch (mode)
				{
					case AnalyzerMode.TokenizeDocument:
						// TODO: skipping terms makes index smaller, but eliminates possibility
						// to do efficient phrase queries as the relative term positions are lost
						return true;
						
					default:
						return false;
				}
			}
			
			public override Lucene.Net.Analysis.TokenStream TokenStream(string fieldName, System.IO.TextReader reader)
			{
				return new ReducingSourceCodeTokenizer(reader, CanSkipDuplicates(Mode), 4 /* max number of digits in a word */);
			}
			
			public override bool SupportsPhrases(AnalyzerMode mode)
			{
				return !CanSkipDuplicates(mode);
			}
		}

		private class ReducingSourceCodeAnalyzerV2: ReducingSourceCodeAnalyzer
		{
			public override Lucene.Net.Analysis.TokenStream TokenStream(string fieldName, System.IO.TextReader reader)
			{
				return new ReducingSourceCodeTokenizer(reader, CanSkipDuplicates(Mode), 8 /* max number of digits in a word */);
			}
		}
		
		private AnalyzerBase CreateAnalyzer(AnalyzerMode mode)
		{
			AnalyzerBase result = null;
			
			switch (ReadConfigString("Index/Analyzer.Version"))
			{
				case "3":
					result = new ReducingSourceCodeAnalyzerV2();
					break;
					
				case "2":
					result = new ReducingSourceCodeAnalyzer();
					break;
					
				case "1":
					result = new SourceCodeAnalyzer();
					break;
					
				default:
					result = new OldSourceCodeAnalyzer();
					break;
			}
			
			if (result == null)
			{
				throw new NotImplementedException();
			}
			result.Mode = mode;
			return result;
		}
		
		private Lucene.Net.Index.Term CreateSearchStartTerm(string field, string searchText, bool useTermSeek)
		{
			if (useTermSeek)
			{
				return new Lucene.Net.Index.Term(field, searchText);
			}
			else
			{
				return new Lucene.Net.Index.Term(field, "");
			}
		}
		#endregion
		
		#region Searching
		public enum SetCombineMode
		{
			None,
			Or,
			And,
			AndNot
		}
		
		private struct SearchContext
		{
			public string Text;
			public System.IO.TextReader TextStream;
			public string Phrase;
			public bool CaseSensitive;
			public bool WholeWords;
			public bool AllOccurencies;
			public int StartIndex;
			public int[] Offsets;
			
			public void AddOffset(int offset)
			{
				if (Offsets == null)
				{
					Offsets = new int[]{offset};
				}
				else
				{
					Array.Resize(ref Offsets, Offsets.Length + 1);
					Offsets[Offsets.Length - 1] = offset;
				}
			}
		}
		
		private static bool WordPartAt(Lucene.Net.Analysis.Analyzer analyzer, string text, int index, bool defaultValue)
		{
			if ((index < 0) || (index >= text.Length))
			{
				return defaultValue;
			}
			Lucene.Net.Analysis.TokenStream tokens = analyzer.TokenStream(null, new System.IO.StringReader(text[index].ToString()));
			try
			{
				return tokens.Next() != null;
			}
			finally
			{
				tokens.Close();
			}
		}
		
		private static bool IdentifierPartAt(string text, int index, bool defaultValue)
		{
			if ((index < 0) || (index >= text.Length))
			{
				return defaultValue;
			}
			char c = text[index];
			return (c == '_') || Char.IsLetterOrDigit(c);
		}
		
		private static int Search(ref SearchContext context)
		{
			if (context.TextStream != null)
			{
				if (context.AllOccurencies)
				{
					context.Text = context.TextStream.ReadToEnd();
					return InternalSearch(ref context);
				}
				else
				{
					while ((context.Text = context.TextStream.ReadLine()) != null)
					{
						context.StartIndex = 0;
						int result = InternalSearch(ref context);
						if (result > 0)
						{
							return result;
						}
					}
					return 0;
				}
			}
			else
			{
				return InternalSearch(ref context);
			}
		}
		
		/// <summary>
		/// works only upon the context.Text
		/// </summary>
		/// <param name="context"></param>
		/// <returns></returns>
		private static int InternalSearch(ref SearchContext context)
		{
			System.StringComparison options;
			if (context.CaseSensitive)
			{
				options = StringComparison.InvariantCulture;
			}
			else
			{
				options = StringComparison.InvariantCultureIgnoreCase;
			}
			int i;
			for(;;)
			{
				if (context.StartIndex >= context.Text.Length)
				{
					return 0;
				}
				i = context.Text.IndexOf(context.Phrase, context.StartIndex, options);
				if (i < 0)
				{
					return 0;
				}
				if (context.WholeWords)
				{
					bool ok = true;
					
					// check start of phrase
					if (IdentifierPartAt(context.Phrase, 0, false))
					{
						if (IdentifierPartAt(context.Text, i - 1, false))
						{
							ok = false;
						}
					}
					
					if (ok)
					{
						// check end of phrase
						int len = context.Phrase.Length;
						if (IdentifierPartAt(context.Phrase, len - 1, false))
						{
							if (IdentifierPartAt(context.Text, i + len, false))
							{
								ok = false;
							}
						}
					}
					if (ok)
					{
						// match found
						break;
					}
					else
					{
						context.StartIndex = i + 1;
					}
				}
				else
				{
					// match found
					break;
				}
			}
			context.AddOffset(i);
			if (context.AllOccurencies)
			{
				context.StartIndex = i + context.Phrase.Length;
				return 1 + Search(ref context);
			}
			else
			{
				return 1;
			}
		}

		public struct SearchResult
		{
			public int Row;
			public int Column;
			public int Length;
			
			public SearchResult(int row, int column, int length)
			{
				Row = row;
				Column = column;
				Length = length;
			}
		}
		
		public abstract class Query
		{
			public abstract string[] Phrases { get; }
			public DocumentCollection Results;
			
			public static Query New(string phrase, bool caseSensitive, bool wholeWords, System.Text.StringFilter fileFilter)
			{
				SimpleQuery result = new SimpleQuery();
				result.Phrase = phrase;
				result.CaseSensitive = caseSensitive;
				result.WholeWords = wholeWords;
				result.FileFilter = fileFilter;
				return result;
			}

			public static Query New(Query left, SetCombineMode mode, string phrase, bool caseSensitive, bool wholeWords, System.Text.StringFilter fileFilter)
			{
				return New(left, mode, New(phrase, caseSensitive, wholeWords, fileFilter));
			}
			
			public static Query New(Query left, SetCombineMode mode, Query right)
			{
				switch (mode)
				{
					case SetCombineMode.None:
						return right;
						
					case SetCombineMode.And:
						if ((left == null) || (right == null))
						{
							return null;
						}
						else
						{
							CombinedQuery result = new CombinedQuery();
							result.CombineMode = mode;
							result.Left = left;
							result.LeftAddsToResultSet = true;
							result.Right = right;
							result.RightAddsToResultSet = true;
							return result;
						}
						
					case SetCombineMode.Or:
						if (left == null)
						{
							return right;
						}
						else if (right == null)
						{
							return left;
						}
						else
						{
							CombinedQuery result = new CombinedQuery();
							result.CombineMode = mode;
							result.Left = left;
							result.LeftAddsToResultSet = true;
							result.Right = right;
							result.RightAddsToResultSet = true;
							return result;
						}
						
					case SetCombineMode.AndNot:
						if (left == null)
						{
							return null;
						}
						else if (right == null)
						{
							return left;
						}
						else
						{
							CombinedQuery result = new CombinedQuery();
							result.CombineMode = mode;
							result.Left = left;
							result.LeftAddsToResultSet = true;
							result.Right = right;
							result.RightAddsToResultSet = false;
							return result;
						}
						
					default:
						// TODO:
						throw new NotImplementedException();
				}
			}
		}
		
		private class SimpleQuery: Query
		{
			public string Phrase = null;
			public bool CaseSensitive = false;
			public bool WholeWords = false;
			public System.Text.StringFilter FileFilter = null;
			
			public override string[] Phrases
			{
				get 
				{
					if (Phrase == null)
					{
						return new string[0];
					}
					else
					{
						return new string[]{Phrase};
					}
				}
			}
		}
		
		private class CombinedQuery: Query
		{
			public SetCombineMode CombineMode = SetCombineMode.None;
			public Query Left = null;
			public Query Right = null;
			public bool LeftAddsToResultSet;
			public bool RightAddsToResultSet;
			
			public override string[] Phrases 
			{
				get 
				{
					System.Collections.Generic.List<string> results = new System.Collections.Generic.List<string>();
					if ((Left != null) && LeftAddsToResultSet)
					{
						foreach(string phrase in Left.Phrases)
						{
							if (!String.IsNullOrEmpty(phrase) && !results.Contains(phrase))
							{
								results.Add(phrase);
							}
						}
					}
					if ((Right != null) && RightAddsToResultSet)
					{
						foreach(string phrase in Right.Phrases)
						{
							if (!String.IsNullOrEmpty(phrase) && !results.Contains(phrase))
							{
								results.Add(phrase);
							}
						}
					}
					return results.ToArray();
				}
			}
		}
		
		private class SearchResults: System.Collections.Generic.List<SearchResult>, System.Collections.Generic.IComparer<SearchResult>
		{
			public int Compare(SearchResult x, SearchResult y)
			{
				int result = x.Row - y.Row;
				if (result == 0)
				{
					result = x.Column - y.Column;
				}
				return result;
			}
			
			public new void Sort()
			{
				base.Sort(this);
			}
		}
		
		public SearchResult[] Execute(Query query, string Text)
		{
			SearchResults results = InternalExecute(query, Text);
			if (results == null)
			{
				return null;
			}
			else
			{
				return results.ToArray();
			}
		}
		
		private SearchResults InternalExecute(Query query, string Text)
		{
			if (query == null)
			{
				return null;
			}
			if (query is SimpleQuery)
			{
				return InternalExecute((SimpleQuery)query, Text);
			}
			else if (query is CombinedQuery)
			{
				return InternalExecute((CombinedQuery)query, Text);
			}
			else
			{
				// TODO:
				throw new NotImplementedException();
			}
		}
		
		private SearchResults InternalExecute(CombinedQuery query, string Text)
		{
			SearchResults results = null;
			if (query.LeftAddsToResultSet)
			{
				results = InternalExecute(query.Left, Text);
			}
			if (query.RightAddsToResultSet)
			{
				SearchResults temp = InternalExecute(query.Right, Text);
				if (temp != null)
				{
					if (results == null)
					{
						results = temp;
					}
					else
					{
						results.AddRange(temp);
						
						// we have to merge overlapping regions
						results.Sort();
						
						int i = 0;
						while (i < results.Count - 1)
						{
							SearchResult current = results[i];
							SearchResult next = results[i + 1];
							if (current.Row == next.Row)
							{
								int currentStartsAt = current.Column;
								int currentEndsBefore = currentStartsAt + current.Length;
								int nextStartsAt = next.Column;
								int nextEndsBefore = nextStartsAt + next.Length;
								
								if (nextStartsAt < currentEndsBefore)
								{
									// 'next' item will be merged into the 'current' item
									int newEndsBefore = currentEndsBefore;
									if (nextEndsBefore > newEndsBefore)
									{
										newEndsBefore = nextEndsBefore;
									}
									current.Length = newEndsBefore - currentStartsAt;
									results[i] = current;
									results.RemoveAt(i + 1);
									continue;
								}
							}
							i++;
						}
					}
				}
			}
			return results;
		}
		
		private SearchResults InternalExecute(SimpleQuery query, string Text)
		{
			SearchResults results = null;
			
			if (!String.IsNullOrEmpty(query.Phrase))
			{
				System.IO.TextReader stream = new System.IO.StringReader(Text);
				try
				{
					int row = 0;
					string line;
					while ((line = stream.ReadLine()) != null)
					{
						SearchContext context = new SearchContext();
						context.Text = line;
						context.Phrase = query.Phrase;
						context.CaseSensitive = query.CaseSensitive;
						context.WholeWords = query.WholeWords;
						context.AllOccurencies = true;
						context.StartIndex = 0;
						
						int hits = Search(ref context);
						if (hits > 0)
						{
							if (results == null)
							{
								results = new SearchResults();
							}
							for(int offsetIndex = 0; offsetIndex < hits; offsetIndex++)
							{
								results.Add(new SearchResult(row, context.Offsets[offsetIndex], query.Phrase.Length));
							}
						}
						row++;
					}
				}
				finally
				{
					stream.Dispose();
				}
			}
			return results;
		}
		
		public DocumentCollection Execute(Query query)
		{
			return Execute(query, Documents);
		}
		
		private DocumentCollection Execute(Query query, DocumentCollection selection)
		{
			if (query == null)
			{
				return new CustomCollection();
			}
			if (query.Results != null)
			{
				return query.Results;
			}
			if (query is SimpleQuery)
			{
				SimpleQuery q = (SimpleQuery)query;
				q.Results = InternalFind(q.Phrase, q.CaseSensitive, q.WholeWords, q.FileFilter, selection);
				return q.Results;
			}
			else if (query is CombinedQuery)
			{
				CombinedQuery q = (CombinedQuery)query;
				if ((q.Left == null) || (q.Right == null))
				{
					throw new ArgumentException();
				}
				if (q.Left.Results == null)
				{
					// TODO: use combined progress to evaluate both left and right.
					// however, the UI doesn't create completely unresolved combined queries right now
					throw new ArgumentNullException();
				}
				if (q.Right.Results == null)
				{
					switch (q.CombineMode)
					{
						case SetCombineMode.And:
						case SetCombineMode.AndNot:
							Execute(q.Right, q.Left.Results);
							break;
							
						default:
							Execute(q.Right, selection);
							break;
					}
				}
				
				switch (q.CombineMode)
				{
					case SetCombineMode.None:
						q.Results = q.Right.Results;
						break;
						
					case SetCombineMode.And:
						{
							CustomCollection output = new CustomCollection(Documents);
							if (q.Left.Results.Count == 0)
							{
								q.Results = output;
							}
							else
							{
								output.BeginUpdate();
								foreach(DocumentHandle leftHandle in q.Left.Results)
								{
									if (q.Right.Results.Contains(leftHandle))
									{
										output.Add(leftHandle);
									}
								}
								output.EndUpdate();
								q.Results = output;
							}
						}
						break;
						
					case SetCombineMode.AndNot:
						{
							CustomCollection output = new CustomCollection(Documents);
							if (q.Left.Results.Count == 0)
							{
								q.Results = output;
							}
							else
							{
								output.BeginUpdate();
								foreach(DocumentHandle leftHandle in q.Left.Results)
								{
									if (!q.Right.Results.Contains(leftHandle))
									{
										output.Add(leftHandle);
									}
								}
								output.EndUpdate();
								q.Results = output;
							}
						}
						break;
						
					case SetCombineMode.Or:
						if (q.Left.Results.Count == 0)
						{
							q.Results = q.Right.Results;
						}
						else
						{
							CustomCollection output = new CustomCollection(Documents);
							output.BeginUpdate();
							foreach(DocumentHandle leftHandle in q.Left.Results)
							{
								output.Add(leftHandle);
							}
							foreach(DocumentHandle rightHandle in q.Right.Results)
							{
								output.Or(rightHandle);
							}
							output.EndUpdate();
							q.Results = output;
						}
						break;
						
					default:
						throw new NotImplementedException();
				}
				return query.Results;
			}
			else
			{
				// TODO:
				throw new NotImplementedException();
			}
		}
		
		private class TermEnumerator: System.Collections.Generic.IEnumerator<string>, System.Collections.Generic.IEnumerable<string>
		{
			private Lucene.Net.Index.TermEnum _enum;
			private Index _parent;
			private string _documentField;
			private string _start;
			private bool _useTermSeek;
			
			public TermEnumerator(Index parent, string documentField, string start, bool useTermSeek)
			{
				_parent = parent;
				_documentField = documentField;
				_start = start;
				_useTermSeek = useTermSeek;
				Reset();
			}
			
			public string Current 
			{
				get 
				{
					return _enum.Term().Text();
				}
			}
			
			object System.Collections.IEnumerator.Current 
			{
				get 
				{
					return _enum.Term().Text();
				}
			}
			
			public void Dispose()
			{
				_enum.Close();
				_enum = null;
			}
			
			public bool MoveNext()
			{
				while (true)
				{
					if (!_currentAvailable)
					{
						if (!_enum.Next())
						{
							return false;
						}
					}
					_currentAvailable = false;
					
					Lucene.Net.Index.Term candidate = _enum.Term();
					if (candidate == null)
					{
						return false;
					}
					if (candidate.Field() == _documentField)
					{
						if (candidate.Text().Length >= _start.Length)
						{
							return true;
						}
					}
				}
			}
			
			private bool _currentAvailable;
			public void Reset()
			{
				Lucene.Net.Index.Term head = _parent.CreateSearchStartTerm(_documentField, _start, _useTermSeek);
				_enum = _parent.Reader.Terms(head);
				_currentAvailable = true;
			}
			
			public System.Collections.Generic.IEnumerator<string> GetEnumerator()
			{
				return this;
			}
			
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
			{
				return this;
			}
		}
		
		private class TermDocsEnumerator: System.Collections.Generic.IEnumerator<DocumentHandle>, System.Collections.Generic.IEnumerable<DocumentHandle>
		{
			private Lucene.Net.Index.TermDocs _enum;
			private Index _parent;
			private string _documentField;
			private string _start;
			
			public TermDocsEnumerator(Index parent, string documentField, string start)
			{
				_parent = parent;
				_documentField = documentField;
				_start = start;
				Reset();
			}
			
			public DocumentHandle Current 
			{
				get 
				{
					return _enum.Doc();
				}
			}
			
			object System.Collections.IEnumerator.Current 
			{
				get 
				{
					return _enum.Doc();
				}
			}
			
			public void Dispose()
			{
				_enum.Close();
				_enum = null;
			}
			
			public bool MoveNext()
			{
				if (!_currentAvailable)
				{
					if (!_enum.Next())
					{
						return false;
					}
				}
				_currentAvailable = false;
				return true;
			}
			
			private bool _currentAvailable;
			public void Reset()
			{
				Lucene.Net.Index.Term head = _parent.CreateSearchStartTerm(_documentField, _start, true);
				_enum = _parent.Reader.TermDocs(head);
				_currentAvailable = false;
			}
			
			public System.Collections.Generic.IEnumerator<DocumentHandle> GetEnumerator()
			{
				return this;
			}
			
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
			{
				return this;
			}
		}
		
		private System.Collections.Generic.IEnumerable<string> GetWordEnumerator()
		{
			return GetWordEnumerator("");
		}
		
		private System.Collections.Generic.IEnumerable<string> GetWordEnumerator(string start)
		{
			return GetWordEnumerator(start, false);
		}
		
		private System.Collections.Generic.IEnumerable<string> GetWordEnumerator(string start, bool useTermSeek)
		{
			return new TermEnumerator(this, DocumentFields.Contents, start, useTermSeek);
		}
		
		private System.Collections.Generic.IEnumerable<DocumentHandle> GetDocumentEnumerator(string start)
		{
			return new TermDocsEnumerator(this, DocumentFields.Contents, start);
		}
		
		private const CachedDocumentId NoCacheId = 0;
			
		private string GetCachedFileName(CachedDocumentId id)
		{
			return id.ToString("X") + ".dat";
		}

		public System.IO.TextReader CreateTextReader(DocumentHandle handle)
		{
			return OpenIndexedContent(handle, Documents.GetPath(handle));
		}
		
		private System.IO.TextReader OpenIndexedContent(string fileName, CachedDocumentId cacheId)
		{
			if (cacheId == NoCacheId)
			{
				return _io.CreateTextReader(fileName);
			}
			else
			{
				return System.Text.FileStream.CreateReader(_store.OpenInputStream(GetCachedFileName(cacheId), IndexStore.InputStreamAccessMode.MultipleSequentialScan));
			}
		}
		
		private System.IO.TextReader OpenIndexedContent(DocumentHandle handle, string fileName)
		{
			DocumentKind kind = Documents.GetKind(handle);
			if (kind == DocumentKind.Indexed)
			{
				return OpenIndexedContent(fileName, NoCacheId);
			}
			else if (kind == (DocumentKind.Indexed | DocumentKind.Cached))
			{
				DocumentFlags flags = Documents.GetFlags(handle);
				return OpenIndexedContent(fileName, flags.CacheId);
			}
			else
			{
				return null;
			}
		}
		
		private System.IO.TextWriter CreateIndexedContent(CachedDocumentId cacheId)
		{
			return new System.IO.StreamWriter(_store.CreateOutputStream(GetCachedFileName(cacheId)), System.Text.Encoding.UTF8);
		}
		
		private static string BuildPhrase(string phrase, bool phrasesSupported)
		{
			const string Quote = "\"";
			
			if (phrasesSupported)
			{
				return Quote + phrase + Quote;
			}
			else
			{
				bool singleTermPhrase = !phrase.Contains(" ");
				
				if (singleTermPhrase)
				{
					return Quote + phrase + Quote;
				}
				else
				{
					return "(" + Quote + phrase.Replace(" ", Quote + " AND " + Quote) + Quote + ")";
				}
			}
		}
		
		private DocumentCollection InternalFind(string phrase, bool caseSensitive, bool wholeWords, System.Text.StringFilter fileNameFilter, DocumentCollection selection)
		{
			SearchProgress progress = new SearchProgress(this, true);
			try
			{
				if (String.IsNullOrEmpty(phrase) && fileNameFilter.AcceptsAll())
				{
					Documents.Sorted = true;
					return Documents;
				}
				
				// TODO: TUNE: step proportions
				System.Threading.SingleProgress indexProgress = progress.Add(10);
				System.Threading.SingleProgress filesProgress = progress.Add(200);
				
				AnalyzerBase analyzer = CreateAnalyzer(AnalyzerMode.TokenizeQuery);

				string normalizedPhrase = "";
				{
					Lucene.Net.Analysis.TokenStream tokens = analyzer.TokenStream(DocumentFields.Contents, new System.IO.StringReader(phrase));
					
					while(true)
					{
						Lucene.Net.Analysis.Token token = tokens.Next();
						if (token == null)
						{
							break;
						}
						
						string tokenText = token.TermText();
						if (!String.IsNullOrEmpty(normalizedPhrase))
						{
							normalizedPhrase = normalizedPhrase + " ";
						}
						normalizedPhrase = normalizedPhrase + tokenText;
					}
				}

				// results of simpleQuery don't have to be verified in the file. Hitting simpleQuery in the index means
				// the file really contained the queried term
				bool simpleQuery = (String.Compare(phrase, normalizedPhrase, true) == 0) && !normalizedPhrase.Contains(" ");
				
				// contains index search result set that needs to be filtered for filename and verified (case sensitive,
				// whole word, special characters embedded)
				DocumentCollection searchResult = null;

				// if exists, contains handles to be included in the search
				FilterCollection searchFilter = null;
				
				if (String.IsNullOrEmpty(normalizedPhrase))
				{
					// TODO: takes too long to process if phrase is not empty
					searchResult = Documents;
					progress.Skip(indexProgress);
				}
				else
				{
					bool endWildcardNeeded;		// if true, StartsWith() matching is needed
					bool startWildcardNeeded;	// if true, EndsWith() matching is needed
					bool singleTermQuery;		// if true, the query contains just one term
					bool containsNeeded;		// if true, the query contains just one term and Contains() matching is needed
					bool phrasesSupported;		// if true, the query can contain phrases as these are supported by the analyzer
					
					phrasesSupported = analyzer.SupportsPhrases(AnalyzerMode.TokenizeDocument);
					
					singleTermQuery = !normalizedPhrase.Contains(" ");
					containsNeeded = false;
					
					if (wholeWords)
					{
						endWildcardNeeded = false;
						startWildcardNeeded = false;
						normalizedPhrase = BuildPhrase(normalizedPhrase, phrasesSupported);
					}
					else
					{
						endWildcardNeeded = WordPartAt(analyzer, phrase, phrase.Length - 1, false);
						startWildcardNeeded = WordPartAt(analyzer, phrase, 0, false);
						
						if (startWildcardNeeded || endWildcardNeeded)
						{
							if (singleTermQuery && startWildcardNeeded && endWildcardNeeded)
							{
								containsNeeded = true;
							}
						}
						else
						{
							normalizedPhrase = BuildPhrase(normalizedPhrase, phrasesSupported);
						}
					}
					
					if (startWildcardNeeded || endWildcardNeeded || containsNeeded)
					{
						// Lucene doesn't support wildcard phrase queries. We have to rewrite the query into
						// multiple boolean clauses
						string[] queryTerms = normalizedPhrase.Split(' ');
						System.Collections.Generic.List<string> queries = new System.Collections.Generic.List<string>();
						
						System.Threading.SingleProgress wordProgress = progress.Add(100);
						wordProgress.Treshold = (ulong)(WordCount / 100);	// TODO: TUNE: how many times the word progress should be called
						wordProgress.Open(0, (ulong)WordCount * 2);
						wordProgress.Increment = 2;
						wordProgress.Write("Words" /* TODO: localize */);
						try
						{
							const int firstTermIndex = 0;
							int lastTermIndex = queryTerms.Length - 1;
							
							// TODO: use some query explosion protection
							
							if (containsNeeded)
							{
								searchFilter = new FilterCollection(Documents);
								
								string term = queryTerms[firstTermIndex];
								foreach(string word in GetWordEnumerator(term))
								{
									if (word.Contains(term))
									{
										foreach(DocumentHandle handle in GetDocumentEnumerator(word))
										{
											searchFilter.Or(handle);
										}
									}
									if (!wordProgress.StepIt())
									{
										// interrupted
										return new CustomCollection();
									}
								}
							}
							else if (startWildcardNeeded && endWildcardNeeded)
							{
								wordProgress.Increment = 1;
								
								FilterCollection tails = new FilterCollection(Documents);
								
								string term = queryTerms[lastTermIndex];
								foreach(string word in GetWordEnumerator(term, true))
								{
									if (word.StartsWith(term))
									{
										foreach(DocumentHandle handle in GetDocumentEnumerator(word))
										{
											tails.Or(handle);
										}
									}
									else
									{
										break;
									}
									if (!wordProgress.StepIt())
									{
										// interrupted
										return new CustomCollection();
									}
								}
								
								if (tails.Count == 0)
								{
									// endWildcard had no hits
									return new CustomCollection();
								}
								
								searchFilter = new FilterCollection(Documents);

								term = queryTerms[firstTermIndex];
								
								// remove the to-be-substituted term from the query
								queries.Add(String.Join(" ", queryTerms, 1, queryTerms.Length - 2));
								
								foreach(string word in GetWordEnumerator(term))
								{
									if (word.EndsWith(term))
									{
										foreach(DocumentHandle handle in GetDocumentEnumerator(word))
										{
											if (tails.Contains(handle))
											{
												searchFilter.Or(handle);
											}
										}
									}
									if (!wordProgress.StepIt())
									{
										// interrupted
										return new CustomCollection();
									}
								}
							}
							else if (startWildcardNeeded)
							{
								searchFilter = new FilterCollection(Documents);
								
								string term = queryTerms[firstTermIndex];
								
								// remove the to-be-substituted term from the query
								queries.Add(String.Join(" ", queryTerms, 1, queryTerms.Length - 1));
								
								foreach(string word in GetWordEnumerator(term))
								{
									if (word.EndsWith(term))
									{
										foreach(DocumentHandle handle in GetDocumentEnumerator(word))
										{
											searchFilter.Or(handle);
										}
									}
									if (!wordProgress.StepIt())
									{
										// interrupted
										return new CustomCollection();
									}
								}
							}
							else if (endWildcardNeeded)
							{
								searchFilter = new FilterCollection(Documents);
								
								string term = queryTerms[lastTermIndex];
								
								// remove the to-be-substituted term from the query
								queries.Add(String.Join(" ", queryTerms, 0, queryTerms.Length - 1));
								
								foreach(string word in GetWordEnumerator(term, true))
								{
									if (word.StartsWith(term))
									{
										foreach(DocumentHandle handle in GetDocumentEnumerator(word))
										{
											searchFilter.Or(handle);
										}
									}
									else
									{
										break;
									}
									if (!wordProgress.StepIt())
									{
										// interrupted
										return new CustomCollection();
									}
								}
							}
						}
						finally
						{
							wordProgress.Close();
						}
						
						{
							// TODO: remove once debugged
							System.Diagnostics.Trace.WriteLine(String.Format("OR clauses: {0}", queries.Count));
							
							// rewrite collected queries into the OR-ed phrase query
							// TODO: use some query optimizer-reducer if there are too many queries
							System.Text.StringBuilder buffer = new System.Text.StringBuilder();
							bool empty = true;
							for(int i = 0; i < queries.Count; i++)
							{
								if (String.IsNullOrEmpty(queries[i]))
								{
									continue;
								}
								if (!empty)
								{
									buffer.Append(" OR ");
								}
								buffer.Append(BuildPhrase(queries[i], phrasesSupported));
								empty = false;
							}
							normalizedPhrase = buffer.ToString();
						}
						
						if (String.IsNullOrEmpty(normalizedPhrase) && (searchFilter == null))
						{
							// word list had no hits
							return new CustomCollection();
						}
					}
					
					{
						// TODO: remove once debugged
						if (searchFilter != null)
						{
							System.Diagnostics.Trace.WriteLine(String.Format("Search filter: {0}", searchFilter.Count));
						}
					}
					
					// TODO: remove once debugged
					// System.Diagnostics.Trace.WriteLine(String.Format("Search \"{0}\"", normalizedPhrase));
				
					if ((searchFilter != null) && String.IsNullOrEmpty(normalizedPhrase))
					{
						searchResult = searchFilter;
						progress.Skip(indexProgress);
					}
					else
					{
						indexProgress.Open(0, 10);
						indexProgress.Write("Index" /* TODO: localize */);
						try
						{
							Lucene.Net.Search.Searcher searcher = new Lucene.Net.Search.IndexSearcher(Reader);
							Lucene.Net.Search.BooleanQuery.SetMaxClauseCount(Int32.MaxValue);
							try
							{
								Lucene.Net.QueryParsers.QueryParser parser = new Lucene.Net.QueryParsers.QueryParser(DocumentFields.Contents, analyzer);
								Lucene.Net.Search.Query query = parser.Parse(normalizedPhrase);
								// TODO: use custom hit collector to interrupt the search if needed
								Lucene.Net.Search.Hits hits;
								if (searchFilter == null)
								{
									hits = searcher.Search(query);
								}
								else
								{
									hits = searcher.Search(query, searchFilter.CreateFilter());
								}
								indexProgress.Step(8);
								searchResult = new SearchResultsCollection(hits, _io);
							}
							finally
							{
								searcher.Close();
							}
						}
						finally
						{
							indexProgress.Close();
						}
					}
					
					// TODO: remove once debugged
					System.Diagnostics.Trace.WriteLine(String.Format("Hits: {0}", searchResult.Count));
					
				}
				Documents.Sorted = true;
				CustomCollection result = new CustomCollection(Documents);
				result.BeginUpdate();
				
				filesProgress.Open(0, (ulong)searchResult.Count);
				filesProgress.Write("Files" /* TODO: localize */);
				try
				{
					bool useTextSearch;
					bool checkFileExists;
					
					if (String.IsNullOrEmpty(phrase))
					{
						useTextSearch = false;
						checkFileExists = false;
					}
					else
					{
						useTextSearch = (!simpleQuery) || wholeWords || caseSensitive;
						checkFileExists = true;
					}
					
					foreach(DocumentHandle handle in searchResult)
					{
						if (selection.Contains(handle))
						{
							string fileName = Documents.GetPath(handle);	// using Documents collection instead of
																			// searchResult.GetItemPath(i);
																			// because Documents uses path cache
	
							if (fileNameFilter.Accepts(fileName))
							{
								int score = 0;
								
								if (useTextSearch)
								{
									try
									{
										System.IO.TextReader r = OpenIndexedContent(handle, fileName);
										if (r != null)
										{
											try
											{
												SearchContext context = new SearchContext();
												context.TextStream = r;
												context.Phrase = phrase;
												context.CaseSensitive = caseSensitive;
												context.WholeWords = wholeWords;
												context.AllOccurencies = false;
												context.StartIndex = 0;
												score = Search(ref context);
											}
											finally
											{
												r.Dispose();
											}
										}
									}
									catch (Exception)
									{
										// TODO: notify user
										// file can not be analyzed now, don't include it
										score = 0;
									}
								}
								else
								{
									if (checkFileExists)
									{
										if ((Documents.GetKind(handle) & DocumentKind.Deleted) == 0)
										{
											score = 1;
										}
									}
									else
									{
										// special case - just file name filter
										score = 1;
									}
								}
								
								if (score > 0)
								{
									result.Add(handle);
									FoundDocumentCountChanged(result.Count, false);
								}
							}
						}
						if (!filesProgress.StepIt())
						{
							// TODO: show "interrupted"
							break;
						}
					}
					FoundDocumentCountChanged(result.Count, true);
					result.EndUpdate();
					return result;
				}
				finally
				{
					filesProgress.Close();
				}
			}
			finally
			{
				progress.SkipAll();
			}
		}
		#endregion

		private bool IsAccentedCharacter(byte[] character)
		{
			char c = System.Text.Encoding.Default.GetString(character)[0];
			string normalized = System.Convert.ToString(c).Normalize(System.Text.NormalizationForm.FormKD);
			System.Text.Encoding removal = System.Text.Encoding.GetEncoding(
				System.Text.Encoding.ASCII.CodePage,
				new System.Text.EncoderReplacementFallback(""),
				new System.Text.DecoderReplacementFallback(""));
		
			byte[] bytes = removal.GetBytes(normalized);
			string decoded = System.Text.Encoding.ASCII.GetString(bytes);
			if (decoded.Length != 1)
			{
				return false;
			}
			char decodedC = decoded[0];
			return 
				((decodedC >= 'a') && (decodedC <= 'z')) ||
				((decodedC >= 'A') && (decodedC <= 'Z'));
		}
		
		private const int FilePreviewBufferLength = 512; 	// TUNE:
		private bool IsBinaryOrEmpty(string fileName, out IParser parser)
		{
			parser = null;
			System.IO.Stream r = _io.OpenFile(fileName, System.IO.FileMode.Open, System.IO.FileAccess.Read, System.IO.FileShare.Read /* TODO: is this the good default mode ?*/);
			try
			{
				if (r.Length == 0)
				{
					return true;
				}
				System.Text.Encoding encoding = System.Text.FileStream.GetEncoding(r, System.Text.Encoding.Default);
				if (encoding == System.Text.Encoding.Default)
				{
					// no special text-encoding preamble was found
					int len = FilePreviewBufferLength;
					if (r.Length < len)
					{
						len = (int)r.Length;
					}
					byte[] buffer = new byte[len];
					len = r.Read(buffer, 0, len);
					
					parser = Parsers.Parsers.Match(fileName, buffer);
					if (parser != null)
					{
						if (parser.SupportsCopy)
						{
							// the file may be binary, but it's parser is known
							return false;
						}
						else
						{
							// the file may look strange (non-ascii characters) but it is known to be valid text
							parser = null;
							return false;
						}
					}
					
					// check if it is valid plain-text file
					int flaws = 0;
					for(int i = 0; i < len; i++)
					{
						if (buffer[i] > 127)
						{
							if (!IsAccentedCharacter(new byte[]{buffer[i]}))
							{
								flaws++;
								if (flaws > 3 /* TODO: TUNE: */)
								{
									return true;
								}
							}
						}
						else if (buffer[i] < 32)
						{
							switch (buffer[i])
							{
								case 13:	// new line
								case 10:	// new line
								case 9:		// tab
								case 12:	// new page
								case 26:	// end-of-file
									break;
									
								default:
									// binary code found
									return true;
							}
						}
					}
					return false;
				}
				else
				{
					// we have a text encoded file. Lets check if there is a parser available for the file
					// so that the indexed content is more human-friendly
					System.IO.TextReader reader =  new System.IO.StreamReader(r, encoding);
					try
					{
						char[] buffer = new char[FilePreviewBufferLength];
						int len = reader.ReadBlock(buffer, 0, buffer.Length);
						parser = Parsers.Parsers.Match(fileName, new String(buffer, 0, len));
						if (parser != null)
						{
							if (!parser.SupportsCopy)
							{
								// we don't need copy-less parser as regular TextReader is enough
								parser = null;
							}
						}
					}
					finally
					{
						reader.Close();
					}
					return false;
				}
			}
			finally
			{
				r.Dispose();
			}
		}
		
		#region Update progress
		private class ProgressAdapter: System.Threading.SingleProgress
		{
			private Index _parent;
			public ProgressAdapter(Index parent)
			{
				_parent = parent;
			}
			
			protected override void DoOpen(ulong min, ulong max)
			{
				if (min != 0)
				{
					throw new ArgumentException();
				}
				_parent.OpenProgress(null, (int)max);
			}
			
			protected override bool DoStep(ulong value)
			{
				return _parent.ShowProgress((int)value);
			}
			
			protected override void DoClose()
			{
				_parent.CloseProgress();
			}
			
			protected override void DoWrite(string message)
			{
				_parent.ShowMessage(message);
			}
			
			public override bool Aborted 
			{
				get 
				{
					return _parent.InterruptRequested(); 
				}
			}
		}
		
		internal System.Threading.IProgress CreateProgressAdapter()
		{
			return new ProgressAdapter(this);
		}
		
		private class Updater: IDisposable
		{
			private int _count;
			private int _step;
			private bool _progressActive;
			private FileCounter _fileCounter;
			private IAsyncResult _fileCounterResult;
			private bool _activateProgress;
			private Index _parent;
			private bool _stopCounting;
			private System.Threading.IProgress _progress;
			public SortedFileNameList Survivers = null;

			public void StopCounting()
			{
				_stopCounting = true;
			}
			
			private void OpenProgress()
			{
				_activateProgress = false;
				_progressActive = true;
				_progress.Open(0, (ulong)_count);
				_parent._updaterVisible = true;
			}
			
			public Updater(Index parent, string[] seeds, System.Threading.IProgress progress)
			{
				_parent = parent;
				if (progress == null)
				{
					_progress = _parent.CreateProgressAdapter();
				}
				else
				{
					_progress = progress;
				}
				System.Collections.Generic.List<string> folders = new System.Collections.Generic.List<string>();
				foreach(string fileName in seeds)
				{
					if (_parent._io.DirectoryExists(fileName))
					{
						folders.Add(fileName);
					}
					else if (_parent._io.FileExists(fileName))
					{
						_count++;
					}
				}
				if (folders.Count == 0)
				{
					OpenProgress();
				}
				else
				{
					_fileCounter = new FileCounter(this.CountFiles);
					_fileCounterResult = _fileCounter.BeginInvoke(_parent._io, folders, new AsyncCallback(this.FileCounterCompleted), null);
				}
			}

			private int _stepLimit = 0;
			public bool Step()
			{
				bool canContinue = true;
				_stepLimit++;
				if (_stepLimit > 10 /* TODO: TUNE: report frequency */)
				{
					_stepLimit = 0;
					if (_activateProgress)
					{
						OpenProgress();
					}
					if (_progressActive)
					{
						canContinue = _progress.Step((ulong)_step);
					}
					else
					{
						canContinue = !_progress.Aborted;
					}
				}
				_step++;
				return canContinue;
			}
			
			private void FileCounterCompleted(IAsyncResult result)
			{
				if (_fileCounter != null)
				{
					_count += _fileCounter.EndInvoke(result);
					_activateProgress = true;
				}
				_fileCounter = null;
			}
			
			private delegate int FileCounter(Explorer.Model.Reader io, System.Collections.Generic.List<string> folders);
			private int CountFiles(Explorer.Model.Reader io, System.Collections.Generic.List<string> folders)
			{
				try
				{
					int result = 0;
					while(folders.Count > 0)
					{
						if (_stopCounting)
						{
							// interrupted
							return 0;
						}
						string folder = folders[0];
						folders.RemoveAt(0);
					
						result += io.GetFiles(folder).Count;
	
						foreach(string subFolder in io.GetDirectories(folder))
						{
							folders.Add(subFolder);
						}
					}
					return result;
				}
				finally
				{
					io.DisconnectThread();
				}
			}
			
			private void Close()
			{
				if (_fileCounter != null)
				{
					// TODO: !! lock the _fileCounter variable
					FileCounter counter = _fileCounter;
					_fileCounter = null;
					_stopCounting = true;
					counter.EndInvoke(_fileCounterResult);
				}
				if (_progressActive)
				{
					_progress.Close();
				}
			}
			
			#region Implementation of IDisposable
			public void Dispose()
			{
				Dispose(true);
				GC.SuppressFinalize(this);
			}
	
			private void Dispose(bool disposing)
			{
				Close();
			}
	
			~Updater()
			{
				Dispose(false);
			}
			#endregion
		}
		
		private Updater _updater = null;
		private bool _updaterVisible = false;
		
		private void UpdaterOpen(string[] seeds, System.Threading.IProgress progress)
		{
			_updater = new Updater(this, seeds, progress);
		}
		
		private void UpdaterOpen(string[] seeds)
		{
			UpdaterOpen(seeds, null);
		}
		
		private void UpdaterClose()
		{
			_updater.Dispose();
			_updater = null;
			_updaterVisible = false;
		}
		
		private void UpdaterStop()
		{
			if (_updater != null)
			{
				_updater.StopCounting();
			}
		}
		
		private bool UpdaterStep()
		{
			if (_updater != null)
			{
				return _updater.Step();
			}
			else
			{
				return true;
			}
		}
		
		private bool UpdaterAcceptsFile(string fileName)
		{
			if (_updater != null)
			{
				if (_updater.Survivers != null)
				{
					if (_updater.Survivers.Contains(fileName))
					{
						return false;
					}
				}
			}
			return true;
		}
		#endregion

		protected enum FileIndexOption
		{
			DoNotInclude,
			IndexHeader,
			IndexHeaderAndContents
		}
		
		protected virtual FileIndexOption AcceptsFile(string fileName, out IParser parser)
		{
			try
			{
				if (IsBinaryOrEmpty(fileName, out parser))
				{
					return FileIndexOption.IndexHeader;
				}
				return FileIndexOption.IndexHeaderAndContents; // TODO: System.IO.Path.GetExtension(fileName);
			}
			catch(Exception)
			{
				// file is not accessible right now
				parser = null;
				if (_io.FileExists(fileName))
				{
					return FileIndexOption.IndexHeader;
				}
				else
				{
					return FileIndexOption.DoNotInclude;
				}
			}
		}
		
		protected FileIndexOption AcceptsFile(string fileName)
		{
			IParser dummy;
			return AcceptsFile(fileName, out dummy);
		}

		private bool InternalAddFile(string fileName)
		{
			bool result = false;
			try
			{
				if (UpdaterAcceptsFile(fileName))
				{
					Explorer.Model.Entry info = _io.GetFileInfo(fileName);
					if ((info != null) && ((info.Attributes & System.IO.FileAttributes.Hidden) == 0))
					{
						IParser parser;
						FileIndexOption option = AcceptsFile(fileName, out parser);
						if (option != FileIndexOption.DoNotInclude)
						{
							result = true;
							Lucene.Net.Documents.Document document;
							try
							{
								document = CreateDocument(fileName, option == FileIndexOption.IndexHeaderAndContents, parser);
								Writer.AddDocument(document);
								Modified = true;
							}
							catch(Exception)
							{
								// TODO: probably the file could not be read right now. should we notify the user?
								result = false;
							}
							if (result)
							{
								if (option == FileIndexOption.IndexHeaderAndContents)
								{
									DocumentsSize = DocumentsSize + info.Size;
								}
								DocumentCountChanged(Writer.DocCount(), !_updaterVisible);
							}
						}
					}
				}
			}
			catch(Exception)
			{
				// TODO: probably the file could not be accessed because of too long path
				result = false;
			}
			Tools.ThrowExceptionIfInterrupted(UpdaterStep());
			return result;
		}
	
		private static bool IsDrive(string folderName)
		{
			return folderName.EndsWith(":\\");
		}
		
		private void InternalAddFolder(string folderName)
		{
			Explorer.Model.Entry info;
			try
			{
				info = _io.GetDirectoryInfo(folderName);
			}
			catch(Exception)
			{
				// probably path too long exception
				// TODO: inform user somehow
				return;
			}
			
			bool accepted = info != null;
			if (accepted)
			{
				if (!IsDrive(info.FullName))
				{
					accepted = (info.Attributes & System.IO.FileAttributes.Hidden) == 0;
				}
			}
			
			if (accepted)
			{
				System.Collections.Generic.ICollection<string> files = _io.GetFiles(folderName);
				foreach(string fileName in files)
				{
					InternalAddFile(fileName);
				}
				System.Collections.Generic.ICollection<string> folders = _io.GetDirectories(folderName);
				foreach(string folder in folders)
				{
					InternalAddFolder(folder);
				}
			}
		}
		
		public bool Mount(string fileName)
		{
			if (AddMount(fileName))
			{
				return true;				
			}
			else
			{
				return false;
			}
		}
		
		public bool Mount(string[] fileNames)
		{
			bool modified = false;
			foreach(string fileName in fileNames)
			{
				if (Mount(fileName))
				{
					modified = true;
				}
			}
			if (modified)
			{
				MountFileSystem();
				Update();
			}
			return modified;
		}
		
		public void Add(string fileName)
		{
			Add(new string[]{fileName});
		}

		public void Add(string[] fileNames)
		{
			UpdaterOpen(fileNames);
			try
			{
				ShowMessage("Prepare");
				_updater.Survivers = new SortedFileNameList(Documents);
	
				BeginUpdate();

				ShowMessage("Add");
				try
				{
					// non-interruptible part
					foreach(string fileName in fileNames)
					{
						AddSeed(fileName);
					}
					
					// interruptible part
					foreach(string fileName in fileNames)
					{
						if (_io.DirectoryExists(fileName))
						{
							InternalAddFolder(fileName);
						}
						else if (_io.FileExists(fileName))
						{
							InternalAddFile(fileName);
						}
						else
						{
							// file or folder is not visible
						}
					}
				}
				finally
				{
					EndUpdate();
				}
			}
			finally
			{
				UpdaterClose();
			}
		}
		
		private class DocumentFields
		{
			public const string Path = "path";
			public const string Contents = "contents";
			public const string Flags = "flags";
		}
		
		internal class DocumentFlags
		{
			public DateTime LastWriteTime;
			public bool Indexed;
			public long Length;
			public CachedDocumentId CacheId;
			
			public bool Valid
			{
				get
				{
					return Length >= 0;
				}
			}
			
			public bool Cached
			{
				get
				{
					return CacheId != NoCacheId;
				}
			}
			
			public DocumentFlags()
			{
			}

			public DocumentFlags(byte[] bytes)
			{
				SetBytes(bytes);
			}

			public DocumentFlags(Explorer.Model.Reader io, string fileName, bool indexContents, CachedDocumentId cacheId)
			{
				Explorer.Model.Entry info = io.GetFileInfo(fileName);
				LastWriteTime = info.LastWriteTime;
				Length = info.Size;
				Indexed = indexContents;
				CacheId = cacheId;
			}
			
			private const byte HeaderBitBinary = 1 << 0;
			private const byte HeaderBitCached = 1 << 1;
			
			private const int HeaderShift = (sizeof(long) * 8) - 8;
			private const long HeaderUnMask = (1L << HeaderShift) - 1;
			
			public byte[] GetBytes()
			{
				byte[] result;
				if (Cached)
				{
					result = new byte[sizeof(long) + sizeof(long) + sizeof(UInt32)];
				}
				else
				{
					result = new byte[sizeof(long) + sizeof(long)];
				}
				System.IO.BinaryWriter w = new System.IO.BinaryWriter(new System.IO.MemoryStream(result));
				try
				{
					byte header = 0;
					if (!Indexed)
					{
						header |= HeaderBitBinary;
					}
					if (Cached)
					{
						header |= HeaderBitCached;
					}
					long firstWord = (((long)header) << HeaderShift) | Length;
					w.Write(firstWord);
					w.Write(LastWriteTime.ToBinary());
					if (Cached)
					{
						w.Write(CacheId);
					}
				}
				finally
				{
					w.Close();
				}
				return result;
			}
			
			public void SetBytes(byte[] value)
			{
				if (value == null)
				{
					Length = -1;
					Indexed = true;
				}
				else
				{
					System.IO.BinaryReader r = new System.IO.BinaryReader(new System.IO.MemoryStream(value));
					try
					{
						long firstWord = r.ReadInt64();
						byte header = (byte)(firstWord >> HeaderShift);
						
						Indexed = (header & HeaderBitBinary) == 0;
						
						Length = firstWord & HeaderUnMask;
						LastWriteTime = DateTime.FromBinary(r.ReadInt64());
						
						if ((header & HeaderBitCached) != 0)
						{
							CacheId = r.ReadUInt32();
						}
					}
					finally
					{
						r.Close();
					}
				}
			}
		}
		
		private Lucene.Net.Documents.Document CreateDocument(string fileName, bool indexContents, IParser parser)
		{
			Lucene.Net.Documents.Document result = new Lucene.Net.Documents.Document();

			// Add the path of the file as a field named "path".  Use a field that is
            // indexed (i.e. searchable), but don't tokenize the field into words.
            result.Add(new Lucene.Net.Documents.Field(DocumentFields.Path, fileName, Lucene.Net.Documents.Field.Store.YES, Lucene.Net.Documents.Field.Index.UN_TOKENIZED));

            CachedDocumentId cacheId = NoCacheId;
            
            if (indexContents && (parser != null))
			{
				try
				{
					cacheId = AllocCachedId();
					System.IO.TextWriter writer = CreateIndexedContent(cacheId);
					try
					{
						if (UsesNativeIO)
						{
							parser.Copy(fileName, writer);
						}
						else
						{
							// we must map the file out to the native file system so that the
							// parser can read it
							const string tempFileName = "parser.input";
							try
							{
								System.IO.Stream binaryReader = _io.OpenFile(fileName, System.IO.FileMode.Open, System.IO.FileAccess.Read, System.IO.FileShare.Read);
								try
								{
									System.IO.Stream binaryWriter = _store.CreateOutputStream(tempFileName);
									try
									{
										byte[] buffer = new byte[4 /* TODO: TUNE: */ * 1024];
										
										while (true)
										{
											int size = binaryReader.Read(buffer, 0, buffer.Length);
											if (size > 0)
											{
												binaryWriter.Write(buffer, 0, size);
											}
											else
											{
												break;
											}
										}
									}
									finally
									{
										binaryWriter.Close();
									}
								}
								finally
								{
									binaryReader.Close();
								}
								
								parser.Copy(_store.GetFullPath(tempFileName), writer);
							}
							finally
							{
								_store.DeleteFile(tempFileName);
							}
						}
					}
					finally
					{
						writer.Close();
					}
					if (_store.FileLength(GetCachedFileName(cacheId)) <= 3 /* TODO: Size of the UTF8 pre-amble. Use constant and share the preample size with it's creator */)
					{
						throw new ArgumentException();
					}
				}
				catch(Exception)
				{
					// TODO: something went wrong, probably the file can not be parsed now or never
					ReleaseCachedId(cacheId);
					cacheId = NoCacheId;
					indexContents = false;
				}
			}
			
			if (indexContents)
			{
	            // Add the contents of the file to a field named "contents".  Specify a Reader,
	            // so that the text of the file is tokenized and indexed, but not stored.
	            result.Add(new Lucene.Net.Documents.Field(DocumentFields.Contents, OpenIndexedContent(fileName, cacheId)));
			}
			
            // Add the last modified date of the file and the binary|text bit in a field named "flags".
			result.Add(new Lucene.Net.Documents.Field(DocumentFields.Flags, (new DocumentFlags(_io, fileName, indexContents, cacheId)).GetBytes(), Lucene.Net.Documents.Field.Store.YES));
			
			return result;
		}
		
		private bool _modified = false;
		public bool Modified
		{
			get
			{
				return _modified;
			}
			set
			{
				_modified = value;
			}
		}
		
		internal class SortedFileNameList
		{
			private System.Collections.Generic.List<string> _items;
			public SortedFileNameList(int capacity)
			{
				_items = new System.Collections.Generic.List<string>(capacity);
			}
			
			public SortedFileNameList(DocumentCollection source)
			{
				int count = source.Count;
				_items = new System.Collections.Generic.List<string>(count);
				source.Sorted = true;
				for(int i = 0; i < count; i++)
				{
					Add(source[i]);
				}
			}
			
			public void Add(string fileName)
			{
				_items.Add(fileName);
			}
			
			private class Comparer: System.Collections.Generic.IComparer<string>
			{
				public int Compare(string x, string y)
				{
					return AllDocumentsCollection.Compare(x, y);
				}
			}
			
			public bool Contains(string fileName)
			{
				return _items.BinarySearch(fileName, new Comparer()) >= 0;
			}
		}

		private bool ReEvaluateBinaryFile(string fileName)
		{
			// this is a binary file. check if its "binary" status didn't change
			switch (AcceptsFile(fileName))
			{
				case FileIndexOption.IndexHeaderAndContents:
					// file will be indexed
					return true;
					
				case FileIndexOption.IndexHeader:
					// no need to change. The DocumentsFlags (timestamp etc.) will be obsolete but
					// is it worth? the speed-up
					return false;
					
				case FileIndexOption.DoNotInclude:
					return true;
					
				default:
					throw new NotImplementedException();
			}
		}
		
		public void Update()
		{
			Update(false);
		}
		
		public void Update(bool skipChecking)
		{
			if (ReadConfigString("Index/Analyzer.Version") != RecommendedAnalyzerVersion)
			{
				Rebuild();
				return;
			}
			
			System.Threading.SlicedProgress progress = new System.Threading.SlicedProgress(CreateProgressAdapter());

			System.Threading.SingleProgress reviewFilesProgress = null;
			System.Threading.SingleProgress deleteFilesProgress = null;
			
			// TODO: TUNE: stage ranges
			if (Documents.Count > 0)
			{
				reviewFilesProgress = progress.Add(800);
				deleteFilesProgress = progress.Add(200);
			}
			System.Threading.SingleProgress addFilesProgress = progress.Add(2000);
			try
			{
				// start the background file counter
				UpdaterOpen(Seeds, addFilesProgress);
				try
				{
					int documentsToKeep = 0;
					int documentsToDelete = Documents.Count;
					
					System.Collections.BitArray deletedHandles = new System.Collections.BitArray(Documents.Count);
						
					DocumentCountChanged(documentsToKeep, true);
						
					#region Review Files stage
					// "review files" stage
					// go through already registered files and check their update status.
					// files that need update will be deleted from the index.
					// in the "add files" stage files kept in the index will not be indexed again
					if (reviewFilesProgress != null)
					{
						reviewFilesProgress.Open(0, (ulong)Documents.Count);
						ShowMessage("Check");
						try
						{
							for(int index = 0; index < Documents.Count; index++)
							{
								DocumentHandle handle = Documents.GetItemHandle(index);
								string fileName = Documents.GetPath(handle);
								
								bool deleteEntry = false;
								DocumentFlags flags = null;
								
								if (!skipChecking)
								{
									Explorer.Model.Entry info = _io.GetFileInfo(fileName);
									
									flags = Documents.GetFlags(handle);
									if (info != null)
									{
										if (!flags.Valid)
										{
											// this is an older version of the index. we have no data to compare so
											// the entry will be deleted and recreated
											deleteEntry = true;
										}
										else
										{
											if ((info.Size != flags.Length) ||
											    (info.LastWriteTime.ToFileTimeUtc() != flags.LastWriteTime.ToFileTimeUtc()))
											{
												// file was changed
												if (flags.Indexed)
												{
													deleteEntry = true;
												}
												else
												{
													deleteEntry = ReEvaluateBinaryFile(fileName);
												}
											}
											else
											{
												// if AcceptsFile() logic would change then we should compare its result to the flags.Indexed
												// property
												if (!flags.Indexed)
												{
													deleteEntry = ReEvaluateBinaryFile(fileName);
												}
											}
										}
									}
									else
									{
										// file does not exist. delete its entry
										deleteEntry = true;
									}
								}
								
								if (deleteEntry)
								{
									deletedHandles[handle] = true;
									if (flags.Valid)
									{
										if (flags.Indexed)
										{
											if (DocumentsSize >= 0)
											{
												DocumentsSize = DocumentsSize - flags.Length;
											}
										}
										if (flags.Cached)
										{
											ReleaseCachedId(flags.CacheId);
										}
									}
								}
								else
								{
									documentsToKeep++;
									documentsToDelete--;
									DocumentCountChanged(documentsToKeep, false);
								}
								
								Tools.ThrowExceptionIfInterrupted(reviewFilesProgress.StepIt());
							}
						}
						finally
						{
							reviewFilesProgress.Close();
						}
					}
					#endregion
					
					#region Before Detach stage
					// transfer lookup list of handles that survived to the _updater object for later
					// use
					{
						ShowMessage("Prepare");
						Documents.Sorted = true;
						SortedFileNameList survivers = new SortedFileNameList(documentsToKeep);
						int toGo = documentsToKeep;
						// go through all file names in the sorted order
						int index = 0;
						while (toGo > 0)
						{
							DocumentHandle handle = Documents.GetItemHandle(index);
							if (!deletedHandles[handle])
							{
								survivers.Add(Documents.GetPath(handle));
								toGo--;
							}
							index++;
						}
						_updater.Survivers = survivers;
					}
					#endregion

					Detach();

					if (DocumentsSize < 0)
					{
						DocumentsSize = 0;
					}

					bool oldModified = Modified;
					// set local detector for the Modified state.
					// when really Modified, the EndUpdate() operation takes longer than necessary
					Modified = false;
					BeginUpdate();
					try
					{
						#region Delete Files stage
						// delete files stage
						// transfer deletedHandles[] bits to the index
						if (deleteFilesProgress != null)
						{
							if (documentsToDelete > 0)
							{
								deleteFilesProgress.Open(0, (ulong)documentsToDelete);
								ShowMessage("Delete");
								try
								{
									DocumentHandle handle = 0;
									while (documentsToDelete > 0)
									{
										if (deletedHandles[handle])
										{
											Writer.DeleteDocument(handle);
											Modified = true;
											
											documentsToDelete--;
											
											Tools.ThrowExceptionIfInterrupted(deleteFilesProgress.StepIt());
										}
										handle++;
									}
								}
								finally
								{
									deleteFilesProgress.Close();
								}
								
								// commit changes
								Writer.Optimize();
							}
							else
							{
								progress.Skip(deleteFilesProgress);
							}
						}
						#endregion

						#region Add Files stage
						ShowMessage("Add");
						foreach(string fileName in Seeds)
						{
							if (_io.DirectoryExists(fileName))
							{
								InternalAddFolder(fileName);
							}
							else if (_io.FileExists(fileName))
							{
								InternalAddFile(fileName);
							}
							else
							{
								// file or folder is not visible
							}
						}
						#endregion
					}
					finally
					{
						EndUpdate();
						if (!Modified)
						{
							Modified = oldModified;
						}
					}
				}
				finally
				{
					UpdaterClose();
				}
			}
			finally
			{
				progress.SkipAll();
			}
		}
		
		public void Rebuild()
		{
			Detach();

			// delete all files
			foreach(string name in _store.List())
			{
				if (String.Compare(name, _settingsFile, true) != 0)
				{
					_store.DeleteFile(name);
				}
			}
			
			WriteConfigString("Index/Analyzer.Version", RecommendedAnalyzerVersion);
			LastCachedDocumentId = NoCacheId;
				
			BeginUpdate();

			DocumentsSize = 0;
			DocumentCountChanged(0, true);
			
			try
			{
				UpdaterOpen(Seeds);
				try
				{
					foreach(string fileName in Seeds)
					{
						if (_io.DirectoryExists(fileName))
						{
							InternalAddFolder(fileName);
						}
						else if (_io.FileExists(fileName))
						{
							InternalAddFile(fileName);
						}
						else
						{
							// file or folder is not visible
						}
					}
				}
				finally
				{
					UpdaterClose();
				}
			}
			finally
			{
				EndUpdate();
			}
		}
		
		private void Close()
		{
			Detach();

			if (_store != null)
			{
				_store.Dispose();
				_store = null;
			}
			
			if (_io != null)
			{
				_io.FileSystem = null;
				_io = null;
			}
		}
		
		public void Save()
		{
			SaveTo(_fileName);
		}
		
		public void SaveAs(string fileName)
		{
			SaveTo(System.IO.Path.GetFullPath(fileName));
		}
		
		private void SaveTo(string fileName)
		{
			Detach();

			string temporaryFile = fileName + ".tmp";
			try
			{
				System.IO.File.Delete(temporaryFile);
				System.IO.Compression.ZipFile.Create(new ZipProgress(this, true), temporaryFile, _store.CreateFileFeeder());
			}
			catch(Exception)
			{
				System.IO.File.Delete(temporaryFile);
				throw;
			}

			_store.CloseArchive();
			
			System.IO.File.Delete(fileName);
			System.IO.File.Move(temporaryFile, fileName);
			_fileName = fileName;
			Modified = false;
		}
		
		#region Implementation of IDisposable
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			Close();
		}

		~Index()
		{
			Dispose(false);
		}
		#endregion
		
		#region ZipFile progress
		private class ZipProgress: System.Threading.SingleProgress
		{
			private Index _parent;
			public ZipProgress(Index parent, bool interruptEnabled)
			{
				_parent = parent;
				_interruptEnabled = interruptEnabled;
				Treshold = 16 * 1024; 	// TODO: TUNE: 
			}
			
			private bool _interruptEnabled;
			
			public bool InterruptEnabled
			{
				get
				{
					return _interruptEnabled;
				}
				set
				{
					_interruptEnabled = value;
				}
			}
			
			protected override void DoOpen(ulong min, ulong max)
			{
				if (min != 0)
				{
					throw new ArgumentException();
				}
				ulong smoothTreshold = max / 500; // TODO: TUNE: how many steps
				if ((smoothTreshold > 0) && (smoothTreshold < Treshold))
				{
					Treshold = smoothTreshold;
				}
				_parent.OpenProgress(null, Convert.ToInt32(max));
			}

			protected override bool DoStep(ulong value)
			{
				bool result = _parent.ShowProgress(Convert.ToInt32(value));
				if (_interruptEnabled)
				{
					return result;
				}
				return true;
			}
			
			protected override void DoClose()
			{
				_parent.CloseProgress();
			}
		}
		#endregion
		
		#region Search progress
		private class SearchProgress: System.Threading.SlicedProgress
		{
			private Index _parent;
			public SearchProgress(Index parent, bool interruptEnabled)
			{
				_parent = parent;
				_interruptEnabled = interruptEnabled;
			}
			
			private bool _interruptEnabled;
			
			public bool InterruptEnabled
			{
				get
				{
					return _interruptEnabled;
				}
				set
				{
					_interruptEnabled = value;
				}
			}
			
			protected override void DoOpen(ulong min, ulong max)
			{
				_parent.OpenProgress(null, Convert.ToInt32(max));
			}

			protected override bool DoStep(ulong value)
			{
				bool result = _parent.ShowProgress(Convert.ToInt32(value));
				if (_interruptEnabled)
				{
					return result;
				}
				return true;
			}
			
			protected override void DoClose()
			{
				_parent.CloseProgress();
			}
			
			protected override void DoWrite(string message)
			{
				_parent.ShowMessage(message);
			}
		}
		#endregion
	}
}
