using System;

namespace Indexer.Engine
{
	public class IndexStore: Lucene.Net.Store.Directory
	{
		#region Tuning
		// TODO: TUNE: following constants can be changed without breaking functionality
		private const long ChunkingFileLengthLimit = 1024 * 1024; 	// for how large files chunking will be used
		private const long ChunkSize = 32 * 1024; 					// size of 1 chunk
		private const bool UseChunkedOutput = true;					// true if chunked output should be created,
																	// false - no chunking
		private const bool UseDelayedAccess = true;					// if true, archive will be "mapped" and cached while
																	// accessed. this mode can improve opening time for
																	// archives with chunked files
		#endregion
		
		public long Size
		{
			get
			{
				long result = 0;
				foreach(string name in List())
				{
					result += FileLength(name);
				}
				return result;
			}
		}

		#region Startup progress
		/// <summary>
		/// Purpose of this class is to measure progress of the IndexStore.OpenArchive() operation in the "delayed" mode.
		/// In that mode the zip archive is attached once quickly and the real progress happens later.
		/// This observer actually measures Indexer.Gui.OpenIndexCommand, which does some reading of the *.cfs file (about 88%
		/// if the file was not chunked, about 12% if the file was chunked) and does some lengthy System.IO.Stream.SetLength() in
		/// the chunked case. So those operations are measured by the StartUpObserver class.
		/// </summary>
		private class StartUpObserver
		{
			private System.Threading.IProgress _outerObserver;
			private IndexStore _store;
			
			public System.Threading.IProgress _commitProgress = null;
			public System.Threading.IProgress _setLengthProgress = null;
			
			public bool Enabled = false;
			
			private ulong _commitSize = 0;
			
			public void CommitPage(int pageSize)
			{
				if (!Enabled)
				{
					return;
				}
				_commitSize += (ulong)pageSize;
				Tools.ThrowExceptionIfInterrupted(_commitProgress.Step(_commitSize));
			}

			public void Open(IndexStore store, System.Threading.IProgress observer)
			{
				_outerObserver = observer;
				_store = store;
				
				if (_outerObserver != null)
				{
					_indexFileName = null;
					foreach(string file in _store.List())
					{
						if (file.EndsWith(".cfs", StringComparison.InvariantCultureIgnoreCase))
						{
							_indexFileName = file;
							break;
						}
					}
					if (_indexFileName == null)
					{
						_commitProgress = _outerObserver;
						_commitProgress.Open(0, (ulong)_store.Size);
					}
					else
					{
						File indexFile = _store._files.Get(_indexFileName);
						
						System.Threading.SlicedProgress progress = new System.Threading.SlicedProgress(_outerObserver);
						// TODO: TUNE:
						_setLengthProgress = progress.Add(300);
						_commitProgress = progress.Add(700);
						
						_setLengthProgress.Open(0, (ulong)indexFile.Length);
						
						if (indexFile.IsChunked)
						{
							_commitProgress.Open(0, Convert.ToUInt64(indexFile.Length * 0.12 /* TODO: heuristic */));
						}
						else
						{
							_commitProgress.Open(0, Convert.ToUInt64(indexFile.Length * 0.88 /* TODO: heuristic */));
						}
					}
					Enabled = true;
				}
			}
			
			public void Close()
			{
				Enabled = false;
				_store = null;
				// TODO: close progresses
			}

			private string _indexFileName = null;
			
			public Worker GetWorker(IndexStore store)
			{
				if (_indexFileName != null)
				{
					return new PageLoader(store, _indexFileName);
				}
				else
				{
					return null;
				}
			}
			
			public void SetStreamLength(System.IO.Stream stream, long length, File item)
			{
				if (Enabled && (_setLengthProgress != null) && String.Equals(item.Name, _indexFileName, StringComparison.InvariantCultureIgnoreCase))
				{
					// TODO: following operation is faster then continual expansion but is less UI responsive as the operation
					// blocks for undetermined amount of time
					//
					// stream.SetLength(length);
					
					long now = stream.Length;
					while (now < length)
					{
						long jump = System.Math.Max(128 * 1024, length / 40); // TODO: TUNE: how many commit steps to use
						
						if (now + jump > length)
						{
							jump = length - now;
						}
						
						now += jump;
						// stream.SetLength(now);
						stream.Position = now - 1;
						stream.WriteByte(0xFF);
						stream.Flush();
						
						if (!_setLengthProgress.Step((ulong)now))
						{
							throw new OperationCanceledException();
						}
					}
				}
				else
				{
					stream.SetLength(length);
					stream.Position = length - 1;
					stream.WriteByte(0xFF);
					stream.Flush();
				}
			}
		}
		
		private StartUpObserver _startup = new StartUpObserver();
		
		public void Initialized()
		{
			_startup.Close();
		}
		#endregion
		
		#region Chunking
		private const long NoOffset = -1;
		private const long NoLength = -1;
		
		private static string GetChunkName(string file, long offset, long length)
		{
 			return String.Format("{0}[{1}]", file, offset, length);			
		}
		
		private static void ParseChunkName(string file, out string baseName, out long offset)
		{
			int i = file.IndexOf("[");
			if (i < 0)
			{
				baseName = file;
				offset = NoOffset;
			}
			else
			{
				baseName = file.Substring(0, i);
				string s = file.Substring(i + 1, file.Length - i - 2);
				if (!long.TryParse(s, out offset))
				{
					baseName = file;
					offset = NoOffset;
				}
			}
		}
		
		private static string ParseChunkName(string file)
		{
			string result;
			long offset;
			ParseChunkName(file, out result, out offset);
			return result;
		}
		#endregion
		
		#region Zip I/O
		private class Feeder: System.IO.Compression.IFileFeeder
		{
			private IndexStore _store;
			private InputStreamAccessMode _access;
			public Feeder(IndexStore store, InputStreamAccessMode access)
			{
				_store = store;
				_access = access;
			}
			
			public string[] GetFiles()
			{
				return _store.List();
			}
			
			public DateTime GetLastWriteTime(string file)
			{
				return IndexStore.FileModifiedToDateTime(_store.FileModified(file));
			}
			
			public long GetLength(string file)
			{
				return _store.FileLength(file);
			}
			
			public System.IO.Stream GetInputStream(string file)
			{
				return _store.OpenInputStream(file, _access);
			}
			
			public string InternalNameToCompressedName(string file)
			{
				return file;
			}
			
			public long GetChunkSize(string file)
			{
				if (UseChunkedOutput && (GetLength(file) > ChunkingFileLengthLimit))
				{
					return ChunkSize;
				}
				else
				{
					return 0;
				}
			}
			
			public string GetChunkName(string file, long offset, long length)
			{
				return IndexStore.GetChunkName(file, offset, length);
			}
		}
		
		private class Loader: System.IO.Compression.IFileWriter
		{
			private IndexStore _store;
			
			public Loader(IndexStore store)
			{
				_store = store;
			}
			
			public void CreateDirectory(string folder)
			{
				throw new NotSupportedException();
			}
			
			public bool AddFile(string compressedName)
			{
				return true;
			}
			
			public System.IO.Stream GetOutputStream(string file, long length)
			{
				string baseName;
				long offset;
				IndexStore.ParseChunkName(file, out baseName, out offset);
				return _store.CreateOutputStream(baseName, offset, length);
			}
			
			public string CompressedNameToInternalName(string compressedName)
			{
				return compressedName;
			}
			
			public void SetLastWriteTime(string file, DateTime stamp)
			{
				_store.SetFileModified(IndexStore.ParseChunkName(file), stamp);
			}
		}
		#endregion
		
		#region Locks
		private class Lock: Lucene.Net.Store.Lock
		{
			private bool _isLocked = false;
			public override bool IsLocked()
			{
				return _isLocked;
			}
			
			public override bool Obtain()
			{
				lock(this)
				{
					if (_isLocked)
					{
						return false;
					}
					else
					{
						_isLocked = true;
						return true;
					}
				}
			}
			
			public override void Release()
			{
				lock(this)
				{
					_isLocked = false;
				}
			}
		}

		private class Locks: System.Collections.Generic.Dictionary<string, Lock>
		{
			public Lock Get(string name)
			{
				name = name.ToUpper();
				Lock result = null;
				if (!TryGetValue(name, out result))
				{
					result = new Lock();
					this.Add(name, result);
				}
				return result;
			}
			
			public void Dispose()
			{
				Clear();
			}
		}

		private Locks _locks = null;
		#endregion

		#region Input
		private class Reader: Lucene.Net.Store.IndexInput
		{
			private System.IO.Stream _stream;
			private File _file;
			public Reader(System.IO.Stream stream, File file)
			{
				_stream = stream;
				_file = file;
			}
			
			public override void Close()
			{
				if (_stream != null)
				{
					_stream.Close();
					_stream = null;
				}
			}
			
			public override long Length()
			{
				return _stream.Length;
			}
			
			public override void Seek(long pos)
			{
				_stream.Position = pos;
			}
			
			public override long GetFilePointer()
			{
				return _stream.Position;
			}
			
			private byte[] _byteBuffer = new byte[1];
			public override byte ReadByte()
			{
				ReadBytes(_byteBuffer, 0, 1);
				return _byteBuffer[0];
			}
			
			public override void ReadBytes(byte[] b, int offset, int len)
			{
				if (_stream.Read(b, offset, len) < len)
				{
					throw new System.IO.IOException("Read past EOF");
				}
			}
			
			public override object Clone()
			{
				Reader result = (Reader)_file.Store.OpenInput(_file.Name);
				result.Seek(GetFilePointer());
				return result;
			}
		}
		#endregion
		
		#region Output
		private class Writer: Lucene.Net.Store.BufferedIndexOutput
		{
			private System.IO.Stream _stream;
			private File _file;
			
			public Writer(System.IO.Stream stream, File file)
			{
				_stream = stream;
				_file = file;
			}
			
			public override void Close()
			{
				if (_stream != null)
				{
					base.Close();
					_stream.Close();
					_stream = null;
				}
			}
			
			public override long Length()
			{
				return _stream.Length;
			}

			public override void Seek(long pos)
			{
				base.Seek(pos);
				_stream.Position = pos;
			}
			
			public override void FlushBuffer(byte[] b, int len)
			{
				if (len > 0)
				{
					_stream.Write(b, 0, len);
				}
			}
		}
		#endregion
		
		#region Files
		private enum FileLocation
		{
			Folder,					// file is in the folder
			Archive,				// file is in the archive
			FolderAndArchive		// file is partially mirrored to the folder
		}

		private class PageBuffer
		{
			public int PageIndex = -1;
			public byte[] Buffer;
			public int Length;
			
			public PageBuffer(int length)
			{
				this.Length = length;
				Buffer = new byte[length];
			}
			
			public void Dispose()
			{
				if (Buffer != null)
				{
					Buffer = null;
				}
			}
		}
		
		private class PageMap
		{
			private const int PageSizeShift = 10;
			private const int PageSize = 1 << PageSizeShift;
			
			private File _file;
			private System.Collections.BitArray _bits;
			private int _pageCount;
			private int _missingPages;
			private int _lastPageSize;
			public PageMap(File file, long length)
			{
				_file = file;
				_pageCount = (int)(length / PageSize);
				long whole = ((long)_pageCount) * PageSize;
				if (length > whole)
				{
					_pageCount++;
					_lastPageSize = (int)(length - whole);
				}
				else
				{
					_lastPageSize = PageSize;
				}
				_missingPages = _pageCount;
				_bits = new System.Collections.BitArray(_pageCount);
			}

			public int Count
			{
				get
				{
					return _pageCount;
				}
			}
			
			private static int OffsetToPageIndex(long offset)
			{
				return (int)(offset >> PageSizeShift);
			}
			
			private static long PageRootOffset(int pageIndex)
			{
				return ((long)pageIndex) << PageSizeShift;
			}
			
			#region Translate cache
			long _offset = -1;
			int _pageIndex;
			int _pageOffset;
			int _pageSize;
			#endregion
			
			public void Translate(long offset, out int pageIndex, out int pageOffset, out int pageSize)
			{
				if (offset == _offset)
				{
					pageIndex = _pageIndex;
					pageOffset = _pageOffset;
					pageSize = _pageSize;
				}
				else
				{
					pageIndex = OffsetToPageIndex(offset);
					long pageRootOffset = PageRootOffset(pageIndex);
					pageOffset = (int)(offset - pageRootOffset);
					pageSize = GetPageSize(pageIndex);
					
					_offset = offset;
					_pageIndex = pageIndex;
					_pageOffset = pageOffset;
					_pageSize = pageSize;
				}
			}
			
			public void Translate(int pageIndex, out long pageRootOffset, out int pageSize)
			{
				pageRootOffset = PageRootOffset(pageIndex);
				pageSize = GetPageSize(pageIndex);
			}
			
			private int GetPageSize(int pageIndex)
			{
				if (pageIndex == _pageCount - 1)
				{
					return _lastPageSize;
				}
				else
				{
					return PageSize;		
				}
			}

			public PageBuffer CreatePageBuffer()
			{
				return new PageBuffer(PageSize);
			}
			
			private PageBuffer _pageBuffer = new PageBuffer(PageSize);

			public bool Read(long position, byte[] buffer, int offset, int count)
			{
				int pageIndex;
				int pageOffset;
				int pageSize;
				
				if (count > PageSize)
				{
					return false;
				}
				
				Translate(position, out pageIndex, out pageOffset, out pageSize);
				if (pageOffset + count <= pageSize)
				{
					if ((_pageBuffer.PageIndex == pageIndex) && _bits[pageIndex])
					{
						System.Array.Copy(_pageBuffer.Buffer, pageOffset, buffer, offset, count);
						return true;
					}
				}
				return false;
			}
			
			public bool LoadPage(int pageIndex, out PageBuffer buffer)
			{
				buffer = _pageBuffer;
				if (_bits[pageIndex])
				{
					if (buffer.PageIndex != pageIndex)
					{
						buffer.PageIndex = pageIndex;
						System.IO.Stream stream = _file.OpenFolderStream();
						try
						{
							stream.Position = PageRootOffset(pageIndex);
							stream.Read(buffer.Buffer, 0, GetPageSize(pageIndex));
						}
						finally
						{
							_file.ReleaseFolderStream();
						}
					}
					return true;
				}
				else
				{
					return false;
				}
			}
			
			public bool Cached(int pageIndex)
			{
				return _bits[pageIndex];
			}
			
			public bool Cached(long offset, int count)
			{
				if (_file.Location == FileLocation.Folder)
				{
					return true;
				}
				int pageIndex;
				int pageOffset;
				int pageSize;
				Translate(offset, out pageIndex, out pageOffset, out pageSize);
				if (!_bits[pageIndex])
				{
					return false;
				}
				if (pageOffset + count <= pageSize)
				{
					// within single page
					return true;
				}
				return Cached(PageRootOffset(pageIndex + 1), count - (pageSize - pageOffset));
			}
			
			public void CommitPage(int pageIndex, PageBuffer buffer)
			{
				buffer.PageIndex = pageIndex;
				System.IO.Stream stream = _file.OpenFolderStream();
				try
				{
					long pageRootOffset = PageRootOffset(pageIndex);
					if (stream.Length < pageRootOffset)
					{
						stream.SetLength(pageRootOffset);
					}
					if (stream.Position != pageRootOffset)
					{
						stream.Position = pageRootOffset;
					}
					int pageSize = GetPageSize(pageIndex);
					stream.Write(buffer.Buffer, 0, pageSize);
					_bits[pageIndex] = true;
					
					_missingPages--;
					if (_missingPages == 0)
					{
						// switch to full folder mode
						_file.CloseChunkStreams();
						_file.Location = FileLocation.Folder;
					}
					
					_file.Store._startup.CommitPage(pageSize);
				}
				finally
				{
					_file.ReleaseFolderStream();
				}
			}
			
			public void Dispose()
			{
				_bits = null;
				if (_pageBuffer != null)
				{
					_pageBuffer.Dispose();
					_pageBuffer = null;
				}
			}
		}
		
		private class File
		{
			internal IndexStore Store;
			
			public FileLocation Location;

			public bool IsInFolder
			{
				get
				{
					return (Location == FileLocation.Folder) || (Location == FileLocation.FolderAndArchive);
				}
			}
			
			public bool IsVolatile;	// if true, file may be modified outside the store

			public string FolderPath
			{
				get
				{
					return System.IO.Path.Combine(Store.Folder, Name);
				}
			}
			
			public string ArchivePath
			{
				get
				{
					return Name;
				}
			}
			
			public string Name;
			private DateTime _lastWriteTime;
			public DateTime LastWriteTime
			{
				get
				{
					if (IsVolatile)
					{
						return (new System.IO.FileInfo(FolderPath)).LastWriteTime;
					}
					else
					{
						return _lastWriteTime;
					}
				}
				set
				{
					_lastWriteTime = value;
				}
			}
			
			private long _length;
			public long Length
			{
				get
				{
					if (IsVolatile)
					{
						return (new System.IO.FileInfo(FolderPath)).Length;
					}
					else
					{
						return _length;
					}
				}
				set
				{
					_length = value;
				}
			}
			
			#region Chunking
			public class Chunk
			{
				public File Owner;
				public long Offset;
				public ICSharpCode.SharpZipLib.Zip.ZipEntry ArchiveHandle;
				public long Length;
				
				internal Chunk()
				{
				}
				
				public Chunk(long offset, ICSharpCode.SharpZipLib.Zip.ZipEntry entry)
				{
					this.Offset = offset;
					this.ArchiveHandle = entry;
					Length = entry.Size;
				}
				
				private System.IO.Stream _stream = null;
				public System.IO.Stream Stream
				{
					get
					{
						if (_stream == null)
						{
							_stream = Owner.Store._archive.GetInputStream(ArchiveHandle);
							if (_stream != null)
							{
								_stream = new System.IO.SeekableInputStream(_stream, (new Rewinder(Owner.Store._archive, ArchiveHandle)).CreateStream, Length);
								Owner.AddChunkStream(this);
							}
							else
							{
								throw new System.IO.FileNotFoundException();
							}
						}
						return _stream;
					}
					set
					{
						if (_stream != value)
						{
							if (_stream != null)
							{
								_stream.Close();
								_stream.Dispose();
								Owner.RemoveChunkStream(this);
							}
							_stream = value;
						}
					}
				}
			}
			
			private class Chunks: System.Collections.Generic.List<Chunk>
			{
				private File _owner;
				public Chunks(File owner)
				{
					_owner = owner;	
				}
				
				private Chunk _searcher = new Chunk();
				
				private class Comparer: System.Collections.Generic.Comparer<Chunk>
				{
					public override int Compare(Chunk x, Chunk y)
					{
						if (x.Offset == y.Offset)
						{
							return 0;
						}
						else if (x.Offset < y.Offset)
						{
							return -1;
						}
						else
						{
							return 1;
						}
					}
				}
				
				public void Dispose()
				{
					foreach(Chunk item in this)
					{
						item.ArchiveHandle = null;
					}
					base.Clear();
				}
				
				public void Add(long offset, ICSharpCode.SharpZipLib.Zip.ZipEntry entry)
				{
					Chunk item = new Chunk(offset, entry);
					item.Owner = _owner;
					int index = BinarySearch(item, new Comparer());
					if (index < 0)
					{
						Insert(~index, item);
					}
				}
				
				public Chunk Find(long position)
				{
					_searcher.Offset = position;
					int index = BinarySearch(_searcher, new Comparer());
					Chunk result;
					if (index < 0)
					{
						index = ~index;
						/* Now 'index' is index of the next element that is larger than value or,
						 * if there is no larger element, then it is the Count
						 */
						result = this[index - 1];
					}
					else
					{
						result = this[index];
					}
					// TODO: remove once verified
					System.Diagnostics.Debug.Assert((result.Offset <= position) && (position < result.Offset + result.Length));
					return result;
				}
			}
			
			public void Add(long chunkOffset, ICSharpCode.SharpZipLib.Zip.ZipEntry chunkEntry)
			{
				Length = Length + chunkEntry.Size;
				if (_chunks == null)
				{
					_chunks = new Chunks(this);
				}
				_chunks.Add(chunkOffset, chunkEntry);
			}
			
			private Chunks _chunks;
			
			public bool IsChunked
			{
				get
				{
					return (_chunks != null) && (_chunks.Count > 1);
				}
			}
			
			public ICSharpCode.SharpZipLib.Zip.ZipEntry ArchiveHandle
			{
				get
				{
					if (IsChunked || (_chunks == null))
					{
						return null;
					}
					else
					{
						return _chunks[0].ArchiveHandle;
					}
				}
			}

			private Chunk _lastChunk = null;
			public Chunk GetChunk(long position, out long offset)
			{
				Chunk result = _lastChunk;
				if (result != null)
				{
					offset = position - result.Offset;
					if ((offset >= 0) && (offset < result.Length))
					{
						return result;
					}
				}
				result = _chunks.Find(position);
				_lastChunk = result;
				offset = position - result.Offset;
				return result;
			}
			
			private System.Collections.Generic.List<Chunk> _chunkStreams = null;
			
			internal void AddChunkStream(Chunk chunk)
			{
				if (_chunkStreams == null)
				{
					_chunkStreams = new System.Collections.Generic.List<Chunk>();
				}
				_chunkStreams.Add(chunk);
			}
			
			internal void RemoveChunkStream(Chunk chunk)
			{
				_chunkStreams.Remove(chunk);
			}
			
			public void CloseChunkStreams()
			{
				if (_chunkStreams != null)
				{
					foreach(Chunk chunk in _chunkStreams.ToArray())
					{
						chunk.Stream = null;
					}
					_chunkStreams.Clear();
					_chunkStreams = null;
				}
			}
			#endregion
			
			public void Dispose()
			{
				if (Cache != null)
				{
					Cache.Dispose();
					Cache = null;
				}
				if (_chunks != null)
				{
					CloseChunkStreams();
					_chunks.Dispose();
					_chunks = null;
				}
				if (_folderStream != null)
				{
					_folderStream.Close();
					_folderStream.Dispose();
					_folderStream = null;
					_folderStreamUseCount = 0;
				}
			}
			
			private System.IO.Stream _folderStream = null;
			private int _folderStreamUseCount = 0;
			public void ReleaseFolderStream()
			{
				_folderStreamUseCount--;
				if (_folderStreamUseCount == 0)
				{
					_folderStream.Close();
					_folderStream.Dispose();
					CloseChunkStreams();
					_folderStream = null;
				}
			}
			
			public System.IO.Stream OpenFolderStream()
			{
				if (_folderStreamUseCount == 0)
				{
					if (Cache == null)
					{
						Cache = new PageMap(this, Length);
					}
					_folderStream = new System.IO.FileStream(FolderPath, System.IO.FileMode.OpenOrCreate, System.IO.FileAccess.ReadWrite);
					if (Length > 0)
					{
						// TODO: write to the stream in order to allocate stream sectors now, so that the
						// speed of next operations can be estimated.
						//
						// This step is optional
						Store._startup.SetStreamLength(_folderStream, Length, this);
					}
				}
				_folderStreamUseCount++;
				return _folderStream;
			}
			
			public PageMap Cache = null;
			
			public void LoadAll()
			{
				if (Location == FileLocation.Folder)
				{
					return;
				}
				System.IO.Stream stream = Store.OpenInputStream(Name, InputStreamAccessMode.Cached);
				try
				{
					PageBuffer page = Cache.CreatePageBuffer();
					try
					{
						for(int pageIndex = 0; pageIndex < Cache.Count; pageIndex++)
						{
							if (!Cache.Cached(pageIndex))
							{
								// reading the page will commit it
								long position;
								int count;
								Cache.Translate(pageIndex, out position, out count);
								
								stream.Position = position;
								stream.Read(page.Buffer, 0, count);
							}
						}
					}
					finally
					{
						page.Dispose();
					}
				}
				finally
				{
					stream.Dispose();
				}
			}
			
			public bool LoadPage(int pageIndex)
			{
				if (_folderStream == null)
				{
					return false;
				}
				System.IO.Stream stream = Store.OpenInputStream(Name, InputStreamAccessMode.Cached);
				try
				{
					PageBuffer page = Cache.CreatePageBuffer();
					try
					{
						// reading the page will commit it
						long position;
						int count;
						Cache.Translate(pageIndex, out position, out count);
						
						stream.Position = position;
						stream.Read(page.Buffer, 0, count);
					}
					finally
					{
						page.Dispose();
					}
				}
				finally
				{
					stream.Dispose();
				}
				return true;
			}
		}
		
		private class Files: System.Collections.Generic.Dictionary<string, File>
		{
			private IndexStore _store;
			public Files(IndexStore store)
			{
				_store = store;
			}
			
			public File Get(string name)
			{
				name = name.ToUpper();
				File result = null;
				TryGetValue(name, out result);
				return result;
			}

			private void Add(File item)
			{
				base.Add(item.Name.ToUpper(), item);
			}
			
			public File Add(string fileName)
			{
				string name = System.IO.Path.GetFileName(fileName);
				File result = Get(name);
				if (result == null)
				{
					result = new File();
					result.Store = _store;
					result.Name = name;
					System.IO.FileInfo info = new System.IO.FileInfo(fileName);
					result.LastWriteTime = info.LastWriteTime;
					result.Length = info.Length;
					result.Location = FileLocation.Folder;
					result.IsVolatile = true;
					Add(result);
				}
				return result;
			}
			
			public File Add(ICSharpCode.SharpZipLib.Zip.ZipEntry entry)
			{
				File result = Get(entry.Name);
				if (result == null)
				{
					result = Add(entry.Name, 0, entry);
				}
				return result;
			}
			
			public File Add(string fileName, long chunkOffset, ICSharpCode.SharpZipLib.Zip.ZipEntry chunkEntry)
			{
				File result = Get(fileName);
				if (result == null)
				{
					result = new File();
					result.Store = _store;
					result.Name = fileName;
					result.Location = FileLocation.Archive;
					result.Length = 0;
					result.LastWriteTime = chunkEntry.DateTime;
					Add(result);
				}
				result.Add(chunkOffset, chunkEntry);
				return result;
			}
			
			public void Remove(File item)
			{
				base.Remove(item.Name.ToUpper());
			}
			
			public string[] List()
			{
				string[] results = new string[Count];
				int index = 0;
				foreach(System.Collections.Generic.KeyValuePair<string, File> entry in this)
				{
					results[index] = entry.Value.Name;
					index++;
				}
				return results;
			}
			
			public void Dispose()
			{
				foreach(System.Collections.Generic.KeyValuePair<string, File> entry in this)
				{
					entry.Value.Dispose();
				}
				Clear();
			}
		}
		
		private Files _files = null;
		#endregion

		private class SmartInputStream: System.IO.Stream
		{
			private System.IO.Stream _folderStream;
			private File _file;
			private long _position;
			
			public SmartInputStream(File file)
			{
				_file = file;
				_folderStream = _file.OpenFolderStream();
				_position = 0;
			}
			
			protected override void Dispose(bool disposing)
			{
				Close();
				base.Dispose(disposing);
			}
			
			public override bool CanRead 
			{
				get 
				{ 
					return true;
				}
			}
			
			public override bool CanSeek 
			{
				get 
				{
					return true;
				}
			}
			
			public override bool CanWrite 
			{
				get 
				{
					return false;
				}
			}
			
			public override void Close()
			{
				if (_folderStream != null)
				{
					_file.ReleaseFolderStream();
					_folderStream = null;
				}
			}
			
			public override long Length 
			{
				get 
				{
					return _file.Length;
				}
			}
			
			public override long Position 
			{
				get 
				{
					return _position;
				}
				set 
				{
					if (value != _position)
					{
						Seek(value, System.IO.SeekOrigin.Begin);
					}
				}
			}
			
			public override void Flush()
			{
			}
			
			public override void Write(byte[] buffer, int offset, int count)
			{
				if (count > 0)
				{
					throw new NotSupportedException();
				}
			}
			
			public override void SetLength(long value)
			{
				if (value != Length)
				{
					throw new NotSupportedException();
				}
			}
			
			public override long Seek(long offset, System.IO.SeekOrigin origin)
			{
				long currentPosition = Position;
				long newPosition;
				
				switch (origin)
				{
					case System.IO.SeekOrigin.Begin:
						newPosition = offset;
						break;
						
					case System.IO.SeekOrigin.Current:
						newPosition = currentPosition + offset;
						break;
						
					case System.IO.SeekOrigin.End:
						newPosition = Length + offset;
						break;
						
					default:
						newPosition = currentPosition;
						break;
				}
				
				if (newPosition < 0)
				{
					throw new ArgumentException("Attempted seeking before the start of the stream");
				}
				
				if (newPosition > Length)
				{
					throw new ArgumentException("Attempted seeking after the end of the stream"); 
				}

				_position = newPosition;
				return _position;
			}
	
			public override int Read(byte[] buffer, int offset, int count)
			{
				if (count == 0)
				{
					return 0;
				}
				if (_position + count > Length)
				{
					count = (int)(Length - _position);
					if (count <= 0)
					{
						return 0;
					}
				}
				if (count == 1)
				{
					buffer[offset] = InternalReadByte(_position);
					_position++;
					return 1;
				}
				if (_file.Cache.Read(_position, buffer, offset, count))
				{
					_position += (long)count;
					return count;
				}
				if (_file.Cache.Cached(_position, count))
				{
					if (_folderStream.Position != _position)
					{
						_folderStream.Position = _position;
					}
					int result = _folderStream.Read(buffer, offset, count);
					_position += (long)result;
					return result;
				}
				else
				{
					int result = count;
					do
					{
						int jump = InternalReadBytes(_position, buffer, offset, count);
						_position += (long)jump;
						offset += jump;
						count -= jump;
					}
					while (count > 0);
					return result;
				}
			}
			
			private void LoadPage(int pageIndex, PageBuffer buffer)
			{
				long pageStartPosition;
				int pageSize;
				
				_file.Cache.Translate(pageIndex, out pageStartPosition, out pageSize);
				
				// we need to populate page 'pageIndex' that goes from 'pageStartposition'..+'pageSize'
				for(int bufferOffset = 0; bufferOffset < pageSize; bufferOffset++)
				{
					buffer.Buffer[bufferOffset] = ReadChunkByte(pageStartPosition);
					pageStartPosition++;
				}
				
				_file.Cache.CommitPage(pageIndex, buffer);
			}
			
			private byte InternalReadByte(long position)
			{
				int pageIndex;
				int pageOffset;
				int pageSize;
				_file.Cache.Translate(position, out pageIndex, out pageOffset, out pageSize);
				PageBuffer page;
				if (!_file.Cache.LoadPage(pageIndex, out page))
				{
					LoadPage(pageIndex, page);
				}
				return page.Buffer[pageOffset];
			}
			
			private int InternalReadBytes(long position, byte[] buffer, int offset, int count)
			{
				int pageIndex;
				int pageOffset;
				int pageSize;
				_file.Cache.Translate(position, out pageIndex, out pageOffset, out pageSize);
				PageBuffer page;
				if (!_file.Cache.LoadPage(pageIndex, out page))
				{
					LoadPage(pageIndex, page);
				}
				
				int result = pageSize - pageOffset;
				if (result > count)
				{
					result = count;
				}
				System.Array.Copy(page.Buffer, pageOffset, buffer, offset, result);
				return result;
			}
			
			private byte ReadChunkByte(long position)
			{
				long chunkOffset;
				File.Chunk chunk = _file.GetChunk(position, out chunkOffset);
				System.IO.Stream stream = chunk.Stream;
				if (stream.Position < chunkOffset)
				{
					int pageIndex;
					int pageOffset;
					int pageSize;
					_file.Cache.Translate(chunk.Offset + stream.Position, out pageIndex, out pageOffset, out pageSize);

					if (pageOffset > 0)
					{
						// chunk stream does not start new page. It meaans the page it starts on is composed of
						// more chunks and loading the page would cause the previous chunk to be loaded which recursively
						// might cause all previous chunks being loaded.
						// so we will skip few bytes and start since the first page that fully covers current chunk.
						long thisPageStartPosition = chunk.Offset + stream.Position - (long)pageOffset;
						long nextPageStartPosition = thisPageStartPosition + (long)pageSize;
						
						long safeChunkOffset = nextPageStartPosition - chunk.Offset;
						
						if (safeChunkOffset >= chunkOffset)
						{
							// we would skip the byte we really want to read
							stream.Position = chunkOffset;
						}
						else
						{
							stream.Position = safeChunkOffset;
						}
					}

					if (stream.Position < chunkOffset)
					{
						PageBuffer buffer = _file.Cache.CreatePageBuffer();
						do
						{
							_file.Cache.Translate(chunk.Offset + stream.Position, out pageIndex, out pageOffset, out pageSize);
							if (_file.Cache.Cached(pageIndex))
							{
								stream.Position = stream.Position - (long)pageOffset + (long)pageSize;
							}
							else
							{
								// stream.Position changes inside the LoadPage()
								LoadPage(pageIndex, buffer);
							}
						}
						while (stream.Position < chunkOffset);
						buffer.Dispose();
					}
				}
				
				if (stream.Position > chunkOffset)
				{
					stream.Position = chunkOffset;
				}
				
				byte result = (byte)stream.ReadByte();
				if (chunkOffset + 1 == chunk.Length)
				{
					// it is not needed anymore
					chunk.Stream = null;
				}
				return result;
			}
		}
		
		#region Folder management and out-of-Lucene interfaces
		private string _folder;
		
		private void MakeSureFolderExists()
		{
			if (Folder == null)
			{
				Folder = null;
			}
		}
		
		public string Folder
		{
			get
			{
				if (String.IsNullOrEmpty(_folder))
				{
					// trigger autocreate
					Folder = null;
				}
				return _folder;
			}
			private set
			{
				Clear();
				if (String.IsNullOrEmpty(value))
				{
					_folder = System.IO.Path.GetTempFileName();
					System.IO.File.Delete(_folder);
					System.IO.Directory.CreateDirectory(_folder);
					if (!_folder.EndsWith("\\"))
					{
						_folder = _folder + "\\";
					}
				}
				else
				{
					if (!value.EndsWith("\\"))
					{
						_folder = value + "\\";
					}
					else
					{
						_folder = value;
					}
					Load();
				}
			}
		}
		
		public void Clear()
		{
			if (_locks != null)
			{
				_locks.Dispose();
			}
			_locks = new Locks();
			if (_files != null)
			{
				_files.Dispose();
			}
			_files = new Files(this);
			if (_archive != null)
			{
				_archive.Close();
				_archive = null;
			}
			if (!String.IsNullOrEmpty(_folder))
			{
				try
				{
					System.IO.Directory.Delete(_folder, true);
				}
				catch(Exception)
				{
					// TODO: at this point some garbage might be left in the temporary folder
				}
				_folder = null;
			}
		}

		private void Load()
		{
			foreach(string fileName in System.IO.Directory.GetFiles(Folder, "*.*", System.IO.SearchOption.TopDirectoryOnly))
			{
				_files.Add(fileName);
			}
		}
		
		public enum InputStreamAccessMode
		{
			SingleShotSequentialScan,		// stream will be read in forward-only manner and probably only once
			SingleShotRandomAccess,			// stream will be read only once with seeking - use local cache for large files
			MultipleSequentialScan,			// stream will be accessed sequentially multiple times - use local caching for large files
			MultipleRandomAccess,			// stream will be accessed by seeking muliple times - use local caching
			
			Cached							// stream will be cached
		}
		
		private class Rewinder
		{
			private WeakReference _fileSystem;
			private ICSharpCode.SharpZipLib.Zip.ZipEntry _entry;
			public Rewinder(ICSharpCode.SharpZipLib.Zip.ZipFile fileSystem, ICSharpCode.SharpZipLib.Zip.ZipEntry entry)
			{
				_entry = entry;
				_fileSystem = new WeakReference(fileSystem, false);
			}
			
			public System.IO.Stream CreateStream()
			{
				ICSharpCode.SharpZipLib.Zip.ZipFile fileSystem = (ICSharpCode.SharpZipLib.Zip.ZipFile)_fileSystem.Target;
				if (fileSystem == null)
				{
					throw new ObjectDisposedException("ICSharpCode.SharpZipLib.Zip.ZipFile");
				}
				return fileSystem.GetInputStream(_entry);
			}
		}
		
		public System.IO.Stream OpenInputStream(string name, InputStreamAccessMode access)
		{
			File item = _files.Get(name);
			if (item == null)
			{
				throw new System.IO.FileNotFoundException();
			}
			
			switch (item.Location)
			{
				case FileLocation.Folder:
					return new System.IO.FileStream(item.FolderPath, System.IO.FileMode.Open, System.IO.FileAccess.Read);
					
				case FileLocation.FolderAndArchive:
				case FileLocation.Archive:
					{
						bool inCache = item.Location == FileLocation.FolderAndArchive;
						bool useCache = false;
						bool largeFile = item.Length >= 100 * 1024 /* TUNE: */;
						
						switch (access)
						{
							case InputStreamAccessMode.SingleShotSequentialScan:
								// stream will be read in forward-only manner and probably only once
								break;
								
							case InputStreamAccessMode.SingleShotRandomAccess:
								// stream will be read only once with seeking - use local cache for large files
								if (largeFile)
								{
									useCache = true;
								}
								break;
								
							case InputStreamAccessMode.MultipleSequentialScan:
								// stream will be accessed sequentially multiple times - use local caching for large files
								if (largeFile)
								{
									useCache = true;
								}
								break;
								
							case InputStreamAccessMode.MultipleRandomAccess:
								// stream will be accessed by seeking muliple times - use local caching
								useCache = true;
								break;
								
							case InputStreamAccessMode.Cached:
								useCache = true;
								break;
						}
						
						if (item.IsChunked)
						{
							// smart stream must be used because simpler SeekableInputStream doesn't support multiple base streams
							useCache = true;
						}
						
						if (inCache || useCache)
						{
							item.Location = FileLocation.FolderAndArchive;
							return new SmartInputStream(item);
						}
						else
						{
							System.IO.Stream archiveStream = _archive.GetInputStream(item.ArchiveHandle);
							if (archiveStream != null)
							{
								archiveStream = new System.IO.SeekableInputStream(archiveStream, (new Rewinder(_archive, item.ArchiveHandle)).CreateStream, item.Length);
							}
							else
							{
								throw new System.IO.FileNotFoundException();
							}
							return archiveStream;
						}
					}
					
				default:
					// TODO: anything to implement?
					throw new NotImplementedException();
			}
		}
		
		public System.IO.Stream CreateOutputStream(string name)
		{
			return CreateOutputStream(name, NoOffset, NoLength);
		}

		public string GetFullPath(string name)
		{
			return System.IO.Path.Combine(Folder, name);
		}
		
		public System.IO.Stream CreateOutputStream(string name, long offset, long length)
		{
			if (length == NoLength)
			{
				length = 0;
			}
			if (offset == NoOffset)
			{
				DeleteFile(name);
				string fileName = System.IO.Path.Combine(Folder, name);
				System.IO.Stream result = new System.IO.FileStream(fileName, System.IO.FileMode.Create);
				if (length > 0)
				{
					result.SetLength(length);
				}
				_files.Add(fileName);
				return result;
			}
			else
			{
				string fileName = System.IO.Path.Combine(Folder, name);
				System.IO.Stream result = new System.IO.FileStream(fileName, System.IO.FileMode.OpenOrCreate, System.IO.FileAccess.ReadWrite);
				File item = _files.Get(name);
				if (item == null)
				{
					_files.Add(fileName);
				}
				if (offset + length > result.Length)
				{
					result.SetLength(offset + length);
				}
				result.Position = offset;
				return result;
			}
		}
		
		public System.IO.Compression.IFileFeeder CreateFileFeeder()
		{
			return new Feeder(this, InputStreamAccessMode.Cached);
		}
		
		#region Archive stuff
		private ICSharpCode.SharpZipLib.Zip.ZipFile _archive = null;
		
		public void OpenArchive(System.Threading.IProgress observer, string fileName)
		{
			fileName = System.IO.Path.GetFullPath(fileName);

			MakeSureFolderExists();
				
			if (!UseDelayedAccess)
			{
				System.IO.Compression.ZipFile.Extract(observer, fileName, new Loader(this));
			}
			else
			{
				_archive = new ICSharpCode.SharpZipLib.Zip.ZipFile(fileName, ICSharpCode.SharpZipLib.Zip.ZipFile.OpenMode.Fast);
				foreach(ICSharpCode.SharpZipLib.Zip.ZipEntry entry in _archive)
				{
					if (entry.IsDirectory)
					{
						continue;
					}
					string baseName;
					long offset;
					ParseChunkName(entry.Name, out baseName, out offset);
					if (offset == NoOffset)
					{
						_files.Add(entry);
					}
					else
					{
						_files.Add(baseName, offset, entry);
					}
				}
				_startup.Open(this, observer);
			}
		}
		
		public void CloseArchive()
		{
			if (_archive != null)
			{
				if (_files != null)
				{
					foreach(string file in List())
					{
						_files.Get(file).LoadAll();
					}
				}
				_archive.Close();
				_archive = null;
			}
		}
		#endregion
		#endregion
		
		#region Tools
		public static DateTime FileModifiedToDateTime(long fileModified)
		{
			return new DateTime(fileModified);
		}
		#endregion
		
		public IndexStore()
		{
			Clear();
		}
		
		public void Dispose()
		{
			Clear();
			_locks = null;
			_files = null;
		}

		#region Worker
		public Worker GetWorker()
		{
			return _startup.GetWorker(this);
		}
		
		private class PageLoader: Worker
		{
			private IndexStore _store;
			private string _fileName;
			private int _pageIndex = 0;
			
			public PageLoader(IndexStore store, string fileName)
			{
				_store = store;
				_fileName = fileName;
			}
			
			public override bool Execute()
			{
				File item = _store._files.Get(_fileName);
				if (item == null)
				{
					return false;
				}
				if (item.Location == FileLocation.Folder)
				{
					return false;
				}
				while (_pageIndex < item.Cache.Count)
				{
					if (!item.Cache.Cached(_pageIndex))
					{
						item.LoadPage(_pageIndex);
						_pageIndex++;
						if (item.Location == FileLocation.Folder)
						{
							return false;
						}
						else
						{
							return true;
						}
					}
					else
					{
						_pageIndex++;
					}
				}
				return false;
			}
			
			public override void Dispose()
			{
				_store = null;
			}
		}
		#endregion
		
		#region Lucene Directory interface
		public override void Close()
		{
		}
		
		public override Lucene.Net.Store.IndexOutput CreateOutput(string name)
		{
			System.IO.Stream stream = CreateOutputStream(name);
			File item = _files.Get(name);
			return new Writer(stream, item);
		}
		
		public override void DeleteFile(string name)
		{
			File item = _files.Get(name);
			if (item != null)
			{
				if (item.IsInFolder)
				{
					System.IO.File.Delete(item.FolderPath);
				}
				_files.Remove(item);
				item.Dispose();
			}
		}
		
		public override bool FileExists(string name)
		{
			return _files.Get(name) != null;
		}
		
		public override long FileLength(string name)
		{
			File item = _files.Get(name);
			return (item == null) ? 0 : item.Length;
		}
		
		public override long FileModified(string name)
		{
			File item = _files.Get(name);
			return (item == null) ? 0 : item.LastWriteTime.Ticks;
		}

		public void SetFileModified(string name, DateTime stamp)
		{
			File item = _files.Get(name);
			if (item != null)
			{
				item.LastWriteTime = stamp;
				if (item.IsInFolder)
				{
					(new System.IO.FileInfo(item.FolderPath)).LastWriteTime = stamp;
				}
			}
		}
		
		public override string[] List()
		{
			return _files.List();
		}
		
		public override Lucene.Net.Store.Lock MakeLock(string name)
		{
			return _locks.Get(name);
		}
		
		public override Lucene.Net.Store.IndexInput OpenInput(string name)
		{
			System.IO.Stream stream = OpenInputStream(name, InputStreamAccessMode.MultipleRandomAccess);
			File item = _files.Get(name);
			return new Reader(stream, item);
		}
		
		public override void RenameFile(string from, string to)
		{
			File item = _files.Get(from);
			if (item != null)
			{
				// TODO: make sure item is in folder
				DeleteFile(to);
				string targetFileName = System.IO.Path.Combine(Folder, to);
				System.IO.File.Move(item.FolderPath, targetFileName);
				_files.Remove(item);
				_files.Add(targetFileName);
			}
		}
		
		public override void TouchFile(string name)
		{
			// TODO: implement
			throw new NotImplementedException();
		}
		#endregion
	}
}
