using System;
using Indexer.Engine;

namespace Indexer.Gui
{
	public abstract class WorkItem
	{
		public abstract void Execute();
		protected abstract void DisconnectThread();
		private System.Exception _error;
		public System.Exception Error
		{
			get
			{
				return _error;
			}
			set
			{
				_error = value;
			}
		}
		private System.Threading.ThreadPriority _oldPriority;
		internal void DoWork()
		{
			_oldPriority = System.Threading.Thread.CurrentThread.Priority;
			try
			{
				Worker.WorkItemRunning();
				
				_error = null;
				try
				{
					Execute();
				}
				catch (Exception error)
				{
					_error = error;
				}
			}
			finally
			{
				// we must restore the priority before the thread returns to the pool
				System.Threading.Thread.CurrentThread.Priority = _oldPriority;
				DisconnectThread();
			}
		}
		private System.Threading.IProgress _progress;
		protected System.Threading.IProgress Progress
		{
			get
			{
				return _progress;
			}
		}
		internal Worker _worker;
		internal Worker Worker
		{
			get
			{
				return _worker;
			}
			set
			{
				_worker = value;
				_progress = value;
			}
		}
	}
	
	public class Worker: IIndexObserver, System.Threading.IProgress, IDisposable
	{
		private bool _isBusy = false;
		public bool IsBusy
		{
			get
			{
				return _isBusy;
			}
		}
		
		private bool _isBackground = false;
		public bool IsBackground
		{
			get
			{
				return _isBackground;
			}
			set
			{
				if (_isBackground != value)
				{
					_isBackground = value;
					SetWorkersPriority();
				}
			}
		}
		
		private IIndexObserver _observer;
		public Worker(IIndexObserver observer)
		{
			_observer = observer;
		}
		
		public void Execute(WorkItem command)
		{
			command.Worker = this;
			try
			{
				System.Windows.Forms.MethodInvoker invoker = new System.Windows.Forms.MethodInvoker(delegate () { command.DoWork(); });
				command.Error = null;
				_isBusy = true;
				Open();
				try
				{
					IAsyncResult request = invoker.BeginInvoke(null, null);
					while (!request.IsCompleted)
					{
						System.Windows.Forms.Application.DoEvents();
						// TODO: some other yield to release the CPU?
						System.Threading.Thread.Sleep(100);
					}
					invoker.EndInvoke(request);
				}
				finally
				{
					Close();
					_isBusy = false;
				}
			}
			finally
			{
				command.Worker = null;
			}
			if (command.Error != null)
			{
				throw command.Error;
			}
		}
		
		private volatile System.Threading.Thread WorkerThread = null;
		
		internal void WorkItemRunning()
		{
			WorkerThread = System.Threading.Thread.CurrentThread;
			SetWorkersPriority();
		}
		
		private void SetWorkersPriority()
		{
			System.Threading.ThreadPriority priority;
			if (IsBackground)
			{
				priority = System.Threading.ThreadPriority.Lowest;
			}
			else
			{
				// TODO: this is some guess...
				if (System.Environment.ProcessorCount > 1)
				{
					priority = System.Threading.ThreadPriority.AboveNormal;
				}
				else
				{
					priority = System.Threading.ThreadPriority.Normal;
				}
			}
			if (WorkerThread != null)
			{
				try
				{
					WorkerThread.Priority = priority;
				}
				catch(Exception)
				{
					// TODO: shouldn't happen, use some locks etc.
				}
			}
		}
		
		#region Locking
		private Object _lock = new Object();
		
		private void Lock()
		{
			System.Threading.Monitor.Enter(_lock);
		}
		
		private void UnLock()
		{
			System.Threading.Monitor.Exit(_lock);
		}
		#endregion

		#region IIndexObserver consumer (running in the UI thread)
		private void Open()
		{
			_lastStatus = Status.New();
			_status = Status.New();
			if (_updater == null)
			{
				_updater = new System.Windows.Forms.Timer();
				_updater.Interval = 300; 	// TUNE: update frequency in miliseconds
				_updater.Tick += delegate(object sender, EventArgs e) { this.Update(); };
			}
			_updater.Enabled = true;
		}
		
		private void Close()
		{
			WorkerThread = null;
			if (_updater != null)
			{
				_updater.Enabled = false;
				Update();
			}
		}
		
		private System.Windows.Forms.Timer _updater = null;
		
		private Status _lastStatus;
		
		private void Update()
		{
			Lock();
			Status newStatus = _status;
			UnLock();

			if (_lastStatus.DocumentCount != newStatus.DocumentCount)
			{
				_lastStatus.DocumentCount = newStatus.DocumentCount;
				if (_observer != null)
				{
					_observer.DocumentCountChanged(_lastStatus.DocumentCount);
				}
			}
			if (_lastStatus.FoundDocumentCount != newStatus.FoundDocumentCount)
			{
				_lastStatus.FoundDocumentCount = newStatus.FoundDocumentCount;
				if (_observer != null)
				{
					_observer.FoundDocumentCountChanged(_lastStatus.FoundDocumentCount);
				}
			}
			if (_lastStatus.ProgressVisible != newStatus.ProgressVisible)
			{
				_lastStatus.ProgressVisible = newStatus.ProgressVisible;
				if (_lastStatus.ProgressVisible)
				{
					_lastStatus.ProgressSize = newStatus.ProgressSize;
					_lastStatus.ProgressMessage = newStatus.ProgressMessage;
					// show progress
					if (_observer != null)
					{
						_observer.OpenProgress(_lastStatus.ProgressMessage, _lastStatus.ProgressSize);
					}
				}
				else
				{
					// hide progress
					if (_observer != null)
					{
						_observer.CloseProgress();
					}
				}
			}
			bool pollInterruptRequested = true;
			if (_lastStatus.ProgressVisible)
			{
				if (_lastStatus.ProgressValue != newStatus.ProgressValue)
				{
					pollInterruptRequested = false;
					_lastStatus.ProgressValue = newStatus.ProgressValue;
					if (_observer != null)
					{
						bool canceled = !_observer.ShowProgress(_lastStatus.ProgressValue);
						if (canceled)
						{
							Lock();
							_status.Cancelled = true;
							UnLock();
						}
					}
				}
			}
			if (pollInterruptRequested)
			{
				if (_observer != null)
				{
					bool canceled = _observer.InterruptRequested();
					if (canceled)
					{
						Lock();
						_status.Cancelled = true;
						UnLock();
					}
				}
			}
			if (_lastStatus.ProgressMessage != newStatus.ProgressMessage)
			{
				_lastStatus.ProgressMessage = newStatus.ProgressMessage;
				if (_observer != null)
				{
					_observer.ShowMessage(_lastStatus.ProgressMessage);
				}
			}
		}
		#endregion
		
		#region IIndexObserver interface implementation
		private struct Status
		{
			public int DocumentCount;
			public int FoundDocumentCount;
			public string ProgressMessage;
			public int ProgressSize;
			public int ProgressValue;
			public bool ProgressVisible;
			public bool Cancelled;
			
			public static Status New()
			{
				Status result = new Status();
				result.DocumentCount = -1;
				result.FoundDocumentCount = -1;
				result.ProgressMessage = null;
				result.ProgressSize = -1;
				result.ProgressValue = 0;
				result.ProgressVisible = false;
				result.Cancelled = false;
				return result;
			}
		}
		
		private Status _status;
		
		void IIndexObserver.DocumentCountChanged(int documentCount)
		{
			Lock();
			_status.DocumentCount = documentCount;
			UnLock();
		}
		
		void IIndexObserver.FoundDocumentCountChanged(int documentCount)
		{
			Lock();
			_status.FoundDocumentCount = documentCount;
			UnLock();
		}
		
		void IIndexObserver.OpenProgress(string message, int count)
		{
			Lock();
			_status.ProgressMessage = message;
			_status.ProgressSize = count;
			_status.ProgressValue = 0;
			_status.ProgressVisible = true;
			UnLock();
		}
		
		bool IIndexObserver.ShowProgress(int index)
		{
			bool result;
			Lock();
			_status.ProgressValue = index;
			result = !_status.Cancelled;
			UnLock();
			return result;
		}
		
		void IIndexObserver.ShowMessage(string message)
		{
			Lock();
			_status.ProgressMessage = message;
			UnLock();
		}
		
		void IIndexObserver.CloseProgress()
		{
			Lock();
			_status.ProgressVisible = false;
			UnLock();
		}
		
		bool IIndexObserver.InterruptRequested()
		{
			bool result;
			Lock();
			result = _status.Cancelled;
			UnLock();
			return result;
		}
		#endregion

		#region IProgress interface implementation
		bool System.Threading.IProgress.Aborted 
		{
			get 
			{
				// TODO: same code is in the InterruptRequested() method - unify it
				bool result;
				Lock();
				result = _status.Cancelled;
				UnLock();
				return result;
			}
		}
		
		void System.Threading.IProgress.Open(ulong min, ulong max)
		{
			if (min != 0)
			{
				throw new ArgumentException();
			}
			
			Lock();
			_status.ProgressSize = (int)max;
			_status.ProgressValue = 0;
			_status.ProgressVisible = true;
			UnLock();
		}
		
		void System.Threading.IProgress.Write(string message)
		{
			Lock();
			_status.ProgressMessage = message;
			UnLock();
		}
		
		bool System.Threading.IProgress.Step(ulong value)
		{
			bool result;
			Lock();
			_status.ProgressValue = (int)value;
			result = !_status.Cancelled;
			UnLock();
			return result;
		}
		
		void System.Threading.IProgress.Abort()
		{
			Lock();
			_status.Cancelled = true;
			UnLock();
		}
		
		void System.Threading.IProgress.Close()
		{
			Lock();
			_status.ProgressVisible = false;
			UnLock();
		}
		#endregion
		
		#region IDisposable interface implementation
		public void Close(bool commit)
		{
			if (_updater != null)
			{
				_updater.Dispose();
				_updater = null;
			}
			// TODO: throw new NotImplementedException();
		}
		
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			Close(disposing);
		}

		~Worker()
		{
			Dispose(false);
		}
		#endregion
	}
}
