using System;

namespace Explorer.Engine
{
	public class MountFileSystem: FileSystemBase
	{
		private const string DummyDrive = "A:\\";
		
		public enum FileSystemMode
		{
			Merge,
			CheckPoint
		}
		
		public enum FileSystemAccess
		{
			Read,
			ReadWrite
		}
		
		public interface IMountPoint
		{
			void Open();
			void Add(IFileSystem fileSystem, string root, FileSystemMode mode, FileSystemAccess access);
			void Clear();
		}
		
		private const string InvalidPath = null;
		
		private abstract class MountItem: IRawFileSystem
		{
			private IFileSystem _fileSystem = null;
			public IFileSystem FileSystem
			{
				get
				{
					return _fileSystem;
				}
				set
				{
					if (value != _fileSystem)
					{
						if (_fileSystem != null)
						{
							_fileSystem.Release();
						}
						_fileSystem = value;
						if (_fileSystem != null)
						{
							_fileSystem.AddRef();
						}
					}
				}
			}
			
			private NameMapper _mapper = new NameMapper(DummyDrive, "");
			public string Path
			{
				get
				{
					return _mapper.InnerPrefix;
				}
				set
				{
					_mapper.InnerPrefix = value;
				}
			}
			public bool ReadOnly;

			public void Open()
			{
				if (_fileSystem != null)
				{
					_fileSystem.Open();
				}
			}
			
			public void Close()
			{
				FileSystem = null;
			}
			
			private string OuterToInner(string path)
			{
				return _mapper.OuterToInner(path);
			}
			
			private string InnerToOuter(string path)
			{
				return _mapper.InnerToOuter(path);
			}
			
			private void InnerToOuter(Model.Entry entry)
			{
				_mapper.InnerToOuter(entry);
			}
			
			private void AccessDenied()
			{
				// TODO: wrap into IOException class
				throw new NotSupportedException();
			}
			
			#region IRawFileSystem interface
			/*
			 * This is partial file system interface working with partially valid path names
			 * 
			 */
			public Model.Entries CreateList(string folder)
			{
				Model.Entries result = FileSystem.CreateList(OuterToInner(folder));
				foreach(Model.Entry entry in result)
				{
					InnerToOuter(entry);
				}
				return result;
			}
			
			public Model.Entry GetInfo(string entryName)
			{
				Model.Entry result = FileSystem.GetInfo(OuterToInner(entryName));
				if (result != null)
				{
					InnerToOuter(result);
				}
				return result;
			}
			
			public void CreateDirectory(string folderName)
			{
				if (ReadOnly)
				{
					AccessDenied();
				}
				else
				{
					FileSystem.CreateDirectory(OuterToInner(folderName));
				}
			}
			
			public void DeleteFile(string fileName)
			{
				if (ReadOnly)
				{
					// TODO: check if file exists. If so, throw access denied
					AccessDenied();
				}
				else
				{
					FileSystem.DeleteFile(OuterToInner(fileName));
				}
			}
			
			public void DeleteDirectory(string folderName)
			{
				if (ReadOnly)
				{
					// TODO: check if directory exists. If so, throw access denied
					AccessDenied();
				}
				else
				{
					FileSystem.DeleteDirectory(OuterToInner(folderName));
				}
			}
			
			public System.IO.Stream OpenFile(string fileName, System.IO.FileMode mode, System.IO.FileAccess access, System.IO.FileShare share)
			{
				if (ReadOnly)
				{
					// TODO: use some more elaborate "security access" layer
					if ((access != System.IO.FileAccess.Read) ||
					    (mode == System.IO.FileMode.Create) ||
					    (mode == System.IO.FileMode.CreateNew) ||
					    (mode == System.IO.FileMode.Truncate))
					{
						AccessDenied();
						return null;
					}
					else
					{
						if (GetInfo(fileName) == null)
						{
							// file does not exist
							if (mode == System.IO.FileMode.OpenOrCreate)
							{
								AccessDenied();
								return null;
							}
						}
						return FileSystem.OpenFile(OuterToInner(fileName), mode, access, share);
					}
				}
				else
				{
					return FileSystem.OpenFile(OuterToInner(fileName), mode, access, share);
				}
			}
			
			public string GetBaseFileName(string fileName, out IRawFileSystem container)
			{
				return FileSystem.GetBaseFileName(OuterToInner(fileName), out container);
			}
			#endregion
		}
		
		private class MergeMountItem: MountItem
		{
		}
		
		private class CheckPointMountItem: MountItem
		{
		}
		
		private class MountItems: System.Collections.Generic.List<MountItem>
		{
			public new void Clear()
			{
				foreach(MountItem item in this)
				{
					item.Close();
				}
				base.Clear();
			}
		}
		
		private class MountPoint: IRawFileSystem, IMountPoint
		{
			#region IMountPoint interface
			public void Add(IFileSystem fileSystem, string root, FileSystemMode mode, FileSystemAccess access)
			{
				if (fileSystem == null)
				{
					return;
				}
				if (_items == null)
				{
					_items = new MountItems();
				}
				MountItem item;
				switch (mode)
				{
					case FileSystemMode.Merge:
						item = new MergeMountItem();
						break;
						
					case FileSystemMode.CheckPoint:
						item = new CheckPointMountItem();
						break;
						
					default:
						// TODO: implement
						throw new NotImplementedException();
				}
				item.FileSystem = fileSystem;
				item.ReadOnly = access == FileSystemAccess.Read;
				item.Path = root;
				_items.Add(item);
			}
			
			public void Clear()
			{
				Close();
			}
			#endregion
			
			public string Root;
			private MountItems _items = null;
			
			public virtual void Open()
			{
				if (_items != null)
				{
					foreach(MountItem item in _items)
					{
						item.Open();
					}
				}
			}
			
			public void Close()
			{
				if (_items != null)
				{
					_items.Clear();
					_items = null;
				}
			}
			
			public virtual string OuterToInner(string path)
			{
				bool hit = false;
				string subPath = null;
				if (String.Compare(path + "\\", Root, true) == 0)
				{
					hit = true;
					subPath = "";
				}
				else if (path.Length >= Root.Length)
				{
					if (String.Compare(path.Substring(0, Root.Length), Root, true) == 0)
					{
						hit = true;
						subPath = path.Remove(0, Root.Length);
					}
				}
				if (hit)
				{
					return DummyDrive + subPath;
				}
				else
				{
					return InvalidPath;
				}
			}
			
			public virtual string InnerToOuter(string path)
			{
				return Root + path.Substring(DummyDrive.Length);
			}
			
			private class NameComparer: System.Collections.Generic.Comparer<Model.Entry>
			{
				public override int Compare(Model.Entry x, Model.Entry y)
				{
					return Model.Tools.NameCompare(x, y);
				}
			}
			
			#region IRawFileSystem interface
			public virtual Model.Entries CreateList(string folder)
			{
				NameComparer nameComparer = new NameComparer();
				
				Model.Entries result = new Model.Entries();
				{
					Model.Entry item = new Model.Entry();
					item.Kind = Model.EntryKind.UpFolder;
					// TODO: is it enough?
					item.Path = folder;
					result.Add(item);
				}

				if (_items != null)
				{
					foreach(MountItem item in _items)
					{
						Model.Entries candidate = item.CreateList(folder);
						
						// merge candidate's items to the result list
						foreach(Model.Entry entry in candidate)
						{
							if (entry.Kind != Model.EntryKind.UpFolder)
							{
								// one up folder is enough
								entry.InternalHandle = item;
								
								int index = result.BinarySearch(entry, nameComparer);
								if (index < 0)
								{
									result.Insert(~index, entry);
								}
								else
								{
									int compare = Model.Tools.DateCompare(result[index], entry);
									
									// keep the smaller one
									if (compare <= 0)
									{
										// keep current one
									}
									else
									{
										// TODO: is this discard-stuff needed? does it work?
										Model.Entry discard = result[index];
										
										result[index] = entry;
										
										discard.Dispose();
										discard = null;
									}
								}
							}
						}
					}
				}
				
				return result;
			}

			private Model.Entry GetWritableParentInfo(string entryName)
			{
				return GetWritableInfo(System.IO.Path.GetDirectoryName(entryName));
			}
			
			public virtual Model.Entry GetWritableInfo(string entryName)
			{
				if (_items == null)
				{
					return null;
				}
				Model.Entry result = null;
				foreach(MountItem item in _items)
				{
					if (item.ReadOnly)
					{
						continue;
					}
					Model.Entry candidate = item.GetInfo(entryName);
					if (candidate != null)
					{
						candidate.InternalHandle = item;
						if (result == null)
						{
							result = candidate;
							// TODO: should we search for a folder kind?
							break;
						}
						// TODO: is this correct? what if both folder and file match the entryName?
						else if (Model.Tools.DateCompare(candidate, result) < 0)
						{
							result = candidate;
						}
					}
				}
				return result;
			}
			
			public virtual Model.Entry GetInfo(string entryName)
			{
				if (_items == null)
				{
					return null;
				}
				Model.Entry result = null;
				foreach(MountItem item in _items)
				{
					Model.Entry candidate = item.GetInfo(entryName);
					if (candidate != null)
					{
						candidate.InternalHandle = item;
						if (result == null)
						{
							result = candidate;
						}
						else if (Model.Tools.DateCompare(candidate, result) < 0)
						{
							result = candidate;
						}
					}
				}
				return result;
			}
			
			public virtual void CreateDirectory(string folderName)
			{
				Model.Entry parentEntry = GetWritableParentInfo(folderName);
				if ((parentEntry == null) || (parentEntry.Kind != Model.EntryKind.Folder))
				{
					// TODO: delagate to checkpoint writer
					throw new System.IO.DirectoryNotFoundException();
				}
				else
				{
					// TODO: delegate to checkpoint writer
					((MountItem)parentEntry.InternalHandle).CreateDirectory(folderName);
				}
			}
			
			public virtual void DeleteFile(string fileName)
			{
				Model.Entry entry = GetInfo(fileName);
				if ((entry == null) || (entry.Kind != Model.EntryKind.File))
				{
					throw new System.IO.FileNotFoundException();
				}
				else
				{
					// TODO: delegate to checkpoint writer
					((MountItem)entry.InternalHandle).DeleteFile(fileName);
				}
			}
			
			public virtual void DeleteDirectory(string folderName)
			{
				Model.Entry entry = GetInfo(folderName);
				if ((entry == null) || (entry.Kind != Model.EntryKind.Folder))
				{
					throw new System.IO.DirectoryNotFoundException();
				}
				else
				{
					// TODO: delegate to checkpoint writer
					((MountItem)entry.InternalHandle).DeleteDirectory(folderName);
				}
			}
			
			public virtual System.IO.Stream OpenFile(string fileName, System.IO.FileMode mode, System.IO.FileAccess access, System.IO.FileShare share)
			{
				Model.Entry entry = GetInfo(fileName);
				if ((entry == null) || (entry.Kind != Model.EntryKind.File))
				{
					Model.Entry parentEntry = GetWritableParentInfo(fileName);
					if ((parentEntry == null) || (parentEntry.Kind != Model.EntryKind.Folder))
					{
						// TODO: delegate to writer
						throw new NotImplementedException();
					}
					else
					{
						return ((MountItem)parentEntry.InternalHandle).OpenFile(fileName, mode, access, share);
					}
				}
				else
				{
					// TODO: delegate write access to writer
					return ((MountItem)entry.InternalHandle).OpenFile(fileName, mode, access, share);
				}
			}

			public string GetBaseFileName(string fileName, out IRawFileSystem container)
			{
				Model.Entry entry = GetInfo(fileName);
				if (entry == null)
				{
					container = null;
					return null;
				}
				return ((MountItem)entry.InternalHandle).GetBaseFileName(fileName, out container);
			}
			#endregion
		}
		
		private class RootMountPoint: MountPoint
		{
			public override string OuterToInner(string path)
			{
				return path;
			}
			
			public override string InnerToOuter(string path)
			{
				return path;
			}
			
			private IFileSystem _fileSystem = null;
			public IFileSystem FileSystem
			{
				get
				{
					return _fileSystem;
				}
				set
				{
					if (value != _fileSystem)
					{
						if (_fileSystem != null)
						{
							_fileSystem.Release();
						}
						_fileSystem = value;
						if (_fileSystem != null)
						{
							_fileSystem.AddRef();
						}
					}
				}
			}

			public override void Open()
			{
				if (_fileSystem != null)
				{
					_fileSystem.Open();
				}
			}
			
			#region IRawFileSystem interface
			public override Model.Entries CreateList(string folder)
			{
				return FileSystem.CreateList(folder);
			}
			
			public override Model.Entry GetInfo(string entryName)
			{
				return FileSystem.GetInfo(entryName);
			}
			
			public override void CreateDirectory(string folderName)
			{
				FileSystem.CreateDirectory(folderName);
			}
			
			public override void DeleteFile(string fileName)
			{
				FileSystem.DeleteFile(fileName);
			}
			
			public override void DeleteDirectory(string folderName)
			{
				FileSystem.DeleteDirectory(folderName);
			}
			
			public override System.IO.Stream OpenFile(string fileName, System.IO.FileMode mode, System.IO.FileAccess access, System.IO.FileShare share)
			{
				return FileSystem.OpenFile(fileName, mode, access, share);
			}
			#endregion
		}
		
		private class MountPoints: System.Collections.Generic.List<MountPoint>
		{
			private static string Normalized(string mountPoint)
			{
				if (String.IsNullOrEmpty(mountPoint))
				{
					return "";
				}
				if (!mountPoint.EndsWith("\\"))
				{
					mountPoint = mountPoint + "\\";
				}
				return mountPoint;				
			}
			
			public MountPoint Find(string mountPoint)
			{
				if (String.IsNullOrEmpty(mountPoint))
				{
					return null;
				}
				mountPoint = Normalized(mountPoint);
				foreach(MountPoint item in this)
				{
					if (String.Compare(item.Root, mountPoint, true) == 0)
					{
						return item;
					}
				}
				return null;
			}
			
			private class Comparer: System.Collections.Generic.Comparer<MountPoint>
			{
				private static int CountOfPathComponents(string path)
				{
					int result = 0;
					foreach(char c in path)
					{
						if (c == '\\')
						{
							result++;
						}
					}
					return result;
				}
				
				public override int Compare(MountPoint x, MountPoint y)
				{
					return -(CountOfPathComponents(x.Root) - CountOfPathComponents(y.Root));
				}
			}
			
			public MountPoint Add(string mountPoint)
			{
				MountPoint item = new MountPoint();
				item.Root = Normalized(mountPoint);
				Add(item);
				// sort mount points so that the least-generic come first
				Sort(new Comparer());
				return item;
			}
			
			public new void Clear()
			{
				foreach(MountPoint item in this)
				{
					item.Close();
				}
				base.Clear();
			}
			
			public void Open()
			{
				foreach(MountPoint item in this)
				{
					item.Open();
				}
			}
		}
		
		private class BackboneFileSystem: TreeMappedFileSystem
		{
			public BackboneFileSystem(): base('A')
			{
			}
			
			public string AddFolder(string folder)
			{
				string result = Drive.ToString() + folder.Substring(1);
				RegisterFolder(result);
				return result;
			}
		}
		
		private RootMountPoint _baseMount;
		private BackboneFileSystem _backbone;
		
		public MountFileSystem(IFileSystem fileSystem)
		{
			_baseMount = new RootMountPoint();
			_backbone = new BackboneFileSystem();
			_baseMount.FileSystem = fileSystem;
		}

		public MountFileSystem(char drive)
		{
			_baseMount = new RootMountPoint();
			_backbone = new BackboneFileSystem();
			_backbone.AddRef();
			_baseMount.FileSystem = _backbone;
		}
		
		public override void Open()
		{
			_backbone.Open();
			_baseMount.Open();
			_mounts.Open();
		}
		
		protected override void Close()
		{
			base.Close();
			_mounts.Clear();
			_baseMount.FileSystem = null;
			_backbone.Release();
			// TODO: is everything really closed?
		}
		
		private MountPoints _mounts = new MountPoints();

		#region Configuration
		public IMountPoint this[string mountPoint]
		{
			get
			{
				return _mounts.Find(mountPoint);
			}
		}
		
		public IMountPoint Add(string mountPoint, FileSystemAccess defaultAccessMode)
		{
			MountPoint result = _mounts.Find(mountPoint);
			if (result == null)
			{
				result = _mounts.Add(mountPoint);
				_backbone.AddFolder(result.Root);
				// by default mount points work in see-through mode
				result.Add(_baseMount.FileSystem, result.Root, FileSystemMode.Merge, defaultAccessMode);
				// TODO: somehow add _backbone to the _baseMount.FileSystem (for the case
				// when constructor MountFileSystem(IFileSystem) is used
			}
			return result;
		}
		#endregion
		
		private string OuterToInner(string path, out MountPoint mount)
		{
			mount = _baseMount;
			if (String.IsNullOrEmpty(path))
			{
				return path;
			}
			foreach(MountPoint item in _mounts)
			{
				string result = item.OuterToInner(path);
				if (result != InvalidPath)
				{
					mount = item;
					return result;
				}
			}
			return path;
		}
		
		private string InnerToOuter(string path, MountPoint mount)
		{
			return mount.InnerToOuter(path);
		}
		
		private void InnerToOuter(Model.Entry entry, MountPoint mount)
		{
			entry.Path = mount.InnerToOuter(entry.Path);
		}
		
		public override Model.Entries CreateList(string folder)
		{
			MountPoint mount;
			folder = OuterToInner(folder, out mount);
			Model.Entries result = mount.CreateList(folder);
			foreach(Model.Entry entry in result)
			{
				InnerToOuter(entry, mount);
			}
			return result;
		}
		
		public override void CreateDirectory(string folderName)
		{
			MountPoint mount;
			folderName = OuterToInner(folderName, out mount);
			mount.CreateDirectory(folderName);
		}
		
		public override void DeleteDirectory(string folderName)
		{
			MountPoint mount;
			folderName = OuterToInner(folderName, out mount);
			mount.DeleteDirectory(folderName);
		}
		
		public override void DeleteFile(string fileName)
		{
			MountPoint mount;
			fileName = OuterToInner(fileName, out mount);
			mount.DeleteFile(fileName);
		}
		
		public override Model.Entry GetInfo(string entryName)
		{
			MountPoint mount;
			entryName = OuterToInner(entryName, out mount);
			Model.Entry entry = mount.GetInfo(entryName);
			if (entry != null)
			{
				InnerToOuter(entry, mount);
			}
			return entry;
		}
		
		public override System.IO.Stream OpenFile(string fileName, System.IO.FileMode mode, System.IO.FileAccess access, System.IO.FileShare share)
		{
			MountPoint mount;
			fileName = OuterToInner(fileName, out mount);
			return mount.OpenFile(fileName, mode, access, share);
		}
		
		public override string GetBaseFileName(string fileName, out IRawFileSystem container)
		{
			MountPoint mount;
			fileName = OuterToInner(fileName, out mount);
			return mount.GetBaseFileName(fileName, out container);
		}
	}
}
