using System;

namespace Explorer.Engine
{
	public class ZipFileSystem: TreeMappedFileSystem
	{
		ICSharpCode.SharpZipLib.Zip.ZipFile _base = null;

		public ZipFileSystem(char drive, string fileName): base(drive)
		{
			bool fileNameValid = !String.IsNullOrEmpty(fileName);
			if (fileNameValid)
			{
				fileName = System.IO.Path.GetFullPath(fileName);
				fileNameValid = System.IO.File.Exists(fileName);
			}
			
			if (fileNameValid)
			{
				try
				{
					_base = new ICSharpCode.SharpZipLib.Zip.ZipFile(System.IO.Path.GetFullPath(fileName), ICSharpCode.SharpZipLib.Zip.ZipFile.OpenMode.Safe);
				}
				catch
				{
					_base = ICSharpCode.SharpZipLib.Zip.ZipFile.Create(new System.IO.MemoryStream());
					_base.IsStreamOwner = true;
				}
			}
			else
			{
				_base = ICSharpCode.SharpZipLib.Zip.ZipFile.Create(new System.IO.MemoryStream());
				_base.IsStreamOwner = true;
			}
		}

		private static string CleanZipName(string name)
		{
			if (name.StartsWith("/"))
			{
				return name.Substring(1);
			}
			else
			{
				return name;
			}
		}
		
		protected override void BuildIndexes()
		{
			int entryCount = Convert.ToInt32(_base.Count);
			int fileCount = entryCount;
			
			// register folders and count files
			for(int i = 0; i < entryCount; i++)
			{
				string fullName = CleanZipName(_base[i].Name);
				
				if (String.IsNullOrEmpty(fullName))
				{
					// root entry - considered invalid
					fileCount--;
					continue;
				}
				
				if (fullName.EndsWith("/"))
				{
					fileCount--;
					Folders.Add(fullName.Substring(0, fullName.Length - 1));
				}
				else
				{
					int j = fullName.LastIndexOf("/");
					if (j >= 0)
					{
						Folders.Add(fullName.Substring(0, j));
					}
				}
			}
			
			// register files
			InitializeFiles(fileCount, this.EntryIndexToFileName);
			for(int i = 0; i < entryCount; i++)
			{
				string fullName = CleanZipName(_base[i].Name);
				if (!fullName.EndsWith("/"))
				{
					int j = fullName.LastIndexOf("/");
					if (j >= 0)
					{
						Files.Add(Folders.Find(fullName.Substring(0, j)), i);
					}
					else
					{
						Files.Add(Folders, i);
					}
				}
			}
		}
		
		private string EntryIndexToFileName(int index)
		{
			ICSharpCode.SharpZipLib.Zip.ZipEntry entry = _base[index];
			string result = CleanZipName(entry.Name);
			int i = result.LastIndexOf("/");
			if (i >= 0)
			{
				return result.Substring(i + 1);
			}
			else
			{
				return result;
			}
		}
		
		protected override void Close()
		{
			base.Close();
			if (_base != null)
			{
				_base.Close();
				_base = null;
			}
		}
		
		private string ZipToInner(string path)
		{
			return _innerRoot + path;
		}
		
		private string InnerToZip(string path)
		{
			return path.Substring(_innerRoot.Length);
		}
		
		protected override void AddFilesToList(NameTree rootTree, Explorer.Model.Entries result)
		{
			foreach(int index in Files.GetEnumerator(rootTree))
			{
				ICSharpCode.SharpZipLib.Zip.ZipEntry entry = _base[index];
				result.Add(GetInfo(entry));
			}
		}
		
		private Model.Entry GetInfo(ICSharpCode.SharpZipLib.Zip.ZipEntry entry)
		{
			if (entry == null)
			{
				return null;
			}
			
			Model.Entry item = new Model.Entry();
			
			if (entry.IsDirectory)
			{
				item.Kind = Model.EntryKind.Folder;
			}
			else
			{
				item.Kind = Model.EntryKind.File;
				item.Size = entry.Size;
				item.LastWriteTime = entry.DateTime;
			}
			
			string name = InnerToOuter(ZipToInner(CleanZipName(entry.Name)));
			if (name.EndsWith("\\"))
			{
				name = name.Substring(0, name.Length - 1);
			}
			item.Name = System.IO.Path.GetFileName(name);
			item.Path = GetPath(name);
			
			item.InternalHandle = entry;
			
			return item;
		}

		protected override Explorer.Model.Entry GetFileInfo(string innerName, int fileIndex)
		{
			return GetInfo(_base[fileIndex]);
		}
		
		private class Rewinder
		{
			private WeakReference _fileSystem;
			private ICSharpCode.SharpZipLib.Zip.ZipEntry _entry;
			public Rewinder(ICSharpCode.SharpZipLib.Zip.ZipFile fileSystem, ICSharpCode.SharpZipLib.Zip.ZipEntry entry)
			{
				_entry = entry;
				_fileSystem = new WeakReference(fileSystem, false);
			}
			
			public System.IO.Stream CreateStream()
			{
				ICSharpCode.SharpZipLib.Zip.ZipFile fileSystem = (ICSharpCode.SharpZipLib.Zip.ZipFile)_fileSystem.Target;
				if (fileSystem == null)
				{
					throw new ObjectDisposedException("ICSharpCode.SharpZipLib.Zip.ZipFile");
				}
				return fileSystem.GetInputStream(_entry);
			}
		}
		
		protected override System.IO.Stream OpenFile(Model.Entry info, System.IO.FileMode mode, System.IO.FileAccess access, System.IO.FileShare share)
		{
			ICSharpCode.SharpZipLib.Zip.ZipEntry entry = (ICSharpCode.SharpZipLib.Zip.ZipEntry)info.InternalHandle;
			
			System.IO.Stream result = _base.GetInputStream(entry);
			if (result != null)
			{
				// TODO: we might skip the seekable wrapper if returned stream supports seeking, but there is off-by-one
				// bug in the PartialInputStream class that throws "Cannot seek past end"
				// if (!result.CanSeek)
				{
					return new System.IO.SeekableInputStream(result, (new Rewinder(_base, entry)).CreateStream, entry.Size);
				}
			}
			return result;
		}
		
		public override void GetContainerInfo(string path, out string container, out string containerPath)
		{
			container = _base.Name;
			containerPath = InnerToZip(OuterToInner(path));
		}
	}
}
