/* KEYWORDS: VERSION, TODO
 * VERSION: 1.3
 * 1.3 / 2006-05-20
 * - new delegate TestCaseWithStrings for test function(string[]) -> bool
 * - new overloaded TestBench.Run() for running TestCaseWithStrings
 */
using System;
using System.Diagnostics;

namespace NUnit
{
	/// <summary>
	/// This function represents single test. TestCase fails either by
	/// returning false or by failing through one of <see cref="@Debug.Assert">Debug.Assert()</see> methods
	/// </summary>
	/// <returns>
	/// true if test succeeded
	/// false if test failed
	/// </returns>
	public delegate bool TestCase();
	public delegate bool TestCaseWithStrings(string[] arguments);

	/// <summary>
	/// This class represents test runner. Individual tests are run through the <see cref="@Run">Run</see> method.
	/// After all tests had been run, <see cref="@Failed">Failed</see> property indicates
	/// either success or failure.
	/// </summary>
	public class TestBench
	{
		private class TestListener: TextWriterTraceListener 
		{
			private bool pFailed = false;

			public TestListener(string logFileName): base(new System.IO.StreamWriter(logFileName))
			{
			}

			public override void Fail(string message, string detailMessage)
			{
				Failed = true;
				base.Fail(message, detailMessage);
			}

			public override void Fail(string message)
			{
				Failed = true;
				base.Fail(message);
			}

			public bool Failed
			{
				get
				{
					return pFailed;
				}
				set
				{
					pFailed = value;
				}
			}
		}

		private TestListener guard;

		public bool Run(string testName, TestCase testMethod)
		{
			Debug.WriteLine(String.Format("TestBench: running {0} test", testName));
			Debug.Indent();

			string errorMessage;
			bool result;
			try
			{
				errorMessage = "";
				guard.Failed = false;
				result = testMethod();
				if(guard.Failed)
				{
					result = false;
				}
			}
			catch(Exception e)
			{
				errorMessage = e.ToString();
				result = false;
			}
			Debug.Unindent();
			Debug.WriteLine("");
			if(result)
			{
				Debug.WriteLine(String.Format("TestBench: {0} test succeeded", testName));
			}
			else
			{
				if(errorMessage != "")
				{
					Debug.WriteLine(String.Format("TestBench: {0} test failed with {1}", testName, errorMessage));
				}
				else
				{
					Debug.WriteLine(String.Format("TestBench: {0} test failed", testName));
				}
			}

			if(!result)
			{
				pFailed = true;
			}
			Debug.Flush();
			return result;
		}

		public bool Run(string testName, TestCaseWithStrings testMethod, string[] arguments)
		{
			Debug.Write(String.Format("TestBench: running {0} test with ", testName));
			{
				for(int i = 0; i < arguments.Length; i++)
				{
					if(i > 0)
					{
						Debug.Write(", ");
					}
					Debug.Write(String.Format("\"{0}\"", arguments[i]));
				}
			}
			Debug.WriteLine("");
			Debug.Indent();

			string errorMessage;
			bool result;
			try
			{
				errorMessage = "";
				guard.Failed = false;
				result = testMethod(arguments);
				if(guard.Failed)
				{
					result = false;
				}
			}
			catch(Exception e)
			{
				errorMessage = e.ToString();
				result = false;
			}
			Debug.Unindent();
			Debug.WriteLine("");
			if(result)
			{
				Debug.WriteLine(String.Format("TestBench: {0} test succeeded", testName));
			}
			else
			{
				if(errorMessage != "")
				{
					Debug.WriteLine(String.Format("TestBench: {0} test failed with {1}", testName, errorMessage));
				}
				else
				{
					Debug.WriteLine(String.Format("TestBench: {0} test failed", testName));
				}
			}

			if(!result)
			{
				pFailed = true;
			}
			Debug.Flush();
			return result;
		}

		public TestBench(string logFileName)
		{
			bool debugMode = false;
			try
			{
				Debug.Assert(false);
			}
			catch(Exception)
			{
				debugMode = true;
			}
			if (!debugMode)
			{
				throw new Exception("Debug support missing");
			}
			Debug.Listeners.Clear();
			Debug.AutoFlush = true;
			guard = new TestListener(logFileName);
			Debug.Listeners.Add(guard);
		}

		private bool pFailed = false;

		public bool Failed
		{
			get
			{
				return pFailed;
			}
		}
	}
}
