using System;

namespace Prolog.Memory
{
	public enum OperatorType
	{
		xfx,
		xfy,
		yfx,
		yfy,
		fx,
		fy,
		xf,
		yf
	}
	
	public class Operator
	{
		private int _precedence;
		private OperatorType _type;
		private string _name;
		
		public int Precedence
		{
			get
			{
				return _precedence;
			}
		}
		
		public OperatorType Type
		{
			get
			{
				return _type;
			}
		}
		
		public string Name
		{
			get
			{
				return _name;
			}
		}
		
		public Operator(int precedence, OperatorType type, string name)
		{
			this._precedence = precedence;
			this._type = type;
			this._name = name;
		}
		
		public bool IsUnary
		{
			get
			{
				switch (_type)
				{
					case OperatorType.fx:
					case OperatorType.fy:
					case OperatorType.xf:
					case OperatorType.yf:
						return true;
						
					default:
						return false;
				}
			}
		}
		
		public bool IsBinary
		{
			get
			{
				switch (_type)
				{
					case OperatorType.xfx:
					case OperatorType.xfy:
					case OperatorType.yfx:
					case OperatorType.yfy:
						return true;
						
					default:
						return false;
				}
				
			}
		}
	}
	
	public class Operators
	{
		private class PrecedenceComparer: System.Collections.Generic.Comparer<Operator>
		{
			public override int Compare(Operator x, Operator y)
			{
				return x.Precedence - y.Precedence;
			}
		}
		
		private System.Collections.Generic.List<Operator> _items = new System.Collections.Generic.List<Operator>();
		
		public int Count
		{
			get
			{
				return _items.Count;
			}
		}
		
		public Operator this[int index]
		{
			get
			{
				return _items[index];
			}
		}
		
		public void Add(Operator item)
		{
			// Operators can have at most one infix declaration and one declaration of either pre- or
			// post-fix in force at any time. Subsequent operator declarations supersede earlier ones.
			if (item.IsBinary)
			{
				for(int i = 0; i < Count; i++)
				{
					Operator old = this[i];
					if ((old.Name == item.Name) && old.IsBinary)
					{
						_items.RemoveAt(i);
						break;
					}
				}
			}
			else if (item.IsUnary)
			{
				for(int i = 0; i < Count; i++)
				{
					Operator old = this[i];
					if ((old.Name == item.Name) && old.IsUnary)
					{
						_items.RemoveAt(i);
						break;
					}
				}
			}
			
			_items.Add(item);
			_items.Sort(new PrecedenceComparer());
		}
		
		public void Add(int precedence, OperatorType type, string name)
		{
			Add(new Operator(precedence, type, name));
		}
		
		public void Add(int precedence, OperatorType type, string[] names)
		{
			foreach(string name in names)
			{
				Add(new Operator(precedence, type, name));
			}
		}
		
		public bool IsBinary(string name, out Operator result)
		{
			for(int i = 0; i < Count; i++)
			{
				result = this[i];
				if ((result.Name == name) && result.IsBinary)
				{
					return true;
				}
			}
			result = null;
			return false;
		}
		
		public bool IsBinary(string name)
		{
			Operator dummy;
			return IsBinary(name, out dummy);
		}
		
		public bool IsUnary(string name, out Operator result)
		{
			for(int i = 0; i < Count; i++)
			{
				result = this[i];
				if ((result.Name == name) && result.IsUnary)
				{
					return true;
				}
			}
			result = null;
			return false;
		}
	}
}
