/* KEYWORDS: VERSION, TODO
 * VERSION: 1.3
 * 1.3 / 2006-12-26
 * - fixed AtomCollection.Save()
 * 1.2 / 2006-05-20
 * - multiline values are internally separated with '|' character.
 *   see EncodedValue(), DecodedValue()
 */
using System;
using System.Collections;
using System.Data;
using System.Diagnostics;
using System.Reflection;

namespace Quest.Atomic
{
	/// <summary>
	/// Description of AtomCollection.
	/// </summary>
	[Serializable]
	public class AtomCollection: ArrayList
	{
		private Type pFactory = null;
		public const int NullID = 0;
		private Binder pBinder = null;
		private int pCodePage = Csv.DefaultCodePage;

		public AtomCollection(Type aFactory)
		{
			pFactory = aFactory;
		}

		public AtomCollection(Type aFactory, Binder aBinder)
		{
			pFactory = aFactory;
			pBinder = aBinder;
			if(pBinder != null) pBinder.Add(this);
		}

		public int CodePage
		{
			get
			{
				return pCodePage;
			}
			set
			{
				pCodePage = value;
			}
		}

		public Type Factory
		{
			get
			{
				return pFactory;
			}
		}

		public Atom New(int ID)
		{
			// try to bind to existing atom
			Atom result = this[ID];
			if (result != null)
			{
				return result;
			}

			// we need to create new one
			ConstructorInfo constructor = pFactory.GetConstructor(new Type[0]);
			Debug.Assert(constructor != null);
			result = (Atom)constructor.Invoke(new object[0]);
			result.Owner = this;

			int newID;
			if(ID == NullID)
			{
				newID = Count + 1;
			}
			else
			{
				newID = ID;
			}
			while(newID > Count)
			{
				Atom temporary = (Atom)constructor.Invoke(new object[0]);
				temporary.ID = Count + 1;
				temporary.Owner = this;
				Add(temporary);
			}
			result.ID = newID;
			base[newID - 1] = result;
			return result;
		}

		public new Atom this[int ID]
		{
			get
			{
				if(ID == NullID)
				{
					return null;
				}
				else
				{
					return (ID > Count) ? null : (Atom)base[ID - 1];
				}
			}
		}

		public Atom New()
		{
			return New(NullID);
		}

		internal void Delete(Atom item)
		{
			if (item.ID == NullID) return;

			if (item.Owner == this)
			{
				// unlink all external pointers pointing to the item
				foreach(AtomCollection container in pBinder)
				{
					if (container != this)
					{
						container.Delete(item);
					}
				}

				// shift items down and change their IDs accordingly			
				int index = (item.ID - 1) + 1;
				while (index < Count)
				{
					Atom moving = (Atom)base[index];
					if (moving.ID != NullID)
					{
						moving.ID = moving.ID - 1;
					}
					base[index - 1] = moving;
					index++;
				}

				base.RemoveAt(Count - 1);

				item.ID = NullID;
				item.Owner = null;
			}
			else
			{
				// unlink all owned pointers pointing to the item
				IList pointers = Select(item);
				if (pointers.Count > 0)
				{
					FieldInfo[] fields = AtomInfo.GetFields(pFactory, AtomInfo.FieldFlags.Linkable);
					foreach(Atom pointer in pointers)
					{
						foreach(FieldInfo info in fields)
						{
							if((Atom)info.GetValue(pointer) == item)
							{
								info.SetValue(pointer, null);
							}
						}
					}
				}
			}
		}

		public string Name
		{
			get
			{
				return pFactory.Name;
			}
		}

		public static string CollectionName(string elementName)
		{
			// construct english plural of elementName
			string collection = elementName;
			if (!collection.EndsWith("s"))
			{
				if (collection.EndsWith("y"))
				{
					collection = collection.Substring(0, collection.Length - 1) + "ies";
				}
				else
				{
					collection = collection + "s";
				}
			}
			else
			{
				collection = collection + "es";
			}
			return collection;
		}

		public string DisplayName
		{
			get
			{
				return CollectionName(Name);
			}
		}

		private FieldInfo[] GetFields(DataColumnCollection columns)
		{
			string[] names = null;
			if(columns != null)
			{
				names = new string[columns.Count];
				for(int i = 0; i < columns.Count; i++)
				{
					names[i] = columns[i].ColumnName;
				}
			}
			return AtomInfo.GetFields(pFactory, AtomInfo.FieldFlags.All, names);
		}

		private FieldInfo[] GetFields()
		{
			return GetFields(null);
		}

		#region Searching
		internal IList InternalSelect(Type aFactory, Atom filter)
		{
			return pBinder.Bind(aFactory).Select(filter);
		}

		internal IList InternalSelect(Type aFactory, Atom filter, FieldInfo field)
		{
			return pBinder.Bind(aFactory).Select(filter, field);
		}

		internal IList InternalSelectCandidates(Atom filter, FieldInfo field)
		{
			return pBinder.Bind(field.FieldType).SelectCandidates(filter);
		}

		public IList Select(Atom filter)
		{
			ArrayList result = new ArrayList();
			FieldInfo[] fields = AtomInfo.GetFields(pFactory, AtomInfo.FieldFlags.Linkable);
			if (fields.Length > 0)
			{
				foreach(Atom candidate in this)
				{
					foreach(FieldInfo info in fields)
					{
						if((Atom)info.GetValue(candidate) == filter)
						{
							result.Add(candidate);
							break;
						}
					}
				}
			}
			return result;
		}

		public IList SelectCandidates(Atom filter)
		{
			ArrayList result = new ArrayList();
			foreach(Atom candidate in this)
			{
				if (candidate != filter)
				{
					result.Add(candidate);
				}
			}
			return result;
		}

		public IList Select(Atom filter, FieldInfo field)
		{
			ArrayList result = new ArrayList();
			foreach(Atom candidate in this)
			{
				if((Atom)field.GetValue(candidate) == filter)
				{
					result.Add(candidate);
				}
			}
			return result;
		}

		public IList Select(Atom filter, string fieldName)
		{
			return Select(filter, Factory.GetField(fieldName));
		}
		#endregion

		#region Internal coding of multi-lines
		private const string defaultNewLineSeparator = "\n";

		private static object DecodedValue(object value, string newLineSeparator)
		{
			if(value.GetType() == typeof(System.String))
			{
				return ((string)value).Replace("|", newLineSeparator);
			}
			else
			{
				return value;
			}
		}

		private static object EncodedValue(object value, string newLineSeparator)
		{
			if(value.GetType() == typeof(System.String))
			{
				return ((string)value).Replace(newLineSeparator, "|").Replace("\r", "");
			}
			else
			{
				return value;
			}
		}
		#endregion

		#region Input/Output
		public void Save()
		{
			Save("");
		}

		public DataTable GetDataTable()
		{
			return GetDataTable(false, defaultNewLineSeparator);
		}

		public DataTable GetDataTable(bool dbExact, string newLineSeparator)
		{
			FieldInfo[] fields = GetFields();
			DataTable table = new DataTable();

			for(int i = 0; i < fields.Length; i++)
			{
				FieldInfo info = fields[i];
				DataColumn c = new DataColumn();
				c.ColumnName = info.Name;
				c.DataType = info.FieldType;

				if (dbExact)
				{
					if (c.DataType == typeof(System.String))
					{
						c.MaxLength = 1;
					}
				}
				table.Columns.Add(c);
			}

			foreach(Atom a in this)
			{
				DataRow r = table.NewRow();
				for(int i = 0; i < fields.Length; i++)
				{
					FieldInfo info = fields[i];
					object value = info.GetValue(a);
					if(value != null)
					{
						if (dbExact)
						{
							DataColumn c = table.Columns[i];
							if (c.DataType == typeof(System.String))
							{
								string s = (string)DecodedValue(value, newLineSeparator);
								int len = s.Length;

								if (len > c.MaxLength)
								{
									c.MaxLength = len;
								}
								r[info.Name] = s;
								continue;
							}
						}

						r[info.Name] = DecodedValue(value, newLineSeparator);
					}
				}
				table.Rows.Add(r);
			}
			table.AcceptChanges();

			return table;
		}

		public void Save(string prefix)
		{
			DataTable table = GetDataTable();
			try
			{
				System.Data.Csv.SaveTable(table, FileName(prefix)/* TODO:, CodePage*/);
			}
			finally
			{
				table.Dispose();
			}
		}

		private string FileName(string prefix)
		{
			string p = prefix;
			if(p != "")
			{
				if((!p.EndsWith(".")) && (!p.EndsWith("\\")))
				{
					p = p + ".";
				}
			}
			return p + Name + ".csv";
		}

		public void Load()
		{
			Load("");
		}

		public virtual void Loaded()
		{
		}

		public void LoadDataTable(DataTable table, string newLineSeparator)
		{
			Clear();
			try
			{
				if (table == null) return;

				FieldInfo[] fields = GetFields(table.Columns);
				DataColumn columnID = null;
				{
					int i = table.Columns.IndexOf("ID");
					if(i >= 0)
					{
						columnID = table.Columns[i];
					}
				}
				foreach(DataRow r in table.Rows)
				{
					int newID = 0;
					if(columnID != null)
					{
						if(!r.IsNull(columnID))
						{
							object value = r[columnID];
							newID = System.Convert.ToInt32(value.ToString()); // TODO: ?
						}
					}
					Atom a = New(newID);
					for(int i = 0; i < fields.Length; i++)
					{
						FieldInfo info = fields[i];
						int ci = table.Columns.IndexOf(info.Name);
						if(!r.IsNull(ci))
						{
							object value = r[ci];
							if(info.FieldType.IsSubclassOf(typeof(Atom)))
							{
								if(pBinder != null)
								{
									pBinder.Bind(a, info, System.Convert.ToInt32(value));
								}
							}
							else
							{
								info.SetValue(a, EncodedValue(System.Convert.ChangeType(value, info.FieldType), newLineSeparator));
							}
						}
					}
				}
			}
			finally
			{
				Loaded();
			}
		}

		public void Load(string prefix)
		{
			Clear();
			DataTable table = System.Data.Csv.LoadTable(FileName(prefix), CodePage);
			try
			{
				LoadDataTable(table, defaultNewLineSeparator);
			}
			finally
			{
				if (table != null)
				{
					table.Dispose();
				}
			}
		}
		#endregion
	}
}
