using System;
using System.Drawing;
using System.Collections;
using Quest.Atomic;
using Quest.Graph2D;

namespace Quest.Atomic.Graph2D
{
	public class FieldTemplate : ICloneable
	{
		public EdgeLayout EdgeLayout = null;
		public string DisplayLabel = null;

		public object Clone()
		{
			FieldTemplate result = new FieldTemplate();
			if (EdgeLayout != null)
			{
				result.EdgeLayout = (EdgeLayout)EdgeLayout.Clone();
			}
			result.DisplayLabel = DisplayLabel;
			return result;
		}
	}

	internal class FieldTemplates : System.Collections.Generic.Dictionary<string, FieldTemplate>, ICloneable
	{
		public object Clone()
		{
			FieldTemplates result = new FieldTemplates();
			foreach (System.Collections.Generic.KeyValuePair<string, FieldTemplate> entry in this)
			{
				result.Add(entry.Key, (FieldTemplate)entry.Value.Clone());
			}
			return result;
		}
	}

	public class NodeTemplate: ICloneable
	{
		internal GraphBuilder Builder;
		
		public NodeLayout NodeLayout = new NodeLayout();
		public EdgeLayout EdgeLayout = new EdgeLayout();
		public string NameField;
		public bool Visible = true;

		private System.Reflection.FieldInfo[] pNameFieldInfo = null;
		private System.Reflection.FieldInfo[] pEdgeFieldInfo = null;

		internal System.Reflection.FieldInfo[] EdgeFields
		{
			get
			{
				return (pEdgeFieldInfo == null) ? new System.Reflection.FieldInfo[0] : pEdgeFieldInfo;
			}
		}

		internal void SetFactory(Type factory)
		{
			if(!String.IsNullOrEmpty(NameField))
			{
				string[] names = NameField.Split(new char[]{'|', ',', ';'});
				pNameFieldInfo = new System.Reflection.FieldInfo[names.Length];
				for(int i = 0; i < names.Length; i++)
				{
					pNameFieldInfo[i] = factory.GetField(names[i]);
				}
			}

			pEdgeFieldInfo = AtomInfo.GetFields(factory, AtomInfo.FieldFlags.Linkable);
			if(pEdgeFieldInfo.Length == 0)
			{
				pEdgeFieldInfo = null;
			}
		}

		public virtual string NodeName(Quest.Atomic.Atom item)
		{
			if(pNameFieldInfo == null) return "";

			string result = "";
			for(int i = 0; i < pNameFieldInfo.Length; i++)
			{
				object rawValue = pNameFieldInfo[i].GetValue(item);
				string value = "";
				if (rawValue != null)
				{
					if (rawValue is Quest.Atomic.Atom)
					{
						Quest.Atomic.Atom link = (Quest.Atomic.Atom)rawValue;
						NodeTemplate linkTemplate = Builder[link.GetType()];
						if (linkTemplate != null)
						{
							value = linkTemplate.NodeName(link);
						}
						else
						{
							// TODO: use some atom's pretty name
							value = System.Convert.ToString(rawValue);
						}
					}
					else
					{
						value = System.Convert.ToString(rawValue);
					}
				}
				if(i == 0)
				{
					result = value;
				}
				else
				{
					if(value != "")
					{
						result += "|+" + value;
					}
				}
			}
			return result;
		}

		public IList NodeEdges(Quest.Atomic.Atom item)
		{
			ArrayList result = new ArrayList();
			if(pEdgeFieldInfo != null)
			{
				foreach(System.Reflection.FieldInfo info in pEdgeFieldInfo)
				{
					Atom value = (Atom)info.GetValue(item);
					if(value != null)
					{
						NodeEdge edge = new NodeEdge();
						edge.Field = info;
						edge.Value = value;
						result.Add(edge);
					}
				}
			}
			return result;
		}

		private FieldTemplates pFieldTemplates = null;

		private FieldTemplate GetFieldTemplate(System.Reflection.FieldInfo field, bool canCreate)
		{
			return GetFieldTemplate(field.Name, canCreate);
		}

		private FieldTemplate GetFieldTemplate(string field, bool canCreate)
		{
			if (pFieldTemplates == null)
			{
				if (canCreate)
				{
					pFieldTemplates = new FieldTemplates();
				}
				else
				{
					return null;
				}
			}
			FieldTemplate result = null;
			if (!pFieldTemplates.TryGetValue(field, out result))
			{
				if (canCreate)
				{
					result = new FieldTemplate();
					if (pEdgeFieldInfo != null)
					{
						foreach (System.Reflection.FieldInfo info in pEdgeFieldInfo)
						{
							if (info.Name == field)
							{
								result.EdgeLayout = (EdgeLayout)EdgeLayout.Clone();
								break;
							}
						}
					}
					pFieldTemplates.Add(field, result);
				}
			}
			return result;
		}

		public FieldTemplate this[System.Reflection.FieldInfo field]
		{
			get
			{
				return GetFieldTemplate(field, true);
			}
		}

		public FieldTemplate this[string field]
		{
			get
			{
				return GetFieldTemplate(field, true);
			}
		}

		public EdgeLayout GetLayout(System.Reflection.FieldInfo field)
		{
			FieldTemplate template = GetFieldTemplate(field, false);
			if (template == null)
			{
				return EdgeLayout;
			}
			else
			{
				return template.EdgeLayout;
			}
		}

		internal string GetLabel(System.Reflection.FieldInfo field)
		{
			FieldTemplate template = GetFieldTemplate(field, false);
			if (template == null)
			{
				return field.Name;
			}
			else if (template.DisplayLabel == null)
			{
				return field.Name;
			}
			else
			{
				return template.DisplayLabel;
			}
		}

		public string GetDisplayLabel(System.Reflection.FieldInfo field)
		{
			string result = GetLabel(field);
			if (String.IsNullOrEmpty(result))
			{
				return field.Name;
			}
			else
			{
				return result;
			}
		}

		public NodeTemplate()
		{
		}

		public NodeTemplate(string fieldName)
		{
			NameField = fieldName;
		}

		public object Clone()
		{
			NodeTemplate result = new NodeTemplate();
			result.Builder = Builder;
			result.NameField = NameField;
			result.Visible = Visible;
			if(NodeLayout != null)
			{
				result.NodeLayout = (NodeLayout)NodeLayout.Clone();
			}
			if(EdgeLayout != null)
			{
				result.EdgeLayout = (EdgeLayout)EdgeLayout.Clone();
			}
			if (pFieldTemplates != null)
			{
				result.pFieldTemplates = (FieldTemplates)pFieldTemplates.Clone();
			}
			return result;
		}
	}

	public struct NodeEdge
	{
		public System.Reflection.FieldInfo Field;
		public Atom Value;
	}

	public class GraphBuilder
	{
		#region Delayed update
		internal class UpdateRequest
		{
			public Node Node;
			public NodeEdge Link;
			public EdgeLayout Layout;
			public string Label;
		}

		private int updateLevel = 0;
		private ArrayList updateRequests = null;

		public void BeginUpdate()
		{
			if(updateLevel == 0)
			{
				updateRequests = new ArrayList();
			}
			updateLevel++;
		}

		public void EndUpdate()
		{
			updateLevel--;
			if(updateLevel == 0)
			{
				Graph.BeginUpdate();
				try
				{
					foreach(UpdateRequest request in updateRequests)
					{
						Edge edge = Graph.New(request.Node, request.Link.Field.Name, request.Link.Value);
						if(edge != null)
						{
							edge.Label = request.Label;
							edge.Layout = request.Layout;
						}
					}
				}
				finally
				{
					updateRequests.Clear();
					updateRequests = null;
					Graph.EndUpdate();
				}
			}
		}

		private void AddUpdateRequest(Node node, NodeEdge link, EdgeLayout layout, string label)
		{
			UpdateRequest request = new UpdateRequest();
			request.Node = node;
			request.Link = link;
			request.Layout = layout;
			request.Label = label;
			updateRequests.Add(request);
		}
		#endregion

		public GraphBuilder()
		{
		}

		public Graph Graph;

		private Hashtable NodeTemplates = new Hashtable();

		protected NodeTemplate AddTemplate(Type factory, string fieldName, Color backgroundColor)
		{
			NodeTemplate template = new NodeTemplate(fieldName);
			template.Builder = this;
			template.NodeLayout = new NodeLayout(backgroundColor, Color.Black, Color.Black);
			this[factory] = template;
			return this[factory];
		}

		protected NodeTemplate AddTemplate(Type factory, string fieldName)
		{
			return AddTemplate(factory, fieldName, Color.White);
		}

		protected NodeTemplate AddTemplate(Type factory, Color backgroundColor)
		{
			return AddTemplate(factory, null, backgroundColor);
		}

		protected NodeTemplate AddTemplate(Type factory)
		{
			return AddTemplate(factory, null, Color.White);
		}

		#region Style helpers
		public enum ArrowStyle
		{
			None,
			Normal
		};

		protected void ModifyLayout(EdgeLayout layout, Color lineColor)
		{
			layout.LineStyle = (Pen)layout.LineStyle.Clone();
			layout.LineStyle.Color = lineColor;
		}

		protected void ModifyLayout(EdgeLayout layout, ArrowStyle arrow)
		{
			layout.LineStyle = (Pen)layout.LineStyle.Clone();
			switch(arrow)
			{
				case ArrowStyle.None:
					{
						layout.LineStyle.StartCap = System.Drawing.Drawing2D.LineCap.NoAnchor;
						layout.LineStyle.EndCap = System.Drawing.Drawing2D.LineCap.NoAnchor;
					}
					break;

				case ArrowStyle.Normal:
					{
						layout.LineStyle.StartCap = System.Drawing.Drawing2D.LineCap.NoAnchor;
						System.Drawing.Drawing2D.AdjustableArrowCap a = new System.Drawing.Drawing2D.AdjustableArrowCap(7, 10, false);  // TUNE:
						layout.LineStyle.CustomEndCap = a;
					}
					break;
			}
		}
		#endregion

		public NodeTemplate this[Type factory]
		{
			get
			{
				return (NodeTemplate)NodeTemplates[factory];
			}
			set
			{
				if(value == null)
				{
					NodeTemplates.Remove(factory);
				}
				else
				{
					NodeTemplate template = (NodeTemplate)value.Clone();
					template.SetFactory(factory);
					NodeTemplates[factory] = template;
				}
			}
		}

		public bool Update(Node node)
		{
			NodeTemplate template = this[node.Data.GetType()];
			string newLabel = Graph.NodeLabel(node.Data, template.NodeName(node.Data));
			if(node.Label == newLabel)
			{
				return false;
			}
			else
			{
				node.Label = newLabel;
				return true;
			}
		}

		public void Add(Quest.Atomic.Database db)
		{
			BeginUpdate();
			try
			{
				foreach(AtomCollection items in db.Collections)
				{
					Add(items);
				}
			}
			finally
			{
				EndUpdate();
			}
		}

		public void Add(Quest.Atomic.AtomCollection items)
		{
			NodeTemplate template = (NodeTemplate)NodeTemplates[items.Factory];
			if((template != null) && template.Visible)
			{
				foreach(Atom item in items)
				{
					Add(item, template);
				}
			}
		}

		public Node Add(Quest.Atomic.Atom item)
		{
			NodeTemplate template = (NodeTemplate)NodeTemplates[item.GetType()];
			return (template == null) ? null : Add(item, template);
		}

		protected Node Add(Quest.Atomic.Atom item, NodeTemplate template)
		{
			Node node = Graph.New(item, template.NodeName(item));
			node.Layout = template.NodeLayout;

			IList edges = template.NodeEdges(item);
			foreach(NodeEdge link in edges)
			{
				if(updateLevel > 0)
				{
					AddUpdateRequest(node, link, template.GetLayout(link.Field), template.GetLabel(link.Field));
				}
				else
				{
					Edge edge = Graph.New(node, link.Field.Name, link.Value);
					if(edge != null)
					{
						edge.Label = template.GetLabel(link.Field);
						edge.Layout = template.GetLayout(link.Field);
					}
				}
			}

			return node;
		}

		private System.Reflection.FieldInfo[] EdgeFields(Node node, Type connector)
		{
			return (node == null) ? null : EdgeFields(node.Data, connector);
		}

		private System.Reflection.FieldInfo[] EdgeFields(Type connector, Node node)
		{
			return (node == null) ? null : EdgeFields(connector, node.Data);
		}

		private System.Reflection.FieldInfo[] EdgeFields(Quest.Atomic.Atom atom, Type connector)
		{
			return (atom == null) ? null : EdgeFields(atom.GetType(), connector);
		}

		private System.Reflection.FieldInfo[] EdgeFields(Type connector, Quest.Atomic.Atom atom)
		{
			return (atom == null) ? null : EdgeFields(connector, atom.GetType());
		}

		private System.Reflection.FieldInfo[] EdgeFields(Type fromAtom, Type toAtom)
		{
			if((fromAtom == null) || (toAtom == null)) return null;
			NodeTemplate template = this[fromAtom];
			if(template == null) return null;

			ArrayList fields = null;

			foreach(System.Reflection.FieldInfo info in template.EdgeFields)
			{
				if (info.FieldType == toAtom)
				{
					if (fields == null)
					{
						fields = new ArrayList(template.EdgeFields.Length);
					}
					fields.Add(info);
				}
			}

			if (fields == null) return null;

			System.Reflection.FieldInfo[] result = new System.Reflection.FieldInfo[fields.Count];
			for(int i = 0; i < fields.Count; i++)
			{
				result[i] = (System.Reflection.FieldInfo)fields[i];
			}
			return result;
		}

		public class LinkInfo
		{
			public struct Entry
			{
				internal Node pNodeFrom;
				internal Node pNodeTo;
				internal string pLabel;
				internal System.Reflection.FieldInfo pFieldTo;

				public Node NodeFrom 
				{ 
					get 
					{ 
						return pNodeFrom; 
					}
				}
				public Node NodeTo 
				{ 
					get 
					{ 
						return pNodeTo;
					}
				}
				public System.Reflection.FieldInfo FieldTo 
				{ 
					get 
					{ 
						return pFieldTo;
					}
				}
				public string Label
				{
					get
					{
						return pLabel;
					}
				}
				public bool Linked
				{
					get
					{
						return FieldTo.GetValue(NodeFrom.Data) == NodeTo.Data;
					}
				}
			}

			public System.Collections.Generic.List<Entry> Items = new System.Collections.Generic.List<Entry>();

			internal void Add(Node aNodeFrom, Node aNodeTo, System.Reflection.FieldInfo aFieldTo, bool onlyLinked, string aLabel)
			{
				Entry entry = new Entry();
				entry.pNodeFrom = aNodeFrom;
				entry.pNodeTo = aNodeTo;
				entry.pFieldTo = aFieldTo;
				entry.pLabel = aLabel;
				if ((!onlyLinked) || entry.Linked)
				{
					Items.Add(entry);
				}
			}
		}

		public LinkInfo GetLinkInfo(Node a, Node b)
		{
			return GetLinkInfo(a, b, false);
		}

		private LinkInfo GetLinkInfo(Node a, Node b, bool onlyLinked)
		{
			if((a == null) || (b == null) || (a.Data == null) || (b.Data == null)) return null;

			LinkInfo result = new LinkInfo();

			System.Reflection.FieldInfo[] infos;

			infos = EdgeFields(a, b.Data.GetType());
			if (infos != null)
			{
				foreach(System.Reflection.FieldInfo info in infos)
				{
					result.Add(a, b, info, onlyLinked, this[a.Data.GetType()].GetDisplayLabel(info));
				}
			}

			infos = EdgeFields(b, a.Data.GetType());
			if (infos != null)
			{
				foreach(System.Reflection.FieldInfo info in infos)
				{
					result.Add(b, a, info, onlyLinked, this[b.Data.GetType()].GetDisplayLabel(info));
				}
			}

			return (result.Items.Count == 0) ? null : result;
		}

		public bool Linkable(Node a, Node b)
		{
			return GetLinkInfo(a, b, false) != null;
		}

		public bool Linkable(Node a, Quest.Graph2D.INodeList b)
		{
			foreach(Node node in b)
			{
				if(Linkable(a, node)) return true;
			}
			return false;
		}

		public bool Linkable(Node node, Type connector)
		{
			return (EdgeFields(node, connector) != null) || (EdgeFields(connector, node) != null);
		}

		public bool Linked(Node a, Node b)
		{
			return GetLinkInfo(a, b, true) != null;
		}

		public IEdgeList Linked(Node a, Node b, bool link)
		{
			LinkInfo info = GetLinkInfo(a, b, false);
			if(info == null) return null;

			IEdgeList result = new EdgeCollectionLight();
			foreach (LinkInfo.Entry entry in info.Items)
			{
				Edge edge = Linked(entry, link);
				if (edge != null)
				{
					result.Add(edge);
				}
			}
			if (result.Count == 0)
			{
				return null;
			}
			else
			{
				return result;
			}
		}

		public Edge Linked(LinkInfo.Entry entry, bool link)
		{
			Atom oldToData = (Atom)entry.FieldTo.GetValue(entry.NodeFrom.Data);
			if ((oldToData == entry.NodeTo.Data) == link) return null;

			if (link)
			{
				entry.FieldTo.SetValue(entry.NodeFrom.Data, entry.NodeTo.Data);
				NodeTemplate template = this[entry.NodeFrom.Data.GetType()];
				Edge edge = Graph.New(entry.NodeFrom, entry.FieldTo.Name ,entry.NodeTo);
				edge.Label = template.GetLabel(entry.FieldTo);
				edge.Layout = template.GetLayout(entry.FieldTo);

				if (oldToData != null)
				{
					Graph.Remove(entry.NodeFrom, entry.FieldTo.Name, oldToData);
				}

				return edge;
			}
			else
			{
				entry.FieldTo.SetValue(entry.NodeFrom.Data, null);
				Graph.Remove(entry.NodeFrom, entry.FieldTo.Name, entry.NodeTo);

				return null;
			}
		}

		public bool Update(Atom atom, System.Reflection.FieldInfo field, Atom newValue)
		{
			Atom oldValue = (Atom)field.GetValue(atom);
			if (oldValue != newValue)
			{
				if (oldValue != null)
				{
					Graph.Remove(Graph[atom], field.Name, oldValue);
				}
				field.SetValue(atom, newValue);
				if (newValue != null)
				{
					NodeTemplate template = this[atom.GetType()];
					Edge edge = Graph.New(Graph[atom], field.Name, newValue);
					if (edge != null)
					{
						edge.Label = template.GetLabel(field);
						edge.Layout = template.GetLayout(field);
					}
					else
					{
						// TODO: should we return false as no visible part of the graph was changed?
					}
				}
				return true;
			}
			else
			{
				return false;
			}
		}

		public void Delete(Node a)
		{
			Graph.Remove(a);
			a.Data.Delete();
			a.Data = null;
		}
	}
}
