using System;
using System.Collections;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace Quest.Graph2D
{
	public interface IEdgeList: IList
	{
		new Edge this[int index] { get; }
		void Add(IEdgeList list);		
	}

	public class EdgeLayout: ICloneable
	{
		public Pen LineStyle;
		public Font Font;
		public Brush TextStyle;

		private static bool UseArrow(LineCap arrow)
		{
			if ((arrow == LineCap.Flat) || (arrow == LineCap.NoAnchor))
			{
				return false;
			}
			else
			{
				return true;
			}
		}

		public bool UseFromArrow
		{
			get
			{
				return UseArrow(LineStyle.StartCap);
			}
		}

		public bool UseToArrow
		{
			get
			{
				return UseArrow(LineStyle.EndCap);
			}
		}

		public object Clone()
		{
			EdgeLayout result = new EdgeLayout();
			result.LineStyle = LineStyle;
			result.Font = Font;
			result.TextStyle = TextStyle;
			return result;
		}

		private void Setup(Color lineColor, Font aFont, Single fontSizeShift, Color textColor)
		{
			if (aFont != null)
			{
				if (fontSizeShift == 0)
				{
					Font = (Font)aFont.Clone();
				}
				else
				{
					Font = new Font(aFont.Name, aFont.Size + fontSizeShift, aFont.Style, aFont.Unit);
				}
			}
			if (lineColor != Color.Empty)
			{
				LineStyle = new Pen(lineColor);
			}
			if (textColor != Color.Empty)
			{
				TextStyle = new SolidBrush(textColor);
			}
		}

		public EdgeLayout(Color lineColor)
		{
			Setup(lineColor, System.Windows.Forms.Control.DefaultFont, -1, Color.FromArgb(0x80, 0x80, 0x80));
		}

		public EdgeLayout()
		{
			Setup(Color.Black, System.Windows.Forms.Control.DefaultFont, -1, Color.FromArgb(0x80, 0x80, 0x80));
		}
	}

	public class Edge: IGraphElement
	{
		public Node From = null;
		public Node To = null;
		public string Key = null;

		public EdgeLayout Layout;

		public Edge()
		{
		}

		protected int LabelHeight;
		private string pLabel = "";   // TODO: remove
		private string pPrivateLabel = "";
		public string Label
		{
			get
			{
				return pLabel;
			}
			set
			{
				pLabel = (value == null) ? "" : value;
				pPrivateLabel = pLabel;
			}
		}

		internal void SetAsGroupMaster(IEdgeList siblings)
		{
			if (siblings == null)
			{
				pLabel = pPrivateLabel;
			}
			else
			{
				// group master displays labels of all group members
				ArrayList labels = new ArrayList();

				labels.Add(pPrivateLabel);
				foreach(Edge sibling in siblings)
				{
					string label = sibling.pPrivateLabel;
					if (!String.IsNullOrEmpty(label))
					{
						labels.Add(label);
					}
				}
				labels.Sort();

				pLabel = "";
				foreach(string label in labels)
				{
					if (pLabel != "")
					{
						pLabel += "\n";
					}
					pLabel += label;
				}
			}
		}

		internal void SetAsGroupSlave()
		{
			// group slaves don't use their own labels
			pLabel = "";
		}

		#region Quest.Graph2D.IGraphElement interface implementation
		public Rectangle Bounds 
		{
			get
			{
				int x1 = From.Position.X;
				int y1 = From.Position.Y;
				int x2 = To.Position.X;
				int y2 = To.Position.Y;

				int left, top, right, bottom;
				if(x1 < x2)
				{
					left = x1;
					right = x2;
				}
				else
				{
					left = x2;
					right = x1;
				}
				if(y1 < y2)
				{
					top = y1;
					bottom = y2;
				}
				else
				{
					top = y2;
					bottom = y1;
				}
				return new Rectangle(left, top, right - left, bottom - top);
			}
		}

		public object HitTest(int aX, int anY) 
		{
			// TODO: implement line hit test with a few pixels +/- tolerance
			return null;
		}
		#endregion

		public virtual void Calibrate(Graphics canvas)
		{
			if ((Layout == null) || (Layout.Font == null))
			{
				LabelHeight = 0;
			}
			else
			{
				LabelHeight = (int)canvas.MeasureString("M", Layout.Font).Height;
			}
		}

		public virtual void PaintTo(Graphics canvas)
		{
			if(Layout != null)
			{
				if((From != null) && (To != null) && (Layout.LineStyle != null))
				{
					if(!From.Bounds.IntersectsWith(To.Bounds))
					{
						Point fromPosition;
						Point toPosition;

						Pen lineStyle;
						if(From.Selected || To.Selected)
						{
							lineStyle = (Pen)Layout.LineStyle.Clone();
							lineStyle.Color = Color.Blue; // TUNE: TODO:
							lineStyle.Width = 2;
						}
						else
						{
							lineStyle = Layout.LineStyle;
						}

						// TODO: use reduced painting if current PixelSize is too small (as in Node.PaintTo)

						bool drawLabel = (Layout.Font != null) && (Layout.TextStyle != null) && (pLabel != "");
						bool lowResolution = !(GraphElement.PixelSize(canvas, LabelHeight) >= 5 /* TUNE: */);

						if (lowResolution) drawLabel = false;

						if((!Layout.UseFromArrow) && (!Layout.UseToArrow) && (!drawLabel))
						{
							// speed optimization: don't bother with calculation of proper drawing
							fromPosition = From.Position;
							toPosition = To.Position;
						}
						else
						{
							fromPosition = From.BoundaryHit(To.Position);
							toPosition = To.BoundaryHit(From.Position);
						}

						// TODO: TUNE:
						// lineStyle.CustomEndCap = new System.Drawing.Drawing2D.AdjustableArrowCap(4, 2, true);
						canvas.DrawLine
							(
							 lineStyle,
							 fromPosition.X,
							 fromPosition.Y,
							 toPosition.X,
							 toPosition.Y
							);

						if (drawLabel)
						{
							PointF center;
							SizeF textSize;

							{
								const int LabelBorder = 3 /* TUNE: how many pixels keep between label and node */ * 2;

								textSize = canvas.MeasureString(pLabel, Layout.Font);
								textSize.Width += LabelBorder;
								textSize.Height += LabelBorder;

								center = new PointF(fromPosition.X, fromPosition.Y);

								PointF direction = new PointF(toPosition.X - fromPosition.X, toPosition.Y - fromPosition.Y);

								// now we need to move the 'bounds' rectangle (center +- textSize/2) in the 'direction' direction so
								// that it doesn't intersect the 'From.Bounds'

								Single t = Single.NaN;
								Single tx;
								Single ty;

								if (direction.X != 0)
								{
									tx = Math.Abs((textSize.Width / 2) / direction.X);
									if (Single.IsNaN(t) || (t > tx)) t = tx;
								}

								if (direction.Y != 0)
								{
									ty = Math.Abs((textSize.Height / 2) / direction.Y);
									if (Single.IsNaN(t) || (t > ty)) t = ty;
								}

								if (!Single.IsNaN(t))
								{
									center.X = (Single)(center.X + t * direction.X);
									center.Y = (Single)(center.Y + t * direction.Y);
								}
							}

							StringFormat format = new StringFormat();
							format.Alignment = StringAlignment.Center;
							format.LineAlignment = StringAlignment.Center;

							canvas.DrawString(
								pLabel,
								Layout.Font,
								(From.Selected || To.Selected) ? Brushes.Blue : Layout.TextStyle,
								center,
								format);

							/*
							canvas.DrawRectangle(
								Layout.LineStyle,
								center.X - textSize.Width / 2,
								center.Y - textSize.Height / 2,
								textSize.Width,
								textSize.Height);
							 */
						}
					}
				}
			}
		}

		internal bool Mark = false;
	}

	public class EdgeCollectionLight : ArrayList, IEdgeList
	{
		public new Edge this[int index]
		{
			get
			{
				return (Edge)base[index];
			}
		}

		public void Add(IEdgeList list)
		{
			foreach(Edge e in list)
			{
				Add(e);
			}
		}
	}

	public class EdgeCollectionBase: EdgeCollectionLight
	{
		public void Calibrate(Graphics canvas)
		{
			foreach(Edge e in this)
			{
				e.Calibrate(canvas);
			}
		}

		public void PaintTo(Graphics canvas)
		{
			foreach(Edge e in this)
			{
				e.PaintTo(canvas);
			}
		}

		private int updateLevel = 0;
		private bool invalidateUpdateRequest;

		public void BeginUpdate()
		{
			if(updateLevel == 0)
			{
				invalidateUpdateRequest = false;
			}
			updateLevel++;
		}

		public void EndUpdate()
		{
			updateLevel--;
			if(updateLevel == 0)
			{
				if (invalidateUpdateRequest)
				{
					Invalidate();
				}
			}
		}

		public void Invalidate()
		{
			if (updateLevel == 0)
			{
				InternalInvalidate();
			}
			else
			{
				invalidateUpdateRequest = true;
			}
		}

		protected virtual void InternalInvalidate()
		{
		}

		public Edge Find(Node nodeFrom, string key, Node nodeTo)
		{
			foreach(Edge e in this)
			{
				if(((e.From == nodeFrom) && (e.Key == key) && (e.To == nodeTo)))
				{
					return e;
				}
			}
			return null;
		}

		protected IEdgeList FindSiblings(Edge template)
		{
			// TODO: don't search for siblings if underlying Atom structure
			// doesn't contain siblings (no two fields exist with same To-type)
			EdgeCollectionLight result = new EdgeCollectionLight();
			foreach(Edge e in this)
			{
				if (e == template) continue;

				if((e.From == template.From) && (e.To == template.To))
				{
					result.Add(e);
				}
			}
			if (result.Count == 0)
			{
				return null;
			}
			else
			{
				return result;
			}
		}

		public IEdgeList Find(Node root)
		{
			EdgeCollectionLight result = new EdgeCollectionLight();
			foreach(Edge e in this)
			{
				if ((e.From == root) || (e.To == root))
				{
					result.Add(e);
				}
			}
			return result;
		}

		public INodeList NeighboursOf(Node root)
		{
			NodeCollectionLight result = new NodeCollectionLight();
			foreach(Edge e in this)
			{
				if(e.From == root)
				{
					result.Add(e.To);
				}
				else if(e.To == root)
				{
					result.Add(e.From);
				}
			}
			return result;
		}

		public void Remove(Node nodeFrom, string key, Node nodeTo)
		{
			Edge edge = Find(nodeFrom, key, nodeTo);
			if(edge != null)
			{
				this.Remove(edge);
			}
		}

		#region Change reflection
		public override object Clone()
		{
			EdgeCollectionBase result = new EdgeCollectionBase();
			foreach (Edge edge in this)
			{
				result.Add(edge);
			}
			return result;
		}

		protected virtual void Changed()
		{
			Invalidate();
		}

		protected virtual void Link(Edge value, int index)
		{
		}

		protected virtual void Unlink(Edge value)
		{
		}

		public override int Add(object value)
		{
			int result = base.Add(value);
			Link((Edge)value, result);
			Changed();
			return result;
		}

		public override void Clear()
		{
			while (Count > 0) RemoveAt(Count - 1);
		}

		public override void Remove(object obj)
		{
			RemoveAt(IndexOf(obj));
		}

		public override void RemoveAt(int index)
		{
			if ((index >= 0) && (index < Count))
			{
				Unlink(this[index]);
				base.RemoveAt(index);
				Changed();
			}
		}
		#endregion
	}

	public class EdgeCollection : EdgeCollectionBase
	{
		private IGraph pGraph;

		public EdgeCollection(IGraph owner)
		{
			pGraph = owner;
		}

		protected override void InternalInvalidate()
		{
			IEdgeList list = new EdgeCollectionLight();
			list.Add(this);

			int i = 0;
			while (i < list.Count)
			{
				Edge current = list[i];
				IEdgeList siblings = FindSiblings(current);

				if (siblings == null)
				{
					// this edge has no siblings

					current.SetAsGroupMaster(null);
					i++;
				}
				else
				{
					// current edge has siblings. Mark the current one as master
					// and siblings as slaves

					current.SetAsGroupMaster(siblings);
					foreach(Edge sibling in siblings)
					{
						sibling.SetAsGroupSlave();
					}
					// remove all edges processed so far, including the current one
					while (i >= 0)
					{
						list.RemoveAt(i);
						i--;
					}
					// remove all sibling edges because those were processed too
					foreach(Edge sibling in siblings)
					{
						list.Remove(sibling);
					}
					// restart the loop with remaining edges
					i = 0;
				}
			}
		}
	}
}
