using System;
using System.Configuration;
using System.Collections;
using System.Drawing;

namespace Quest.Graph2D
{
	public class Relaxer
	{
		public Graph Graph;

		public int dn = 20;   // field size
		public float rn = 0.6f; // force strength factor
		public float environmentResistance = 1f;
		public float forceQuantum = 8f;
		public int edgeLength = 60;
		public float cf = 0.2f;
		public float cfn = 0.01f;
		public float rotationForce = 1.0f;
		public float edgeForce = 1.0f;

		public Relaxer()
		{
			ISettingsReader settings = new ApplicationSettings("Relaxer");
			settings.Read("dn", ref dn);
			settings.Read("rn", ref rn);
			settings.Read("environmentResistance", ref environmentResistance);
			settings.Read("forceQuantum", ref forceQuantum);
			settings.Read("edgeLength", ref edgeLength);
			settings.Read("cf", ref cf);
			settings.Read("cfn", ref cfn);
			settings.Read("rotationForce", ref rotationForce);
			settings.Read("edgeForce", ref edgeForce);
		}

		#region Utilities
		internal struct PolarPoint
		{
			public double Radius;
			public double Angle;
		}

		private double NormalizedAngle(double angle)
		{
			double result = angle;
			while(result < 0)
			{
				result += 2 * Math.PI;
			}
			while(result >= 2 * Math.PI)
			{
				result -= 2 * Math.PI;
			}
			return result;
		}

		private PolarPoint Convert(Point p)
		{
			PolarPoint result = new PolarPoint();
			result.Radius = Distance(p);
			int x = p.X;
			int y = p.Y;
			if(x != 0)
			{
				result.Angle = Math.Acos(x / result.Radius);
				if(y < 0)
				{
					result.Angle = 2 * Math.PI - result.Angle;
				}
			}
			else if(y != 0)
			{
				result.Angle = Math.Asin(y / result.Radius);
				if(x < 0)
				{
					result.Angle = Math.PI - result.Angle;
				}
			}
			result.Angle = NormalizedAngle(result.Angle);
			return result;
		}

		private Point Convert(PolarPoint p)
		{
			return new Point((int)(p.Radius * Math.Cos(p.Angle)), (int)(p.Radius * Math.Sin(p.Angle)));
		}

		private float Distance(Point a, Point b)
		{
			int dx = b.X - a.X;
			int dy = b.Y - a.Y;
			return (float)Math.Sqrt(dx * dx + dy * dy);
		}

		private float Distance(Point d)
		{
			return (float)Math.Sqrt(d.X * d.X + d.Y * d.Y);
		}

		private float Distance(Rectangle r)
		{
			return Distance(new Point(r.Width, r.Height));
		}

		private Point Direction(Point a, Point b)
		{
			return new Point(b.X - a.X, b.Y - a.Y);
		}

		private Point Direction(Node a, Node b)
		{
			if(!a.Bounds.IntersectsWith(b.Bounds))
			{
				return Direction
					(
					 a.BoundaryHit(b.Position),
					 b.BoundaryHit(a.Position)
					);
			}
			else
			{
				return new Point(0, 0);
			}
		}

		private PointF UnitDirection(Point a, Point b)
		{
			return UnitDirection(Direction(a, b));
		}

		private PointF UnitDirection(Point d)
		{
			float factor = (float)(1 / Distance(d));
			return new PointF(d.X * factor, d.Y * factor);
		}

		private Rectangle FieldBounds(Rectangle root, int fieldSize)
		{
			int left = root.Left - fieldSize;
			int top = root.Top - fieldSize;
			int right = root.Right + fieldSize;
			int bottom = root.Bottom + fieldSize;
			return new Rectangle(left, top, right - left, bottom - top);
		}

		private float Mass(Rectangle r)
		{
			return r.Width * r.Height;
		}

		private Point Center(Rectangle r)
		{
			return new Point
				(
				 r.Left + (r.Width / 2),
				 r.Top + (r.Height / 2)
				);
		}
		#endregion

		private Random generator = new Random();

		/// In order to prevent node overlapping, every node has a geometry field
		/// that generates repulsion forces on pieces that come into that field.
		/// The field is homogenous and rectangular shape. Parameter 'dn' determines
		/// range of the field beyond node bounds.
		/// 
		/// this function calculates force that is generated by 'node' and works on
		/// 'neighbour'
		private PointF Repulsion(Node node, Node neighbour)
		{
			float dx = 0;
			float dy = 0;

			// field within which repulsion force will be generated
			Rectangle field = FieldBounds(node.Bounds, dn);
			// alien object within field
			Rectangle alien = neighbour.Bounds;
			if(field.IntersectsWith(alien))
			{
				alien.Intersect(field);
				// the force will work only on the piece that is within the field

				Point forceDirection = Direction(node.Position, Center(alien));

				while((forceDirection.X == 0) && (forceDirection.Y == 0))
				{
					const int f = 25;
					forceDirection.X = generator.Next(2 * f + 1) - f;
					forceDirection.Y = generator.Next(2 * f + 1) - f;
				}

				if((forceDirection.X == 0) || (forceDirection.Y == 0))
				{
					do
					{
						// random factor
						const int f = 2;

						forceDirection.X += generator.Next(2 * f + 1) - f;
						forceDirection.Y += generator.Next(2 * f + 1) - f;
					}
					while((forceDirection.X == 0) && (forceDirection.Y == 0));
				}

				PointF forceVector = UnitDirection(forceDirection);
//				float forceStrength = rn * (float)Math.Sqrt(Mass(alien));
				float forceStrength = rn * Distance(alien); // * ((Mass(alien) / Mass(neighbour.Bounds)));
				dx = forceVector.X * forceStrength;
				dy = forceVector.Y * forceStrength;
			}
			return new PointF(dx, dy);
		}

		private double ClockwiseAngleDistance(PolarPoint a, PolarPoint b)
		{
			double angleA = a.Angle;
			double angleB = b.Angle;
			while (angleB < angleA)
			{
				angleB += 2 * Math.PI;
			}
			return NormalizedAngle(angleB - angleA);
		}

		private struct PolarNode
		{
			public PolarPoint Point;
			public Node Node;
		}

		private class PolarNodeComparer: IComparer
		{
			public int Compare(object x, object y)
			{
				double dif = ((PolarNode)x).Point.Angle - ((PolarNode)y).Point.Angle;
				if(dif > 0)
				{
					return 1;
				}
				else if(dif < 0)
				{
					return -1;
				}
				else
				{
					return 0;
				}
			}
		}

		private void StarRepulse(Node root, PolarNode a, PolarNode b, int beamCount)
		{
			PolarPoint pa = a.Point;
			PolarPoint pb = b.Point;

			double idealAngle = (2 * Math.PI) / beamCount;
			double angle = (ClockwiseAngleDistance(pa, pb) - idealAngle) / 2;
			pb.Angle -= angle;
			pa.Angle += angle;

			float f = rotationForce * (float)Math.Abs(NormalizedAngle(angle) / (2 * idealAngle)); // TUNE:
			double r = (pa.Radius + pb.Radius) / 2;
			float fa = (float)(f * (pa.Radius / r));
			float fb = (float)(f * (pb.Radius / r));

			Point na = Convert(pa);
			na.X += root.Position.X - a.Node.Position.X;
			na.Y += root.Position.Y - a.Node.Position.Y;

			a.Node.Momentum.X += fa * na.X;
			a.Node.Momentum.Y += fa * na.Y;

			Point nb = Convert(pb);
			nb.X += root.Position.X - b.Node.Position.X;
			nb.Y += root.Position.Y - b.Node.Position.Y;

			b.Node.Momentum.X += fb * nb.X;
			b.Node.Momentum.Y += fb * nb.Y;
		}

		public PointF ForceLimit(float dx, float dy, float limit)
		{
			double sizeSqr = dx * dx + dy * dy;
			if(sizeSqr > limit * limit)
			{
				double f = limit / Math.Sqrt(sizeSqr);
				return new PointF((float)(dx * f), (float)(dy * f));
			}
			else
			{
				return new PointF(dx, dy);
			}
		}

		public void Relax()
		{
			// Edge generates force that becomes stable when edge's length is
			// 'edgeLength'. Otherwise the force works on both end points either
			// in attracting or distracting direction
			foreach(Edge edge in Graph.Edges)
			{
				if((edge.From == null) || (edge.To == null)) continue;

				double vx;
				double vy;
				double len;

				// default calculation
				vx = edge.To.Position.X - edge.From.Position.X;
				vy = edge.To.Position.Y - edge.From.Position.Y;
				len = 0.0001;

				if(!edge.From.Bounds.IntersectsWith(edge.To.Bounds))
				{
					Point toPos = edge.To.BoundaryHit(edge.From.Position);
					Point fromPos = edge.From.BoundaryHit(edge.To.Position);
					if(!toPos.Equals(fromPos))
					{
						vx = toPos.X - fromPos.X;
						vy = toPos.Y - fromPos.Y;
						len = Math.Sqrt(vx * vx + vy * vy);
					}
				}

				double f = edgeForce * ((edgeLength - len) / len);
				if (edge.From.Selected || edge.To.Selected)
				{
					f = f * 150; // TUNE: apply more force for selected edges 
				}
				else if (edge.From.Pin || edge.To.Pin)
				{
					f = f * 40;  // TUNE: apply more force for pinned edges
				}
				float dx = (float)(f * vx);
				float dy = (float)(f * vy);
				edge.To.Momentum.X += dx;
				edge.To.Momentum.Y += dy;
				edge.From.Momentum.X -= dx;
				edge.From.Momentum.Y -= dy;
			}

			// for each node we'll calibrate geometry of the link star
			// so that the link beams are distributed equally
			foreach(Node node in Graph.Nodes)
			{
				INodeList star = Graph.NeighboursOf(node);
				if(star.Count > 1)
				{
					ArrayList list = new ArrayList();
					foreach(Node neighbour in star)
					{
						if(Distance(node.Position, neighbour.Position) > 0)
						{
							PolarNode item = new PolarNode();
							item.Node = neighbour;
							item.Point = Convert(Direction(node.Position, neighbour.Position));
							list.Add(item);
						}
					}
					if(list.Count > 1)
					{
						list.Sort(0, list.Count - 1, new PolarNodeComparer());
						for(int i = 1; i < list.Count; i++)
						{
							StarRepulse(node, (PolarNode)list[i - 1], (PolarNode)list[i], list.Count);
						}
						if(list.Count > 2)
						{
							StarRepulse(node, (PolarNode)list[list.Count - 1], (PolarNode)list[0], list.Count);
						}
					}
				}
			}

			// for every two nodes we'll calculate repulsion forces
			foreach(Node nA in Graph.Nodes)
			{
				foreach(Node nB in Graph.SelectNodes(FieldBounds(nA.Bounds, dn), false))
				{
					if(nA != nB)
					{
						PointF dif;
						dif = Repulsion(nA, nB);
						nB.Momentum.X += dif.X;
						nB.Momentum.Y += dif.Y;

						dif = Repulsion(nB, nA);
						nA.Momentum.X += dif.X;
						nA.Momentum.Y += dif.Y;
					}
				}
			}

			// apply forces. single jump is limited to (-'forceQuantum',+'forceQuantum').
			// node's momentum slows down correspondingly to 'environmentResistance'
			foreach(Node node in Graph.Nodes)
			{
				if(!(node.Selected || node.Pin))
				{
					PointF dif = ForceLimit(node.Momentum.X, node.Momentum.Y, forceQuantum);
					node.Position.X = (int)(node.Position.X + dif.X);
					node.Position.Y = (int)(node.Position.Y + dif.Y);
					node.Momentum.X *= (1 - environmentResistance);
					node.Momentum.Y *= (1 - environmentResistance);
				}
				else
				{
					node.Momentum.X = 0;
					node.Momentum.Y = 0;
				}
			}
		}
	}
}
