/* KEYWORDS: VERSION
 * VERSION: 1.6
 * 1.6 / 2008-10-11
 * - support for overriding the file access in IniFile.CreateInputStream(), IniFile.CreateOutputStream()
 * 1.5 / 2008-07-02
 * - fixed: IniFile.WriteString() preserves leading and trailing spaces
 * - fixed: IniFile.ReadString() -> '"' failed
 * 1.4 / 2007-12-28
 * - changed return type of IniFile.ReadSection() -> NameValueCollection
 *   in order to preserve exact key names
 * 1.3 / 2007-01-03
 * - new function IniFile.Delete(key)
 * 1.2 / 2006-05-22
 * - new function IniFile.ReadSection() -> StringDictionary
 * - new quoted strings are read as unquoted
 */
using System;
using System.IO;
using System.Collections.Specialized;

namespace System.Configuration
{
	using Section = NameValueCollection;

	public class IniFile: SettingsIOBase
	{
		private class Sections: NameObjectCollectionBase
		{
			public Section Get(string key, bool canCreate)
			{
				Section result = (Section)BaseGet(key);
				if((result == null) && canCreate)
				{
					result = new Section();
					BaseSet(key, result);
				}
				return result;
			}

			public Section Get(string key)
			{
				return Get(key, !IsReadOnly);
			}

			public void Clear()
			{
				BaseClear();
			}
		}

		private Sections pSections = new Sections();
		private bool Modified = false;
		private string pFileName;
		private System.Text.Encoding pEncoding;

		public IniFile(string fileName)
		{
			pFileName = fileName;
			pEncoding = null;
			LoadFile();
		}

		public IniFile(string fileName, System.Text.Encoding encoding)
		{
			pFileName = fileName;
			pEncoding = encoding;
			LoadFile();
		}
		
		protected IniFile()
		{
			pFileName = null;
			pEncoding = null;
		}
		
		protected IniFile(System.Text.Encoding encoding)
		{
			pFileName = null;
			pEncoding = encoding;
		}
		
		protected void Initialize()
		{
			LoadFile();
		}
		
		#region IniFile IO
		private void LoadStream(Stream input)
		{
			pSections.Clear();
			TextReader stream = null;
			if (input != null)
			{
				try
				{
					stream = System.Text.FileStream.CreateReader(input, pEncoding);
				}
				catch(Exception)
				{
					stream = null;
				}
			}
			if(stream != null)
			{
				try
				{
					Section sect = null;
					string line;
					while(true)
					{
						line = stream.ReadLine();
						if(line == null) break;

						line = line.Trim();
						if(line == "") continue;
						if(line[0] == ';') continue;
						if(line[0] == '[')
						{
							string sectionName = line.Substring(1, line.Length - 2);
							sect = pSections.Get(sectionName, true);
							continue;
						}
						if(sect == null)
						{
							sect = pSections.Get("", true);
						}
						int i = line.IndexOf('=');
						if(i < 0)
						{
							sect.Set(line, "");
						}
						else
						{
							string key = line.Substring(0, i).Trim();
							line = line.Remove(0, i + 1).Trim();
							sect.Set(key, line);
						}
					}
				}
				finally
				{
					stream.Close();
				}
			}
			Modified = false;
		}
		
		protected virtual Stream CreateOutputStream()
		{
			if (String.IsNullOrEmpty(pFileName))
			{
				throw new ArgumentException();
			}
			return new System.IO.FileStream(pFileName, FileMode.Create, FileAccess.Write);
		}

		protected virtual Stream CreateInputStream()
		{
			if (String.IsNullOrEmpty(pFileName))
			{
				throw new ArgumentException();
			}
			return new System.IO.FileStream(pFileName, FileMode.Open, FileAccess.Read);
		}
		
		private void LoadFile()
		{
			Stream input;
			try
			{
				input = CreateInputStream();
			}
			catch(Exception)
			{
				input = null;
			}
			try
			{
				LoadStream(input);
			}
			finally
			{
				if (input != null)
				{
					input.Dispose();
				}
			}
		}

		private void UpdateFile()
		{
			if(!Modified) return;

			Stream output = CreateOutputStream();
			try
			{
				UpdateStream(output);
			}
			finally
			{
				output.Dispose();
			}
			Modified = false;
		}
		
		private void UpdateStream(Stream output)
		{
			StreamWriter stream = new StreamWriter(output, (pEncoding == null) ? System.Text.Encoding.Default : pEncoding);
			try
			{
				foreach(string sectionName in pSections.Keys)
				{
					Section sect = pSections.Get(sectionName, false);
					stream.WriteLine("[{0}]", sectionName);
					foreach(string key in sect.Keys)
					{
						string value = sect.Get(key);
						stream.WriteLine("{0}={1}", key, value);
					}
				}
			}
			finally
			{
				stream.Close();
			}
			Modified = false;
		}
		#endregion

		private void Split(out string section, ref string key)
		{
			int split = key.IndexOfAny(new char[]{'/', '\\'});
			if(split < 0)
			{
				section = "";
			}
			else
			{
				section = key.Substring(0, split);
				key = key.Remove(0, split + 1);
			}
		}

		public NameValueCollection ReadSection(string key)
		{
			NameValueCollection result = new NameValueCollection();
			Section sect = pSections.Get(key, false);
			if(sect != null)
			{
				foreach(string name in sect.Keys)
				{
					result.Add(name, System.Convert.ToString(StringToObject(sect.Get(name))));
				}
			}
			return result;
		}

		protected static object StringToObject(string value)
		{
			if(value == null)
			{
				return null;
			}
			else if (value == "")
			{
				return value;
			}
			else if(value[0] == '"')
			{
				if(value.EndsWith("\"") && (value.Length > 1))
				{
					return value.Substring(1, value.Length - 2);
				}
				else
				{
					// TODO: ? unterminated string literal
					return value;	// return value.Substring(1, value.Length - 1);
				}
			}
			else
			{
				// TODO: guess value type
				return value;
			}
		}

		protected override object ReadObject(string key, object defaultValue)
		{
			string sectionName;
			string keyName = key;
			Split(out sectionName, ref keyName);
			Section sect = pSections.Get(sectionName, false);
			if(sect == null)
			{
				return defaultValue;
			}
			string value = sect.Get(keyName);
			return (value == null) ? defaultValue : StringToObject(value);
		}

		public override void WriteString(string key, string value)
		{
			string sectionName;
			string keyName = key;
			Split(out sectionName, ref keyName);
			Section sect = pSections.Get(sectionName, true);
			if(sect != null)
			{
				string oldValue = sect.Get(keyName);
				if(value != oldValue)
				{
					// enquote value if needed
					if (!String.IsNullOrEmpty(value) && (value.StartsWith(" ") || value.EndsWith(" ")))
					{
						value = "\"" + value + "\"";
					}
					sect.Set(keyName, value);
					Modified = true;
				}
			}
		}

		public override void Delete(string key)
		{
			string sectionName;
			string keyName = key;
			Split(out sectionName, ref keyName);
			Section sect = pSections.Get(sectionName, true);
			if (sect != null)
			{
				if (String.IsNullOrEmpty(keyName))
				{
					sect.Clear();
					Modified = true;
				}
				else
				{
					string oldValue = sect.Get(keyName);
					if (oldValue != null)
					{
						sect.Remove(keyName);
						Modified = true;
					}
				}
			}
		}

		#region System.IDisposable interface implementation
		protected override void Dispose(bool disposing)
		{
			if(disposing)
			{
				if(pSections != null)
				{
					UpdateFile();
					pSections = null;
				}
			}
			base.Dispose(disposing);
		}
		#endregion
	}
}
