using System;
using System.Collections.Specialized;

namespace System.Configuration
{
	public interface ISettingsReader: IDisposable
	{
		string ReadString(string key, string defaultValue);
		bool ReadBoolean(string key, bool defaultValue);
		float ReadSingle(string key, float defaultValue);
		int ReadInt32(string key, int defaultValue);
		string[] ReadStrings(string key);
		
		string Read(string key, ref string value);
		bool Read(string key, ref bool value);
		float Read(string key, ref float value);
		int Read(string key, ref int value);
	}

	public interface ISettingsWriter: IDisposable
	{
		void WriteString(string key, string value);
		void WriteBoolean(string key, bool value);
		void WriteSingle(string key, float value);
		void WriteInt32(string key, int value);

		void Write(string key, string value);
		void Write(string key, bool value);
		void Write(string key, float value);
		void Write(string key, int value);

		void Delete(string key);
	}

	public class SettingsIOBase: ISettingsReader, ISettingsWriter
	{
		protected SettingsIOBase()
		{
		}

		#region Implementation of System.IDisposable
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		~SettingsIOBase()
		{
			Dispose(false);
		}

		protected virtual void Dispose(bool disposing)
		{
		}
		#endregion

		#region System.Configuration.ISettingsReader interface implementation
		protected virtual object ReadObject(string key, object defaultValue)
		{
			return defaultValue;
		}

		public string ReadString(string key, string defaultValue) 
		{
			object result = ReadObject(key, defaultValue);
			return (result == null) ? null : System.Convert.ToString(result);
		}

		public bool ReadBoolean(string key, bool defaultValue) 
		{
			return System.Convert.ToBoolean(ReadObject(key, defaultValue));
		}

		public float ReadSingle(string key, float defaultValue) 
		{
			return System.Convert.ToSingle(ReadObject(key, defaultValue), System.Globalization.NumberFormatInfo.InvariantInfo);
		}

		public int ReadInt32(string key, int defaultValue) 
		{
			return System.Convert.ToInt32(ReadObject(key, defaultValue));
		}
		
		public string[] ReadStrings(string key)
		{
			string code = ReadString(key, null);
			if (String.IsNullOrEmpty(code))
			{
				return null;
			}
			// TODO: use quote skipping parser
			string[] result = code.Split(',');
			for(int i = 0; i < result.Length; i++)
			{
				string item = result[i].Trim();
				if (item.StartsWith("\"") && item.EndsWith("\""))
				{
					item = item.Substring(1, item.Length - 2);
				}
				result[i] = item;
			}
			return result;
		}

		public string Read(string key, ref string value) 
		{
			value = ReadString(key, value);
			return value;
		}

		public bool Read(string key, ref bool value) 
		{
			value = ReadBoolean(key, value);
			return value;
		}

		public float Read(string key, ref float value) 
		{
			value = ReadSingle(key, value);
			return value;
		}

		public int Read(string key, ref int value) 
		{
			value = ReadInt32(key, value);
			return value;
		}
		#endregion

		#region System.Configuration.ISettingsWriter interface implementation
		public virtual void WriteString(string key, string value)
		{
		}

		public void WriteBoolean(string key, bool value) 
		{
			WriteString(key, value.ToString());
		}

		public void WriteSingle(string key, float value) 
		{
			WriteString(key, value.ToString(System.Globalization.NumberFormatInfo.InvariantInfo));
		}

		public void WriteInt32(string key, int value) 
		{
			WriteString(key, value.ToString());
		}

		public void Write(string key, string value) 
		{
			WriteString(key, value);
		}

		public void Write(string key, bool value) 
		{
			WriteBoolean(key, value);
		}

		public void Write(string key, float value) 
		{
			WriteSingle(key, value);
		}

		public void Write(string key, int value) 
		{
			WriteInt32(key, value);
		}

		public virtual void Delete(string key)
		{
		}
		#endregion
	}

	public class ApplicationSettings: SettingsIOBase
	{
		private NameValueCollection list;

		public ApplicationSettings()
		{
			list = ConfigurationSettings.AppSettings;
		}

		public ApplicationSettings(string sectionName)
		{
			list = (NameValueCollection)ConfigurationSettings.GetConfig(sectionName);
			if(list == null)
			{
				list = new NameValueCollection();
			}
		}

		protected override object ReadObject(string key, object defaultValue)
		{
			string value = list.Get(key);
			return (value == null) ? defaultValue : value;
		}
	}
}
