using System;

namespace System.IO
{
	public class SeekableInputStream: Stream
	{
		private Stream _base;
		private long _position;
		private long _baseLength;
		private StreamCreator _rewinder;
		
		public delegate Stream StreamCreator();
		
		public SeekableInputStream(Stream baseStream, StreamCreator rewinder, long baseLength)
		{
			_base = baseStream;
			_position = 0;
			_rewinder = rewinder;
			_baseLength = baseLength;
		}
		
		public SeekableInputStream(Stream baseStream, StreamCreator rewinder)
		{
			_base = baseStream;
			_position = 0;
			_rewinder = rewinder;
			_baseLength = -1;
		}
		
		protected override void Dispose(bool disposing)
		{
			if (_base != null)
			{
				_base.Dispose();
				_base = null;
			}
			base.Dispose(disposing);
		}
		
		public override bool CanRead 
		{
			get 
			{ 
				return _base.CanRead; 
			}
		}
		
		public override bool CanSeek 
		{
			get 
			{
				return true;
			}
		}
		
		public override bool CanWrite 
		{
			get 
			{
				return false;
			}
		}
		
		public override void Close()
		{
			if (_base != null)
			{
				_base.Close();
				_base.Dispose();
				_base = null;
			}
		}
		
		public override long Length 
		{
			get 
			{
				if (_baseLength < 0)
				{
					return _base.Length;
				}
				else
				{
					return _baseLength;
				}
			}
		}
		
		public override long Position 
		{
			get 
			{
				return _position;
			}
			set 
			{
				if (value != _position)
				{
					Seek(value, SeekOrigin.Begin);
				}
			}
		}
		
		public override void Flush()
		{
		}
		
		public override void Write(byte[] buffer, int offset, int count)
		{
			if (count > 0)
			{
				throw new NotSupportedException();
			}
		}
		
		public override void SetLength(long value)
		{
			if (value != Length)
			{
				throw new NotSupportedException();
			}
		}
		
		public override long Seek(long offset, SeekOrigin origin)
		{
			long currentPosition = Position;
			long newPosition;
			
			switch (origin)
			{
				case SeekOrigin.Begin:
					newPosition = offset;
					break;
					
				case SeekOrigin.Current:
					newPosition = currentPosition + offset;
					break;
					
				case SeekOrigin.End:
					newPosition = Length + offset;
					break;
					
				default:
					newPosition = currentPosition;
					break;
			}
			
			if (newPosition == currentPosition)
			{
				return newPosition;
			}
			
			if (newPosition < 0)
			{
				throw new ArgumentException("Attempted seeking before the start of the stream");
			}
			
			if (newPosition > Length)
			{
				throw new ArgumentException("Attempted seeking after the end of the stream"); 
			}
			
			if (newPosition < currentPosition)
			{
				if (_rewinder == null)
				{
					throw new NotSupportedException();
				}
				
				_base.Close();
				_base.Dispose();
				_base = null;
				_position = 0;
				_base = _rewinder();
				return Seek(newPosition, SeekOrigin.Begin);
			}
			else
			{
				byte[] buffer = new byte[1024 /* TODO: TUNE: */];
				while (newPosition > currentPosition)
				{
					long toRead = newPosition - currentPosition;
					if (toRead > buffer.Length)
					{
						toRead = buffer.Length;
					}
					// TODO: what if Read() fails to advance?
					currentPosition = currentPosition + Read(buffer, 0, Convert.ToInt32(toRead));
				}
				return newPosition;
			}
		}

		public override int Read(byte[] buffer, int offset, int count)
		{
			if (count <= 0)
			{
				return 0;
			}
			long available = Length - _position;
			if ((long)count > available)
			{
				count = (int)available;
				if (count <= 0)
				{
					return 0;
				}
			}
			int result = _base.Read(buffer, offset, count);
			_position += result;
			return result;
		}
	}
}
