using System;

namespace System.Threading
{
	using RecordId = UInt32;
	
	public class MessageBoard: IDisposable
	{
		public const string ToEveryone = "*";
		
		public class Entry: IDisposable
		{
			#region System.IDisposable interface implementation
			~Entry()
			{
				Dispose(false);
			}
	
			public void Dispose()
			{
				Dispose(true);
				GC.SuppressFinalize(this);
			}
	
			private bool _disposed = false;
			private void Dispose(bool commit)
			{
				if (!_disposed)
				{
					_disposed = true;
					if (commit)
					{
						if (this.Discard)
						{
							Delete();
						}
					}
					this.From = null;
					this.To = null;
					this.Subject = null;
					this.Body = null;
				}
			}
			#endregion
			
			private MessageBoard _owner;
			internal Entry(MessageBoard owner)
			{
				_owner = owner;
			}

			internal Entry(MessageBoard owner, RecordId handle)
			{
				_owner = owner;
				this.Handle = handle;
			}
			
			private string _from;
			private string _to;
			private string _subject;
			private string _body;
			
			public string From
			{
				get
				{
					return _from;
				}
				internal set
				{
					_from = value;
				}
			}
			
			public string To
			{
				get
				{
					return _to;
				}
				internal set
				{
					_to = value;
				}
			}
			
			public string Subject
			{
				get
				{
					return _subject;
				}
				internal set
				{
					_subject = value;
				}
			}
			
			public string Body
			{
				get
				{
					return _body;
				}
				internal set
				{
					_body = value;
				}
			}
			
			internal RecordId Handle;
			
			internal string FileName
			{
				get
				{
					return _owner.GetFullPath(Handle.ToString());
				}
			}
			
			internal void Save()
			{
				System.IO.TextWriter writer = new System.IO.StreamWriter(FileName, false, System.Text.Encoding.Unicode);
				try
				{
					writer.WriteLine(To);
					writer.WriteLine(Subject);
					writer.Write(Body);
				}
				finally
				{
					writer.Dispose();
				}
			}
			
			public void Delete()
			{
				Discard = false;
				_owner.InternalDelete(this);
			}
			
			private System.IO.TextReader _reader = null;
			
			internal void OpenReader()
			{
				_reader = new System.IO.StreamReader(FileName, System.Text.Encoding.Unicode, true);
			}
			
			internal void CloseReader()
			{
				_reader.Dispose();
				_reader = null;
			}
			
			internal void ReadHeader()
			{
				To = _reader.ReadLine();
				Subject = _reader.ReadLine();
			}
			
			internal void ReadBody()
			{
				Body = _reader.ReadToEnd();
			}
			
			internal bool Discard;
		}
		
		private Mutex _lock = null;
		private string _sharedFolder = null;
		private string _signature;
		
		private System.IO.PersistentList _queue = null;
		private System.Collections.Generic.IEnumerator<RecordId> _queueReader = null;

		private MessageBoard()
		{
		}
		
		public MessageBoard(string id)
		{
			Initialize(id);
		}
		
		public MessageBoard(Guid id)
		{
			Initialize(id.ToString("N").ToLower());
		}
		
		private void Initialize(string id)
		{
			bool createdNew;

			InitializeSharedFolder(id);
			
			_lock = new Mutex(true, "MessageBoard" + id, out createdNew);
			
			_signature = Guid.NewGuid().ToString("N").ToLower();
		
			if (createdNew)
			{
				_lockCount = 1;
				try
				{
					// clear previous crash data
					string[] files = System.IO.Directory.GetFiles(_sharedFolder, "*.*");
					foreach(string fileName in files)
					{
						System.IO.File.Delete(fileName);
					}
					
					Initialize();
					Connect();
				}
				finally
				{
					UnLock();
				}
			}
			else
			{
				Lock();
				try
				{
					Connect();
				}
				finally
				{
					UnLock();
				}
			}
		}

		public string Signature
		{
			get
			{
				return _signature;
			}
		}
		
		public void Send(string to, string subject, string body)
		{
			Entry message = new Entry(this);
			try
			{
				message.From = _signature;
				if (String.IsNullOrEmpty(to))
				{
					message.To = ToEveryone;
				}
				else
				{
					message.To = to;
				}
				message.Subject = subject;
				message.Body = body;
				Lock();
				try
				{
					message.Handle = _queue.Add(message.From);
					message.Save();
				}
				finally
				{
					UnLock();
				}
			}
			finally
			{
				message.Dispose();
			}
		}
		
		private void InternalDelete(Entry message)
		{
			if (message.Handle != 0)
			{
				Lock();
				try
				{
					System.IO.File.Delete(message.FileName);
					_queue.Remove(message.Handle);
					message.Handle = 0;
				}
				finally
				{
					UnLock();
				}
			}
		}
		
		public Entry Receive()
		{
			Lock();
			try
			{
				while(true)
				{
					if (_queueReader.MoveNext())
					{
						string sender = _queue.GetDataAsString(_queueReader.Current);
						if (sender != this.Signature)
						{
							Entry result = new Entry(this, _queueReader.Current);
							result.From = sender;
							
							bool acceptMessage = false;
							bool discardMessage = false;
							
							result.OpenReader();
							try
							{
								result.ReadHeader();
								
								if (result.To == ToEveryone)
								{
									acceptMessage = true;
								}
								else if (result.To == this.Signature)
								{
									acceptMessage = true;
									discardMessage = true;
								}
								
								if (acceptMessage)
								{
									result.ReadBody();
								}
							}
							finally
							{
								result.CloseReader();
							}
							
							if (acceptMessage)
							{
								result.Discard = discardMessage;
								return result;
							}
						}
					}
					else
					{
						return null;
					}
				}
			}
			finally
			{
				UnLock();
			}
		}
		
		private void InitializeSharedFolder(string id)
		{
			id = id.Replace("{", "").Replace("}", "").Replace("-", "");
			
			string folder = System.Environment.GetFolderPath(System.Environment.SpecialFolder.CommonApplicationData);
			if (String.IsNullOrEmpty(folder))
			{
				throw new AccessViolationException("Can not create shared folder");
			}
			folder = System.IO.Path.Combine(folder, "System.Threading.MessageBoard");
			System.IO.Directory.CreateDirectory(folder);
			folder = System.IO.Path.Combine(folder, id);
			System.IO.Directory.CreateDirectory(folder);
			if (folder.EndsWith("\\"))
			{
				folder = folder.Substring(0, folder.Length - 1);
			}
			_sharedFolder = folder;
		}

		protected string GetFullPath(string fileName)
		{
			return System.IO.Path.Combine(_sharedFolder, fileName);
		}
		
		private int _lockCount = 0;
		protected void Lock()
		{
			lock(this)
			{
				if (_lockCount == 0)
				{
					if (!_lock.WaitOne())
					{
						throw new ApplicationException("Lock failed");
					}
				}
				_lockCount++;
			}
		}
		
		protected void UnLock()
		{
			lock(this)
			{
				_lockCount--;
				if (_lockCount == 0)
				{
					_lock.ReleaseMutex();
				}
			}
		}

		/// <summary>
		/// Populates the tuple space with initial data
		/// </summary>
		protected virtual void Initialize()
		{
		}
		
		/// <summary>
		/// Connects to already initialized shared folder
		/// </summary>
		protected virtual void Connect()
		{
			_queue = new System.IO.PersistentList(GetFullPath("messages.queue"), (uint)(sizeof(char) * 2 * 16));
			_queueReader = _queue.GetEnumerator();
		}
		
		/// <summary>
		/// Disconnects from shared folder previously opened with Connect()
		/// </summary>
		protected virtual void Disconnect()
		{
			if (_queueReader != null)
			{
				_queueReader.Dispose();
				_queueReader = null;
			}
			if (_queue != null)
			{
				_queue = null;
			}
		}
		
		#region System.IDisposable interface implementation
		~MessageBoard()
		{
			Dispose(false);
		}

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private bool _disposed = false;
		private void Dispose(bool commit)
		{
			if (!_disposed)
			{
				_disposed = true;
				if (commit)
				{
					try
					{
						Lock();
						try
						{
							Disconnect();
						}
						finally
						{
							UnLock();
						}
					}
					catch(Exception)
					{
					}
				}
				if (_lock != null)
				{
					_lock.Close();
					_lock = null;
				}
			}
		}
		#endregion
	}
}
