using System;

namespace System.Threading
{
	public interface IProgress
	{
		void Open(ulong min, ulong max);
		void Write(string message);
		bool Step(ulong value);
		void Abort();
		bool Aborted { get; }
		void Close();
	}

	public abstract class AbstractProgress: IProgress
	{
		public abstract void Open(ulong min, ulong max);
		public abstract void Write(string message);
		public abstract bool Step(ulong value);
		private bool _aborted = false;
		public virtual void Abort()
		{
			_aborted = true;
		}
		public virtual bool Aborted
		{
			get
			{
				return _aborted;
			}
		}
		public abstract void Close();
	}

	public class NullProgress: AbstractProgress
	{
		public override void Open(ulong min, ulong max)
		{
		}
		
		public override void Close()
		{
		}
		
		public override void Write(string message)
		{
		}
		
		public override bool Step(ulong value)
		{
			return !Aborted;
		}
	}
	
	public class SingleProgress: AbstractProgress
	{
		private IProgress _parent;
		protected ulong _parentMin;
		protected ulong _parentMax;
		protected ulong _progress = 0;
		protected ulong _maxProgress;
		protected bool _parentRangeDefined = false;
		
		protected void SetParentRange(ulong min, ulong max)
		{
			_parentMin = min;
			_parentMax = max;
			_maxProgress = _parentMax - _parentMin;
			_parentRangeDefined = true;
		}
		
		public SingleProgress(IProgress parent, ulong min, ulong max)
		{
			_parent = parent;
			SetParentRange(min, max);
		}
		
		public SingleProgress(IProgress parent)
		{
			_parent = parent;
		}
		
		protected SingleProgress()
		{
		}
		
		/// <summary>
		/// 0-based progress in parent's coordinates
		/// </summary>
		public ulong Progress
		{
			get
			{
				return _progress;
			}
		}
		
		#region IProgress interface implementation
		protected ulong _min;
		protected ulong _max;
		private double _scale;
		private bool _useScale;
		private ulong _lastStep;
		private ulong _lastStepIt;
		private bool _opened = false;
		public override void Open(ulong min, ulong max)
		{
			if (_opened)
			{
				if ((min != _min) || (max != _max))
				{
					throw new ArgumentException();
				}
				return;
			}
			_opened = true;
			_min = min;
			_max = max;
			if (!_parentRangeDefined)
			{
				SetParentRange(_min, _max);
				_useScale = false;
			}
			else
			{
				_scale = ((double)(_parentMax - _parentMin)) / ((double)(_max - _min));
				// TODO: _useScale if _scale != 1 +- epsilon
				_useScale = true;
			}
			_lastStep = _min;
			_lastStepIt = _min;
			DoOpen(_parentMin, _parentMax);
		}
		
		public override void Write(string message)
		{
			DoWrite(message);
		}
		
		private ulong _treshold = 1;
		public ulong Treshold
		{
			get
			{
				return _treshold;
			}
			set
			{
				_treshold = value;
			}
		}
		
		private ulong _increment = 1;
		public ulong Increment
		{
			get
			{
				return _increment;
			}
			set
			{
				_increment = value;
			}
		}
		
		public bool StepIt()
		{
			_lastStepIt += _increment;
			return Step(_lastStepIt);
		}
		
		public override bool Step(ulong value)
		{
			if (value < _min)
			{
				value = _min;
			}
			if (value > _max)
			{
				value = _max;
			}
			
			if (value < _lastStep + _treshold)
			{
				return !Aborted;
			}
			
			_lastStep = value;
			
			if (_useScale)
			{
				_progress = Convert.ToUInt64((value - _min) * _scale);
			}
			else
			{
				_progress = Convert.ToUInt64(value - _min);
			}
			if (_progress > _maxProgress)
			{
				_progress = _maxProgress;
			}
			if (DoStep(_parentMin + _progress))
			{
				return !Aborted;
			}
			else
			{
				return false;
			}
		}
		
		public override void Close()
		{
			if (_opened)
			{
				_opened = false;
				_progress = _maxProgress;
				DoClose();
			}
			else
			{
				if (!_parentRangeDefined)
				{
					// use any value
					SetParentRange(0, 10);
				}
				_progress = _maxProgress;
			}
		}
		
		public override bool Aborted 
		{
			get 
			{
				if (_parent == null)
				{
					return base.Aborted;
				}
				else
				{
					return base.Aborted || _parent.Aborted;
				}
			}
		}
		#endregion
		
		#region Raw interface to override
		protected virtual void DoOpen(ulong min, ulong max)
		{
			_parent.Open(min, max);
		}
		
		protected virtual void DoClose()
		{
			_parent.Close();
		}
		
		protected virtual void DoWrite(string message)
		{
			_parent.Write(message);
		}
		
		protected virtual bool DoStep(ulong value)
		{
			return _parent.Step(value);
		}
		#endregion
	}
	
	public class SlicedProgress: SingleProgress
	{
		public SlicedProgress(IProgress parent, ulong min, ulong max): base(parent, min, max)
		{
		}
		
		public SlicedProgress(IProgress parent): base(parent)
		{
		}

		protected SlicedProgress()
		{
		}
		
		private System.Collections.Generic.List<SingleProgress> _items = new System.Collections.Generic.List<SingleProgress>();

		public int Count
		{
			get
			{
				return _items.Count;
			}
		}
		
		public SingleProgress this[int index]
		{
			get
			{
				return _items[index];
			}
		}
		
		private ulong _addedMax = 0;
		public SingleProgress Add(ulong range)
		{
			SingleProgress result = new SingleProgress(this, _addedMax, _addedMax + range);
			_addedMax += range;
			_items.Add(result);
			return result;
		}
		
		private bool InternalStep()
		{
			ulong value = 0;
			foreach(SingleProgress item in _items)
			{
				value += item.Progress;
			}
			return base.Step(_parentMin + value);
		}
		
		private bool _opened = false;
		public override void Open(ulong min, ulong max)
		{
			if (!_opened)
			{
				_opened = true;
				base.Open(0, _addedMax);
			}
			InternalStep();
		}
		
		public override bool Step(ulong value)
		{
			return InternalStep();
		}
		
		public override void Close()
		{
			InternalStep();
			// TODO: close if all sup-progresses closed
			ulong max = _maxProgress;
			if (!_parentRangeDefined)
			{
				max = _addedMax;
			}
			if (_progress == max)
			{
				base.Close();
			}
		}
		
		public void Skip(SingleProgress item)
		{
			// TODO: use something more descriptive
			item.Close();
			Close();
		}
		
		public void SkipAll()
		{
			foreach(SingleProgress item in _items)
			{
				Skip(item);
			}
		}
	}
}
