using System;
using System.Runtime.InteropServices;

namespace System.Windows.Forms
{
	public class CloseButton
	{
		private const int SC_CLOSE = 0xF060;
		private const int MF_BYCOMMAND = 0x0;
		private const int MF_GRAYED = 0x1;
		private const int MF_ENABLED = 0x0;

		[DllImport("user32.dll")]
		static extern bool EnableMenuItem(IntPtr hMenu, uint uIDEnableItem, uint uEnable);

		[DllImport("user32.dll")]
		static extern IntPtr GetSystemMenu(IntPtr hWnd, bool bRevert);		

		private class Disablers: System.Collections.Generic.Dictionary<Form, Disabler>
		{
		}
		
		private static Disablers _disablers = null;
		
		private class Disabler
		{
			private Form _owner;
			public Disabler(Form owner)
			{
				_owner = owner;
				_owner.SizeChanged += this.HandleSizeChanged;
			}
			
			public void HandleSizeChanged(object sender, EventArgs e)
			{
				EnableMenuItem(GetSystemMenu(_owner.Handle, false), SC_CLOSE, MF_BYCOMMAND | MF_GRAYED);
			}
			
			public void Close()
			{
				_owner.SizeChanged -= this.HandleSizeChanged;
			}
		}
		
		public static void SetEnabled(Form owner, bool value)
		{
			if (value)
			{
				EnableMenuItem(GetSystemMenu(owner.Handle, false), SC_CLOSE, MF_BYCOMMAND | MF_ENABLED);
				if (_disablers != null)
				{
					Disabler hook;
					if (_disablers.TryGetValue(owner, out hook))
					{
						hook.Close();
						_disablers.Remove(owner);
					}
				}
			}
			else
			{
				EnableMenuItem(GetSystemMenu(owner.Handle, false), SC_CLOSE, MF_BYCOMMAND | MF_GRAYED);
				if (_disablers == null)
				{
					_disablers = new Disablers();
				}
				if (!_disablers.ContainsKey(owner))
				{
					_disablers.Add(owner, new Disabler(owner));
				}
			}
		}
		
		public static bool GetEnabled(Form owner)
		{
			if (_disablers != null)
			{
				if (_disablers.ContainsKey(owner))
				{
					return true;
				}
			}
			return false;
		}
	}
	
	public class LockUpdate: IDisposable
	{
		IntPtr pWindow = IntPtr.Zero;

		public LockUpdate(Form form)
		{
			pWindow = form.Handle;
			BeginUpdate();
		}

		[DllImport("user32.dll")] 
		private static extern Int32 LockWindowUpdate(IntPtr wnd);

		private void BeginUpdate()
		{
			LockWindowUpdate(pWindow);
		}

		private void EndUpdate()
		{
			LockWindowUpdate(IntPtr.Zero);
		}

		#region Implementation of IDisposable
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			if (pWindow != IntPtr.Zero)
			{
				EndUpdate();
				pWindow = IntPtr.Zero;
			}
		}

		~LockUpdate()
		{
			Dispose(false);
		}
		#endregion
	}

	/// <summary>
	/// Implementation of a fix described in article http://forums.microsoft.com/MSDN/ShowPost.aspx?PostID=3739275&SiteID=1
	/// </summary>
	public class FocusFix
	{
		private Form _owner;
		private bool _enabled;
		
		public FocusFix(Form form)
		{
			_owner = form;
			_owner.Shown += this.ApplyFix;
			_enabled = true;
		}

		#region Win32 API
        // Messages
        private const int WM_UPDATEUISTATE = 0x0128;

        // UI State Parameters
        private const int UIS_SET = 1;
        private const int UIS_CLEAR = 2;

        // UI State Flags
        private const int UISF_HIDEFOCUS = 0x1;

        // Helper Functions
        private static int MakeParam(int loWord, int hiWord)
        {
        	return (hiWord << 16) | (loWord & 0xffff);
        }

        [DllImport("user32.dll")]
        private extern static int SendMessage(IntPtr hWnd, uint msg, int wParam, int lParam);
		#endregion      
                        
        public void ApplyFix(object sender, EventArgs e)
		{
			if (!_enabled)
			{
				return;
			}
			_enabled = false;
			
       		// Ensure focus rectange is shown 
            SendMessage(_owner.Handle, WM_UPDATEUISTATE, MakeParam(UIS_CLEAR, UISF_HIDEFOCUS), 0);
		}
	}
}
